=== modified file 'NEWS'
--- NEWS	2007-07-19 06:42:37 +0000
+++ NEWS	2007-07-19 13:13:06 +0000
@@ -40,6 +40,9 @@
     * The ``plugins`` command is now public with improved help.
       (Ian Clatworthy)
 
+    * New bundle and merge directive formats are faster to generate, and
+      more robust against email mangling.  (Aaron Bentley)
+
   LIBRARY API BREAKS:
 
     * Deprecated dictionary ``bzrlib.option.SHORT_OPTIONS`` removed.

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2007-07-19 07:37:23 +0000
+++ bzrlib/builtins.py	2007-07-19 13:13:06 +0000
@@ -53,7 +53,6 @@
     urlutils,
     )
 from bzrlib.branch import Branch
-from bzrlib.bundle.apply_bundle import install_bundle, merge_bundle
 from bzrlib.conflicts import ConflictList
 from bzrlib.revisionspec import RevisionSpec
 from bzrlib.smtp_connection import SMTPConnection
@@ -618,7 +617,9 @@
             if revision is not None:
                 raise errors.BzrCommandError(
                     'Cannot use -r with merge directives or bundles')
-            revision_id = mergeable.install_revisions(branch_to.repository)
+            mergeable.install_revisions(branch_to.repository)
+            base_revision_id, revision_id, verified = \
+                mergeable.get_merge_request(branch_to.repository)
             branch_from = branch_to
         else:
             branch_from = Branch.open(location)
@@ -2641,6 +2642,7 @@
             ):
         from bzrlib.tag import _merge_tags_if_possible
         other_revision_id = None
+        base_revision_id = None
         if merge_type is None:
             merge_type = _mod_merge.Merge3Merger
 
@@ -2666,52 +2668,57 @@
                 if revision is not None:
                     raise errors.BzrCommandError(
                         'Cannot use -r with merge directives or bundles')
-                other_revision_id = mergeable.install_revisions(
-                    tree.branch.repository)
-                revision = [RevisionSpec.from_string(
-                    'revid:' + other_revision_id)]
-
-        if revision is None \
-                or len(revision) < 1 or revision[0].needs_branch():
-            branch = self._get_remembered_parent(tree, branch, 'Merging from')
-
-        if revision is None or len(revision) < 1:
-            if uncommitted:
-                base = [branch, -1]
-                other = [branch, None]
+                mergeable.install_revisions(tree.branch.repository)
+                base_revision_id, other_revision_id, verified =\
+                    mergeable.get_merge_request(tree.branch.repository)
+                if base_revision_id in tree.branch.repository.get_ancestry(
+                    tree.branch.last_revision(), topo_sorted=False):
+                    base_revision_id = None
+                other_branch = None
+                path = ''
+                other = None
+                base = None
+
+        if other_revision_id is None:
+            verified = 'inapplicable'
+            if revision is None \
+                    or len(revision) < 1 or revision[0].needs_branch():
+                branch = self._get_remembered_parent(tree, branch,
+                    'Merging from')
+
+            if revision is None or len(revision) < 1:
+                if uncommitted:
+                    base = [branch, -1]
+                    other = [branch, None]
+                else:
+                    base = [None, None]
+                    other = [branch, -1]
+                other_branch, path = Branch.open_containing(branch)
             else:
-                base = [None, None]
-                other = [branch, -1]
-            other_branch, path = Branch.open_containing(branch)
-        else:
-            if uncommitted:
-                raise errors.BzrCommandError('Cannot use --uncommitted and'
-                                             ' --revision at the same time.')
-            branch = revision[0].get_branch() or branch
-            if len(revision) == 1:
-                base = [None, None]
-                if other_revision_id is not None:
-                    other_branch = None
-                    path = ""
-                    other = None
-                else:
+                if uncommitted:
+                    raise errors.BzrCommandError('Cannot use --uncommitted and'
+                        ' --revision at the same time.')
+                branch = revision[0].get_branch() or branch
+                if len(revision) == 1:
+                    base = [None, None]
                     other_branch, path = Branch.open_containing(branch)
                     revno = revision[0].in_history(other_branch).revno
                     other = [branch, revno]
-            else:
-                assert len(revision) == 2
-                if None in revision:
-                    raise errors.BzrCommandError(
-                        "Merge doesn't permit empty revision specifier.")
-                base_branch, path = Branch.open_containing(branch)
-                branch1 = revision[1].get_branch() or branch
-                other_branch, path1 = Branch.open_containing(branch1)
-                if revision[0].get_branch() is not None:
-                    # then path was obtained from it, and is None.
-                    path = path1
+                else:
+                    assert len(revision) == 2
+                    if None in revision:
+                        raise errors.BzrCommandError(
+                            "Merge doesn't permit empty revision specifier.")
+                    base_branch, path = Branch.open_containing(branch)
+                    branch1 = revision[1].get_branch() or branch
+                    other_branch, path1 = Branch.open_containing(branch1)
+                    if revision[0].get_branch() is not None:
+                        # then path was obtained from it, and is None.
+                        path = path1
 
-                base = [branch, revision[0].in_history(base_branch).revno]
-                other = [branch1, revision[1].in_history(other_branch).revno]
+                    base = [branch, revision[0].in_history(base_branch).revno]
+                    other = [branch1,
+                             revision[1].in_history(other_branch).revno]
 
         if ((tree.branch.get_parent() is None or remember) and
             other_branch is not None):
@@ -2731,6 +2738,7 @@
             try:
                 conflict_count = _merge_helper(
                     other, base, other_rev_id=other_revision_id,
+                    base_rev_id=base_revision_id,
                     check_clean=(not force),
                     merge_type=merge_type,
                     reprocess=reprocess,
@@ -2741,6 +2749,8 @@
                     change_reporter=change_reporter)
             finally:
                 pb.finished()
+            if verified == 'failed':
+                warning('Preview patch does not match changes')
             if conflict_count != 0:
                 return 1
             else:
@@ -3574,6 +3584,10 @@
 
     takes_args = ['submit_branch?', 'public_branch?']
 
+    hidden = True
+
+    _see_also = ['submit']
+
     takes_options = [
         RegistryOption.from_kwargs('patch-type',
             'The type of patch to include in the directive',
@@ -3595,8 +3609,11 @@
     def run(self, submit_branch=None, public_branch=None, patch_type='bundle',
             sign=False, revision=None, mail_to=None, message=None):
         from bzrlib.revision import ensure_null, NULL_REVISION
-        if patch_type == 'plain':
-            patch_type = None
+        include_patch, include_bundle = {
+            'plain': (False, False),
+            'diff': (True, False),
+            'bundle': (True, True),
+            }[patch_type]
         branch = Branch.open('.')
         stored_submit_branch = branch.get_submit_branch()
         if submit_branch is None:
@@ -3614,25 +3631,29 @@
             public_branch = stored_public_branch
         elif stored_public_branch is None:
             branch.set_public_branch(public_branch)
-        if patch_type != "bundle" and public_branch is None:
+        if not include_bundle and public_branch is None:
             raise errors.BzrCommandError('No public branch specified or'
                                          ' known')
+        base_revision_id = None
         if revision is not None:
-            if len(revision) != 1:
+            if len(revision) > 2:
                 raise errors.BzrCommandError('bzr merge-directive takes '
-                    'exactly one revision identifier')
-            else:
-                revision_id = revision[0].in_history(branch).rev_id
+                    'at most two one revision identifiers')
+            revision_id = revision[-1].in_history(branch).rev_id
+            if len(revision) == 2:
+                base_revision_id = revision[0].in_history(branch).rev_id
+                base_revision_id = ensure_null(base_revision_id)
         else:
             revision_id = branch.last_revision()
         revision_id = ensure_null(revision_id)
         if revision_id == NULL_REVISION:
             raise errors.BzrCommandError('No revisions to bundle.')
-        directive = merge_directive.MergeDirective.from_objects(
+        directive = merge_directive.MergeDirective2.from_objects(
             branch.repository, revision_id, time.time(),
             osutils.local_time_offset(), submit_branch,
-            public_branch=public_branch, patch_type=patch_type,
-            message=message)
+            public_branch=public_branch, include_patch=include_patch,
+            include_bundle=include_bundle, message=message,
+            base_revision_id=base_revision_id)
         if mail_to is None:
             if sign:
                 self.outf.write(directive.to_signed(branch))
@@ -3644,6 +3665,118 @@
             s.send_email(message)
 
 
+class cmd_submit(Command):
+    """Create a merge-directive for submiting changes.
+
+    A merge directive provides many things needed for requesting merges:
+    - A machine-readable description of the merge to perform
+    - An optional patch that is a preview of the changes requested
+    - An optional bundle of revision data, so that the changes can be applied
+      directly from the merge directive, without retrieving data from a
+      branch.
+
+    If --no-bundle is specified, then public_branch is needed (and must be
+    up-to-date), so that the receiver can perform the merge using the
+    public_branch.  The public_branch is always included if known, so that
+    people can check it later.
+
+    The submit branch defaults to the parent, but can be overridden.  Both
+    submit branch and public branch will be remembered if supplied.
+
+    If a public_branch is known for the submit_branch, that public submit
+    branch is used in the merge instructions.  This means that a local mirror
+    can be used as your actual submit branch, once you have set public_branch
+    for that mirror.
+    """
+
+    encoding_type = 'exact'
+
+    aliases = ['bundle', 'bundle-revisions']
+
+    _see_also = ['merge']
+
+    takes_args = ['submit_branch?', 'public_branch?']
+    takes_options = [
+        Option('no-bundle',
+               help='Do not include a bundle in the merge directive.'),
+        Option('no-patch', help='Do not include a preview patch in the merge'
+               ' directive.'),
+        Option('remember',
+               help='Remember submit and public branch.'),
+        Option('from',
+               help='Branch to generate the submission from, '
+               'rather than the one containing the working directory.',
+               short_name='f',
+               type=unicode),
+        Option('output', short_name='o', help='Write directive to this file.',
+               type=unicode),
+        'revision',
+        ]
+
+    def run(self, submit_branch=None, public_branch=None, no_bundle=False,
+            no_patch=False, revision=None, remember=False, output=None,
+            **kwargs):
+        from bzrlib.revision import ensure_null, NULL_REVISION
+        if output is None:
+            outfile = self.outf
+        else:
+            outfile = open(output, 'wb')
+        try:
+            from_ = kwargs.get('from', '.')
+            branch = Branch.open_containing(from_)[0]
+            if remember and submit_branch is None:
+                raise errors.BzrCommandError(
+                    '--remember requires a branch to be specified.')
+            stored_submit_branch = branch.get_submit_branch()
+            remembered_submit_branch = False
+            if submit_branch is None:
+                submit_branch = stored_submit_branch
+                remembered_submit_branch = True
+            else:
+                if stored_submit_branch is None or remember:
+                    branch.set_submit_branch(submit_branch)
+            if submit_branch is None:
+                submit_branch = branch.get_parent()
+                remembered_submit_branch = True
+            if submit_branch is None:
+                raise errors.BzrCommandError('No submit branch known or'
+                                             ' specified')
+            if remembered_submit_branch:
+                note('Using saved location: %s', submit_branch)
+
+            stored_public_branch = branch.get_public_branch()
+            if public_branch is None:
+                public_branch = stored_public_branch
+            elif stored_public_branch is None or remember:
+                branch.set_public_branch(public_branch)
+            if no_bundle and public_branch is None:
+                raise errors.BzrCommandError('No public branch specified or'
+                                             ' known')
+            base_revision_id = None
+            if revision is not None:
+                if len(revision) > 2:
+                    raise errors.BzrCommandError('bzr submit takes '
+                        'at most two one revision identifiers')
+                revision_id = revision[-1].in_history(branch).rev_id
+                if len(revision) == 2:
+                    base_revision_id = revision[0].in_history(branch).rev_id
+                    base_revision_id = ensure_null(base_revision_id)
+            else:
+                revision_id = branch.last_revision()
+            revision_id = ensure_null(revision_id)
+            if revision_id == NULL_REVISION:
+                raise errors.BzrCommandError('No revisions to submit.')
+            directive = merge_directive.MergeDirective2.from_objects(
+                branch.repository, revision_id, time.time(),
+                osutils.local_time_offset(), submit_branch,
+                public_branch=public_branch, include_patch=not no_patch,
+                include_bundle=not no_bundle, message=None,
+                base_revision_id=base_revision_id)
+            outfile.writelines(directive.to_lines())
+        finally:
+            if output is not None:
+                outfile.close()
+
 class cmd_tag(Command):
     """Create a tag naming a revision.
     
@@ -3737,7 +3870,7 @@
                   pull=False,
                   pb=DummyProgress(),
                   change_reporter=None,
-                  other_rev_id=None):
+                  other_rev_id=None, base_rev_id=None):
     """Merge changes into a tree.
 
     base_revision
@@ -3798,7 +3931,12 @@
         else:
             merger.set_other(other_revision)
         merger.pp.next_phase()
-        merger.set_base(base_revision)
+        if base_rev_id is not None:
+            merger.set_base_revision(base_rev_id, this_tree.branch)
+        elif base_revision is not None:
+            merger.set_base(base_revision)
+        else:
+            merger.find_base()
         if merger.base_rev_id == merger.other_rev_id:
             note('Nothing to do.')
             return 0
@@ -3858,7 +3996,9 @@
 # details were needed.
 from bzrlib.cmd_version_info import cmd_version_info
 from bzrlib.conflicts import cmd_resolve, cmd_conflicts, restore
-from bzrlib.bundle.commands import cmd_bundle_revisions
+from bzrlib.bundle.commands import (
+    cmd_bundle_info,
+    )
 from bzrlib.sign_my_commits import cmd_sign_my_commits
 from bzrlib.weave_commands import cmd_versionedfile_list, cmd_weave_join, \
         cmd_weave_plan_merge, cmd_weave_merge_text

=== modified file 'bzrlib/bundle/apply_bundle.py'
--- bzrlib/bundle/apply_bundle.py	2007-02-09 15:56:49 +0000
+++ bzrlib/bundle/apply_bundle.py	2007-06-11 14:59:52 +0000
@@ -25,6 +25,9 @@
 
 
 def install_bundle(repository, bundle_reader):
+    custom_install = getattr(bundle_reader, 'install', None)
+    if custom_install is not None:
+        return custom_install(repository)
     pb = bzrlib.ui.ui_factory.nested_progress_bar()
     repository.lock_write()
     try:

=== modified file 'bzrlib/bundle/bundle_data.py'
--- bzrlib/bundle/bundle_data.py	2007-05-03 16:42:30 +0000
+++ bzrlib/bundle/bundle_data.py	2007-06-28 14:14:16 +0000
@@ -23,6 +23,7 @@
 
 from bzrlib import (
     osutils,
+    timestamp,
     )
 import bzrlib.errors
 from bzrlib.bundle import apply_bundle
@@ -87,12 +88,26 @@
 
         return rev
 
+    @staticmethod
+    def from_revision(revision):
+        revision_info = RevisionInfo(revision.revision_id)
+        date = timestamp.format_highres_date(revision.timestamp,
+                                             revision.timezone)
+        revision_info.date = date
+        revision_info.timezone = revision.timezone
+        revision_info.timestamp = revision.timestamp
+        revision_info.message = revision.message.split('\n')
+        revision_info.properties = [': '.join(p) for p in
+                                    revision.properties.iteritems()]
+        return revision_info
+
 
 class BundleInfo(object):
     """This contains the meta information. Stuff that allows you to
     recreate the revision or inventory XML.
     """
-    def __init__(self):
+    def __init__(self, bundle_format=None):
+        self.bundle_format = None
         self.committer = None
         self.date = None
         self.message = None
@@ -447,6 +462,13 @@
         apply_bundle.install_bundle(target_repo, self)
         return self.target
 
+    def get_merge_request(self, target_repo):
+        """Provide data for performing a merge
+
+        Returns suggested base, suggested target, and patch verification status
+        """
+        return None, self.target, 'inapplicable'
+
 
 class BundleTree(Tree):
     def __init__(self, base_tree, revision_id):

=== modified file 'bzrlib/bundle/commands.py'
--- bzrlib/bundle/commands.py	2007-07-12 07:37:18 +0000
+++ bzrlib/bundle/commands.py	2007-07-17 13:27:14 +0000
@@ -21,14 +21,17 @@
 """
 
 import sys
+from cStringIO import StringIO
 
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), """
 from bzrlib import (
     branch,
     errors,
+    merge_directive,
     revision as _mod_revision,
     urlutils,
+    transport,
     )
 """)
 
@@ -37,163 +40,59 @@
 from bzrlib.trace import note
 
 
-class cmd_send_changeset(Command):
-    """Send a bundled up changset via mail.
-
-    If no revision has been specified, the last commited change will
-    be sent.
-
-    Subject of the mail can be specified by the --message option,
-    otherwise information from the changeset log will be used.
-
-    A editor will be spawned where the user may enter a description
-    of the changeset.  The description can be read from a file with
-    the --file FILE option.
-    """
-    takes_options = ['revision', 'message', 'file']
-    takes_args = ['to?']
-
-    def run(self, to=None, message=None, revision=None, file=None):
-        from bzrlib.errors import BzrCommandError
-        from send_changeset import send_changeset
-        
-        if isinstance(revision, (list, tuple)):
-            if len(revision) > 1:
-                raise BzrCommandError('We do not support rollup-changesets yet.')
-            revision = revision[0]
-
-        b = branch.Branch.open_containing('.')
-
-        if not to:
-            try:
-                to = b.controlfile('x-send-address', 'rb').read().strip('\n')
-            except:
-                raise BzrCommandError('destination address is not known')
-
-        if not isinstance(revision, (list, tuple)):
-            revision = [revision]
-
-        send_changeset(b, revision, to, message, file)
-
-
-class cmd_bundle_revisions(Command):
-    """Generate a revision bundle.
-
-    This bundle contains all of the meta-information of a
-    diff, rather than just containing the patch information.
-
-    You can apply it to another tree using 'bzr merge'.
-
-    bzr bundle-revisions
-        - Generate a bundle relative to a remembered location
-
-    bzr bundle-revisions BASE
-        - Bundle to apply the current tree into BASE
-
-    bzr bundle-revisions --revision A
-        - Bundle to apply revision A to remembered location 
-
-    bzr bundle-revisions --revision A..B
-        - Bundle to transform A into B
-    """
-    takes_options = ['revision', 'remember',
-                     Option("output",
-                            help="Write bundle to specified file.",
-                            type=unicode),
-                     ]
-    takes_args = ['base?']
-    aliases = ['bundle']
+class cmd_bundle_info(Command):
+    """Output interesting stats about a bundle"""
+
+    hidden = True
+    takes_args = ['location']
+    takes_options = []
     encoding_type = 'exact'
 
-    def run(self, base=None, revision=None, output=None, remember=False):
-        from bzrlib import user_encoding
-        from bzrlib.bundle.serializer import write_bundle
-
-        target_branch = branch.Branch.open_containing(u'.')[0]
-
-        if base is None:
-            base_specified = False
-        else:
-            base_specified = True
-
-        if revision is None:
-            target_revision = _mod_revision.ensure_null(
-                target_branch.last_revision())
-        elif len(revision) < 3:
-            target_revision = revision[-1].in_history(target_branch).rev_id
-            if len(revision) == 2:
-                if base_specified:
-                    raise errors.BzrCommandError('Cannot specify base as well'
-                                                 ' as two revision arguments.')
-                base_revision = revision[0].in_history(target_branch).rev_id
-        else:
-            raise errors.BzrCommandError('--revision takes 1 or 2 parameters')
-
-        if revision is None or len(revision) < 2:
-            submit_branch = target_branch.get_submit_branch()
-            if base is None:
-                base = submit_branch
-            if base is None:
-                base = target_branch.get_parent()
-            if base is None:
-                raise errors.BzrCommandError("No base branch known or"
-                                             " specified.")
-            elif not base_specified:
-                # FIXME:
-                # note() doesn't pay attention to terminal_encoding() so
-                # we must format with 'ascii' to be safe
-                note('Using saved location: %s',
-                     urlutils.unescape_for_display(base, 'ascii'))
-            base_branch = branch.Branch.open(base)
-
-            # We don't want to lock the same branch across
-            # 2 different branches
-            if target_branch.base == base_branch.base:
-                base_branch = target_branch 
-            if submit_branch is None or remember:
-                if base_specified:
-                    target_branch.set_submit_branch(base_branch.base)
-                elif remember:
-                    raise errors.BzrCommandError('--remember requires a branch'
-                                                 ' to be specified.')
-            base_last_revision = _mod_revision.ensure_null(
-                base_branch.last_revision())
-            target_branch.repository.fetch(base_branch.repository, 
-                base_last_revision)
-            graph = target_branch.repository.get_graph()
-            base_revision = graph.find_unique_lca(base_last_revision,
-                                                  target_revision)
-
-        if output is not None:
-            fileobj = file(output, 'wb')
-        else:
-            fileobj = sys.stdout
-        target_branch.repository.lock_read()
-        try:
-            write_bundle(target_branch.repository, target_revision,
-                         base_revision, fileobj)
-        finally:
-            target_branch.repository.unlock()
-
-
-class cmd_verify_changeset(Command):
-    """Read a written changeset, and make sure it is valid.
-
-    """
-    takes_args = ['filename?']
-
-    def run(self, filename=None):
-        from read_changeset import read_changeset
-        #from bzrlib.xml import serializer_v4
-
-        b, relpath = branch.Branch.open_containing('.')
-
-        if filename is None or filename == '-':
-            f = sys.stdin
-        else:
-            f = open(filename, 'U')
-
-        cset_info, cset_tree = read_changeset(f, b)
-        # print cset_info
-        # print cset_tree
-        #serializer_v4.write(cset_tree.inventory, sys.stdout)
+    def run(self, location, verbose=False):
+        from bzrlib.bundle.serializer import read_bundle
+        from bzrlib.bundle import read_mergeable_from_url
+        from bzrlib import osutils
+        term_encoding = osutils.get_terminal_encoding()
+        bundle_info = read_mergeable_from_url(location)
+        if isinstance(bundle_info, merge_directive._BaseMergeDirective):
+            bundle_info = read_bundle(StringIO(bundle_info.get_raw_bundle()))
+        reader_method = getattr(bundle_info, 'get_bundle_reader', None)
+        if reader_method is None:
+            raise errors.BzrCommandError('Bundle format not supported')
+
+        by_kind = {}
+        file_ids = set()
+        for bytes, parents, repo_kind, revision_id, file_id\
+            in reader_method().iter_records():
+            by_kind.setdefault(repo_kind, []).append(
+                (bytes, parents, repo_kind, revision_id, file_id))
+            if file_id is not None:
+                file_ids.add(file_id)
+        print >> self.outf, 'Records'
+        for kind, records in sorted(by_kind.iteritems()):
+            multiparent = sum(1 for b, p, k, r, f in records if len(p) > 1)
+            print >> self.outf, '%s: %d (%d multiparent)' % \
+                (kind, len(records), multiparent)
+        print >> self.outf, 'unique files: %d' % len(file_ids)
+        print >> self.outf
+        nicks = set()
+        committers = set()
+        for revision in bundle_info.real_revisions:
+            if 'branch-nick' in revision.properties:
+                nicks.add(revision.properties['branch-nick'])
+            committers.add(revision.committer)
+
+        print >> self.outf, 'Revisions'
+        print >> self.outf, ('nicks: %s'
+            % ', '.join(sorted(nicks))).encode(term_encoding, 'replace')
+        print >> self.outf, ('committers: \n%s' %
+        '\n'.join(sorted(committers)).encode(term_encoding, 'replace'))
+        if verbose:
+            print >> self.outf
+            bundle_file.seek(0)
+            line = bundle_file.readline()
+            line = bundle_file.readline()
+            content = bundle_file.read().decode('bz2')
+            print >> self.outf, "Decoded contents"
+            self.outf.write(content)
+            print >> self.outf

=== removed file 'bzrlib/bundle/common.py'
--- bzrlib/bundle/common.py	2006-10-05 05:37:25 +0000
+++ bzrlib/bundle/common.py	1970-01-01 00:00:00 +0000
@@ -1,35 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-"""\
-Common entries, like strings, etc, for the bundle reading + writing code.
-"""
-
-import bzrlib
-
-header_str = 'Bazaar revision bundle v'
-version = (0, 8)
-
-
-def get_header():
-    return [
-        header_str + '.'.join([str(v) for v in version]),
-        ''
-    ]
-
-      
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()

=== removed directory 'bzrlib/bundle/old'
=== removed file 'bzrlib/bundle/old/send_changeset.py'
--- bzrlib/bundle/old/send_changeset.py	2006-10-05 05:37:25 +0000
+++ bzrlib/bundle/old/send_changeset.py	1970-01-01 00:00:00 +0000
@@ -1,79 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""\
-Generate a changeset and send it by mail.
-"""
-
-import bzrlib, bzrlib.changeset
-import common, smtplib
-
-from email import Encoders
-from email.Message import Message
-from email.MIMEBase import MIMEBase
-from email.MIMEMultipart import MIMEMultipart
-from email.MIMEText import MIMEText
-
-
-def send_changeset(branch, revisions, to_address, message, file):
-    import bzrlib.osutils
-    from bzrlib import find_branch
-    from bzrlib.commands import BzrCommandError
-    import gen_changeset
-    import send_changeset
-    from cStringIO import StringIO
-
-    base_rev_id, target_rev_id = common.canonicalize_revision(branch, revisions)
-    rev = branch.get_revision(target_rev_id)
-    if not message:
-        message = rev.message.split('\n')[0]
-
-    from_address = bzrlib.osutils._get_user_id()
-
-    outer = MIMEMultipart()
-    outer['Subject'] = '[PATCH] ' + message
-    outer['To'] = to_address
-    outer['From'] = from_address
-
-    # Either read the mail body from the specified file, or spawn
-    # an editor and let the user type a description.
-    if file:
-        mail_body = open(file, "U").read()
-    else:
-        info = "Changeset by %s\n" % rev.committer
-        info += "From %s\n" % base_rev_id
-        info += "with the following message:\n"
-        for line in rev.message.split('\n'):
-            info += "  " + line + "\n"
-
-        mail_body = bzrlib.osutils.get_text_message(info)
-        if mail_body is None:
-            raise BzrCommandError("aborted")
-    outer.attach(MIMEText(mail_body))
-    
-    changeset_fp = StringIO()
-    gen_changeset.show_changeset(branch, revisions, to_file=changeset_fp)
-    outer.attach(MIMEText(changeset_fp.getvalue()))
-
-    try:
-        fp = open(os.path.join(bzrlib.osutils.config_dir(), 'smtp-host'), 'U')
-        smtpconn = smtplib.SMTP(fp.readline().strip('\n'))
-    except:
-        smtpconn = smtplib.SMTP()
-
-    smtpconn.connect()
-    smtpconn.sendmail(from_address, to_address, outer.as_string())
-    smtpconn.close()

=== modified file 'bzrlib/bundle/serializer/__init__.py'
--- bzrlib/bundle/serializer/__init__.py	2007-07-11 16:42:37 +0000
+++ bzrlib/bundle/serializer/__init__.py	2007-07-17 13:27:14 +0000
@@ -38,6 +38,10 @@
 
 _serializers = {}
 
+v4_string = '4alpha'
+
+def _get_bundle_header(version):
+    return '%s%s\n' % (BUNDLE_HEADER, version)
 
 def _get_filename(f):
     return getattr(f, 'name', '<unknown>')
@@ -79,6 +83,13 @@
     return serializer.read(f)
 
 
+def get_serializer(version):
+    try:
+        return _serializers[version](version)
+    except KeyError:
+        raise errors.BundleNotSupported(version, 'unknown bundle format')
+
+
 def write(source, revision_ids, f, version=None, forced_bases={}):
     """Serialize a list of bundles to a filelike object.
 
@@ -88,28 +99,15 @@
     :param version: [optional] target serialization version
     """
 
-    if version not in _serializers:
-        raise errors.BundleNotSupported(version, 'unknown bundle format')
-
-    serializer = _serializers[version](version)
     source.lock_read()
     try:
-        return serializer.write(source, revision_ids, forced_bases, f)
+        return get_serializer(version).write(source, revision_ids,
+                                             forced_bases, f)
     finally:
         source.unlock()
 
 
 def write_bundle(repository, revision_id, base_revision_id, out, format=None):
-    """"""
-    repository.lock_read()
-    try:
-        return _write_bundle(repository, revision_id, base_revision_id, out,
-                             format)
-    finally:
-        repository.unlock()
-
-
-def _write_bundle(repository, revision_id, base_revision_id, out, format):
     """Write a bundle of revisions.
 
     :param repository: Repository containing revisions to serialize.
@@ -118,14 +116,12 @@
          applying the bundle.
     :param out: Output file.
     """
-    revision_ids = set(repository.get_ancestry(revision_id, topo_sorted=False))
-    revision_ids.difference_update(repository.get_ancestry(base_revision_id,
-                                   topo_sorted=False))
-    revision_ids = list(repository.get_graph().iter_topo_order(revision_ids))
-    revision_ids.reverse()
-    write(repository, revision_ids, out, format,
-          forced_bases = {revision_id:base_revision_id})
-    return revision_ids
+    repository.lock_read()
+    try:
+        return get_serializer(format).write_bundle(repository, revision_id,
+                                                   base_revision_id, out)
+    finally:
+        repository.unlock()
 
 
 class BundleSerializer(object):
@@ -144,9 +140,21 @@
         """
         raise NotImplementedError
 
+    def write_bundle(self, repository, target, base, fileobj):
+        """Write the bundle to the supplied file.
+
+        :param repository: The repository to retrieve revision data from
+        :param target: The revision to provide data for
+        :param base: The most recent of ancestor of the revision that does not
+            need to be included in the bundle
+        :param fileobj: The file to output to
+        """
+        raise NotImplementedError
+
     def write(self, source, revision_ids, forced_bases, f):
         """Write the bundle to the supplied file.
 
+        DEPRECATED: see write_bundle
         :param source: A source for revision information
         :param revision_ids: The list of revision ids to serialize
         :param forced_bases: A dict of revision -> base that overrides default
@@ -154,6 +162,21 @@
         """
         raise NotImplementedError
 
+    def _write_bundle(self, repository, revision_id, base_revision_id, out):
+        """Helper function for translating write_bundle to write"""
+        forced_bases = {revision_id:base_revision_id}
+        if base_revision_id is NULL_REVISION:
+            base_revision_id = None
+        revision_ids = set(repository.get_ancestry(revision_id,
+                           topo_sorted=False))
+        revision_ids.difference_update(repository.get_ancestry(
+            base_revision_id, topo_sorted=False))
+        revision_ids = list(repository.get_graph().iter_topo_order(
+            revision_ids))
+        revision_ids.reverse()
+        self.write(repository, revision_ids, forced_bases, out)
+        return revision_ids
+
 
 def register(version, klass, overwrite=False):
     """Register a BundleSerializer version.
@@ -195,5 +218,7 @@
 
 register_lazy('0.8', 'bzrlib.bundle.serializer.v08', 'BundleSerializerV08')
 register_lazy('0.9', 'bzrlib.bundle.serializer.v09', 'BundleSerializerV09')
-register_lazy(None, 'bzrlib.bundle.serializer.v09', 'BundleSerializerV09')
+register_lazy(v4_string, 'bzrlib.bundle.serializer.v4',
+              'BundleSerializerV4')
+register_lazy(None, 'bzrlib.bundle.serializer.v4', 'BundleSerializerV4')
 

=== modified file 'bzrlib/bundle/serializer/v08.py'
--- bzrlib/bundle/serializer/v08.py	2007-04-25 16:56:25 +0000
+++ bzrlib/bundle/serializer/v08.py	2007-06-20 00:25:11 +0000
@@ -21,7 +21,7 @@
 
 from bzrlib import errors
 from bzrlib.bundle.serializer import (BundleSerializer,
-                                      BUNDLE_HEADER,
+                                      _get_bundle_header,
                                      )
 from bzrlib.bundle.serializer import binary_diff
 from bzrlib.bundle.bundle_data import (RevisionInfo, BundleInfo, BundleTree)
@@ -128,11 +128,13 @@
         finally:
             source.unlock()
 
+    def write_bundle(self, repository, target, base, fileobj):
+        return self._write_bundle(repository, target, base, fileobj)
+
     def _write_main_header(self):
         """Write the header for the changes"""
         f = self.to_file
-        f.write(BUNDLE_HEADER)
-        f.write('0.8\n')
+        f.write(_get_bundle_header('0.8'))
         f.write('#\n')
 
     def _write(self, key, value, indent=1, trailing_space_when_empty=False):

=== modified file 'bzrlib/bundle/serializer/v09.py'
--- bzrlib/bundle/serializer/v09.py	2006-10-11 23:08:27 +0000
+++ bzrlib/bundle/serializer/v09.py	2007-06-13 03:54:07 +0000
@@ -14,7 +14,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-from bzrlib.bundle.serializer import BUNDLE_HEADER
+from bzrlib.bundle.serializer import _get_bundle_header
 from bzrlib.bundle.serializer.v08 import BundleSerializerV08, BundleReader
 from bzrlib.testament import StrictTestament3
 from bzrlib.bundle.bundle_data import BundleInfo
@@ -38,7 +38,7 @@
     def _write_main_header(self):
         """Write the header for the changes"""
         f = self.to_file
-        f.write(BUNDLE_HEADER + '0.9\n#\n')
+        f.write(_get_bundle_header('0.9') + '#\n')
 
     def _testament_sha1(self, revision_id):
         return StrictTestament3.from_revision(self.source, 

=== added file 'bzrlib/bundle/serializer/v4.py'
--- bzrlib/bundle/serializer/v4.py	1970-01-01 00:00:00 +0000
+++ bzrlib/bundle/serializer/v4.py	2007-07-05 00:17:17 +0000
@@ -0,0 +1,553 @@
+# Copyright (C) 2007 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+from cStringIO import StringIO
+import bz2
+import re
+
+from bzrlib import (
+    diff,
+    errors,
+    iterablefile,
+    multiparent,
+    osutils,
+    pack,
+    revision as _mod_revision,
+    trace,
+    xml_serializer,
+    )
+from bzrlib.bundle import bundle_data, serializer
+from bzrlib.util import bencode
+
+
+class BundleWriter(object):
+    """Writer for bundle-format files.
+
+    This serves roughly the same purpose as ContainerReader, but acts as a
+    layer on top of it.
+
+    Provides ways of writing the specific record types supported this bundle
+    format.
+    """
+
+    def __init__(self, fileobj):
+        self._container = pack.ContainerWriter(self._write_encoded)
+        self._fileobj = fileobj
+        self._compressor = bz2.BZ2Compressor()
+
+    def _write_encoded(self, bytes):
+        """Write bzip2-encoded bytes to the file"""
+        self._fileobj.write(self._compressor.compress(bytes))
+
+    def begin(self):
+        """Start writing the bundle"""
+        self._fileobj.write(serializer._get_bundle_header(
+            serializer.v4_string))
+        self._fileobj.write('#\n')
+        self._container.begin()
+
+    def end(self):
+        """Finish writing the bundle"""
+        self._container.end()
+        self._fileobj.write(self._compressor.flush())
+
+    def add_multiparent_record(self, mp_bytes, sha1, parents, repo_kind,
+                               revision_id, file_id):
+        """Add a record for a multi-parent diff
+
+        :mp_bytes: A multi-parent diff, as a bytestring
+        :sha1: The sha1 hash of the fulltext
+        :parents: a list of revision-ids of the parents
+        :repo_kind: The kind of object in the repository.  May be 'file' or
+            'inventory'
+        :revision_id: The revision id of the mpdiff being added.
+        :file_id: The file-id of the file, or None for inventories.
+        """
+        metadata = {'parents': parents,
+                    'storage_kind': 'mpdiff',
+                    'sha1': sha1}
+        self._add_record(mp_bytes, metadata, repo_kind, revision_id, file_id)
+
+    def add_fulltext_record(self, bytes, parents, repo_kind, revision_id):
+        """Add a record for a fulltext
+
+        :bytes: The fulltext, as a bytestring
+        :parents: a list of revision-ids of the parents
+        :repo_kind: The kind of object in the repository.  May be 'revision' or
+            'signature'
+        :revision_id: The revision id of the fulltext being added.
+        """
+        metadata = {'parents': parents,
+                    'storage_kind': 'mpdiff'}
+        self._add_record(bytes, {'parents': parents,
+            'storage_kind': 'fulltext'}, repo_kind, revision_id, None)
+
+    def add_info_record(self, **kwargs):
+        """Add an info record to the bundle
+
+        Any parameters may be supplied, except 'self' and 'storage_kind'.
+        Values must be lists, strings, integers, dicts, or a combination.
+        """
+        kwargs['storage_kind'] = 'header'
+        self._add_record(None, kwargs, 'info', None, None)
+
+    @staticmethod
+    def encode_name(content_kind, revision_id, file_id=None):
+        """Encode semantic ids as a container name"""
+        assert content_kind in ('revision', 'file', 'inventory', 'signature',
+                                'info')
+
+        if content_kind == 'file':
+            assert file_id is not None
+        else:
+            assert file_id is None
+        if content_kind == 'info':
+            assert revision_id is None
+        else:
+            assert revision_id is not None
+        names = [n.replace('/', '//') for n in
+                 (content_kind, revision_id, file_id) if n is not None]
+        return '/'.join(names)
+
+    def _add_record(self, bytes, metadata, repo_kind, revision_id, file_id):
+        """Add a bundle record to the container.
+
+        Most bundle records are recorded as header/body pairs, with the
+        body being nameless.  Records with storage_kind 'header' have no
+        body.
+        """
+        name = self.encode_name(repo_kind, revision_id, file_id)
+        encoded_metadata = bencode.bencode(metadata)
+        self._container.add_bytes_record(encoded_metadata, [name])
+        if metadata['storage_kind'] != 'header':
+            self._container.add_bytes_record(bytes, [])
+
+
+class BundleReader(object):
+    """Reader for bundle-format files.
+
+    This serves roughly the same purpose as ContainerReader, but acts as a
+    layer on top of it, providing metadata, a semantic name, and a record
+    body
+    """
+
+    def __init__(self, fileobj):
+        line = fileobj.readline()
+        if line != '\n':
+            fileobj.readline()
+        self.patch_lines = []
+        self._container = pack.ContainerReader(
+            iterablefile.IterableFile(self.iter_decode(fileobj)))
+
+    @staticmethod
+    def iter_decode(fileobj):
+        """Iterate through decoded fragments of the file"""
+        decompressor = bz2.BZ2Decompressor()
+        for line in fileobj:
+            yield decompressor.decompress(line)
+
+    @staticmethod
+    def decode_name(name):
+        """Decode a name from its container form into a semantic form
+
+        :retval: content_kind, revision_id, file_id
+        """
+        segments = re.split('(//?)', name)
+        names = ['']
+        for segment in segments:
+            if segment == '//':
+                names[-1] += '/'
+            elif segment == '/':
+                names.append('')
+            else:
+                names[-1] += segment
+        content_kind = names[0]
+        revision_id = None
+        file_id = None
+        if len(names) > 1:
+            revision_id = names[1]
+        if len(names) > 2:
+            file_id = names[2]
+        return content_kind, revision_id, file_id
+
+    def iter_records(self):
+        """Iterate through bundle records
+
+        :return: a generator of (bytes, metadata, content_kind, revision_id,
+            file_id)
+        """
+        iterator = self._container.iter_records()
+        for names, meta_bytes in iterator:
+            if len(names) != 1:
+                raise errors.BadBundle('Record has %d names instead of 1'
+                                       % len(names))
+            metadata = bencode.bdecode(meta_bytes(None))
+            if metadata['storage_kind'] == 'header':
+                bytes = None
+            else:
+                _unused, bytes = iterator.next()
+                bytes = bytes(None)
+            yield (bytes, metadata) + self.decode_name(names[0])
+
+
+class BundleSerializerV4(serializer.BundleSerializer):
+    """Implement the high-level bundle interface"""
+
+    def write(self, repository, revision_ids, forced_bases, fileobj):
+        """Write a bundle to a file-like object
+
+        For backwards-compatibility only
+        """
+        write_op = BundleWriteOperation.from_old_args(repository, revision_ids,
+                                                      forced_bases, fileobj)
+        return write_op.do_write()
+
+    def write_bundle(self, repository, target, base, fileobj):
+        """Write a bundle to a file object
+
+        :param repository: The repository to retrieve revision data from
+        :param target: The head revision to include ancestors of
+        :param base: The ancestor of the target to stop including acestors
+            at.
+        :param fileobj: The file-like object to write to
+        """
+        write_op =  BundleWriteOperation(base, target, repository, fileobj)
+        return write_op.do_write()
+
+    def read(self, file):
+        """return a reader object for a given file"""
+        bundle = BundleInfoV4(file, self)
+        return bundle
+
+    @staticmethod
+    def get_source_serializer(info):
+        """Retrieve the serializer for a given info object"""
+        return xml_serializer.format_registry.get(info['serializer'])
+
+
+class BundleWriteOperation(object):
+    """Perform the operation of writing revisions to a bundle"""
+
+    @classmethod
+    def from_old_args(cls, repository, revision_ids, forced_bases, fileobj):
+        """Create a BundleWriteOperation from old-style arguments"""
+        base, target = cls.get_base_target(revision_ids, forced_bases,
+                                           repository)
+        return BundleWriteOperation(base, target, repository, fileobj,
+                                    revision_ids)
+
+    def __init__(self, base, target, repository, fileobj, revision_ids=None):
+        self.base = base
+        self.target = target
+        self.repository = repository
+        bundle = BundleWriter(fileobj)
+        self.bundle = bundle
+        self.base_ancestry = set(repository.get_ancestry(base,
+                                                         topo_sorted=False))
+        if revision_ids is not None:
+            self.revision_ids = revision_ids
+        else:
+            revision_ids = set(repository.get_ancestry(target,
+                                                       topo_sorted=False))
+            self.revision_ids = revision_ids.difference(self.base_ancestry)
+
+    def do_write(self):
+        """Write all data to the bundle"""
+        self.bundle.begin()
+        self.write_info()
+        self.write_files()
+        self.write_revisions()
+        self.bundle.end()
+        return self.revision_ids
+
+    def write_info(self):
+        """Write format info"""
+        serializer_format = self.repository.get_serializer_format()
+        supports_rich_root = {True: 1, False: 0}[
+            self.repository.supports_rich_root()]
+        self.bundle.add_info_record(serializer=serializer_format,
+                                    supports_rich_root=supports_rich_root)
+
+    def iter_file_revisions(self):
+        """Iterate through all relevant revisions of all files.
+
+        This is the correct implementation, but is not compatible with bzr.dev,
+        because certain old revisions were not converted correctly, and have
+        the wrong "revision" marker in inventories.
+        """
+        transaction = self.repository.get_transaction()
+        altered = self.repository.fileids_altered_by_revision_ids(
+            self.revision_ids)
+        for file_id, file_revision_ids in altered.iteritems():
+            vf = self.repository.weave_store.get_weave(file_id, transaction)
+            yield vf, file_id, file_revision_ids
+
+    def iter_file_revisions_aggressive(self):
+        """Iterate through all relevant revisions of all files.
+
+        This uses the standard iter_file_revisions to determine what revisions
+        are referred to by inventories, but then uses the versionedfile to
+        determine what the build-dependencies of each required revision.
+
+        All build dependencies which are not ancestors of the base revision
+        are emitted.
+        """
+        for vf, file_id, file_revision_ids in self.iter_file_revisions():
+            new_revision_ids = set()
+            pending = list(file_revision_ids)
+            while len(pending) > 0:
+                revision_id = pending.pop()
+                if revision_id in new_revision_ids:
+                    continue
+                if revision_id in self.base_ancestry:
+                    continue
+                new_revision_ids.add(revision_id)
+                pending.extend(vf.get_parents(revision_id))
+            yield vf, file_id, new_revision_ids
+
+    def write_files(self):
+        """Write bundle records for all revisions of all files"""
+        for vf, file_id, revision_ids in self.iter_file_revisions_aggressive():
+            self.add_mp_records('file', file_id, vf, revision_ids)
+
+    def write_revisions(self):
+        """Write bundle records for all revisions and signatures"""
+        inv_vf = self.repository.get_inventory_weave()
+        revision_order = list(multiparent.topo_iter(inv_vf, self.revision_ids))
+        if self.target is not None and self.target in self.revision_ids:
+            revision_order.remove(self.target)
+            revision_order.append(self.target)
+        self.add_mp_records('inventory', None, inv_vf, revision_order)
+        parents_list = self.repository.get_parents(revision_order)
+        for parents, revision_id in zip(parents_list, revision_order):
+            revision_text = self.repository.get_revision_xml(revision_id)
+            self.bundle.add_fulltext_record(revision_text, parents,
+                                       'revision', revision_id)
+            try:
+                self.bundle.add_fulltext_record(
+                    self.repository.get_signature_text(
+                    revision_id), parents, 'signature', revision_id)
+            except errors.NoSuchRevision:
+                pass
+
+    @staticmethod
+    def get_base_target(revision_ids, forced_bases, repository):
+        """Determine the base and target from old-style revision ids"""
+        if len(revision_ids) == 0:
+            return None, None
+        target = revision_ids[0]
+        base = forced_bases.get(target)
+        if base is None:
+            parents = repository.get_revision(target).parent_ids
+            if len(parents) == 0:
+                base = _mod_revision.NULL_REVISION
+            else:
+                base = parents[0]
+        return base, target
+
+    def add_mp_records(self, repo_kind, file_id, vf, revision_ids):
+        """Add multi-parent diff records to a bundle"""
+        revision_ids = list(multiparent.topo_iter(vf, revision_ids))
+        mpdiffs = vf.make_mpdiffs(revision_ids)
+        sha1s = vf.get_sha1s(revision_ids)
+        for mpdiff, revision_id, sha1, in zip(mpdiffs, revision_ids, sha1s):
+            parents = vf.get_parents(revision_id)
+            text = ''.join(mpdiff.to_patch())
+            self.bundle.add_multiparent_record(text, sha1, parents, repo_kind,
+                                               revision_id, file_id)
+
+
+class BundleInfoV4(object):
+
+    """Provide (most of) the BundleInfo interface"""
+    def __init__(self, fileobj, serializer):
+        self._fileobj = fileobj
+        self._serializer = serializer
+        self.__real_revisions = None
+        self.__revisions = None
+
+    def install(self, repository):
+        return self.install_revisions(repository)
+
+    def install_revisions(self, repository):
+        """Install this bundle's revisions into the specified repository"""
+        repository.lock_write()
+        try:
+            ri = RevisionInstaller(self.get_bundle_reader(),
+                                   self._serializer, repository)
+            return ri.install()
+        finally:
+            repository.unlock()
+
+    def get_merge_request(self, target_repo):
+        """Provide data for performing a merge
+
+        Returns suggested base, suggested target, and patch verification status
+        """
+        return None, self.target, 'inapplicable'
+
+    def get_bundle_reader(self):
+        self._fileobj.seek(0)
+        return BundleReader(self._fileobj)
+
+    def _get_real_revisions(self):
+        if self.__real_revisions is None:
+            self.__real_revisions = []
+            bundle_reader = self.get_bundle_reader()
+            for bytes, metadata, repo_kind, revision_id, file_id in \
+                bundle_reader.iter_records():
+                if repo_kind == 'info':
+                    serializer =\
+                        self._serializer.get_source_serializer(metadata)
+                if repo_kind == 'revision':
+                    rev = serializer.read_revision_from_string(bytes)
+                    self.__real_revisions.append(rev)
+        return self.__real_revisions
+    real_revisions = property(_get_real_revisions)
+
+    def _get_revisions(self):
+        if self.__revisions is None:
+            self.__revisions = []
+            for revision in self.real_revisions:
+                self.__revisions.append(
+                    bundle_data.RevisionInfo.from_revision(revision))
+        return self.__revisions
+
+    revisions = property(_get_revisions)
+
+    def _get_target(self):
+        return self.revisions[-1].revision_id
+
+    target = property(_get_target)
+
+
+class RevisionInstaller(object):
+    """Installs revisions into a repository"""
+
+    def __init__(self, container, serializer, repository):
+        self._container = container
+        self._serializer = serializer
+        self._repository = repository
+        self._info = None
+
+    def install(self):
+        """Perform the installation"""
+        current_file = None
+        current_versionedfile = None
+        pending_file_records = []
+        added_inv = set()
+        target_revision = None
+        for bytes, metadata, repo_kind, revision_id, file_id in\
+            self._container.iter_records():
+            if repo_kind == 'info':
+                assert self._info is None
+                self._handle_info(metadata)
+            if repo_kind != 'file':
+                self._install_mp_records(current_versionedfile,
+                    pending_file_records)
+                current_file = None
+                current_versionedfile = None
+                pending_file_records = []
+                if repo_kind == 'inventory':
+                    self._install_inventory(revision_id, metadata, bytes)
+                if repo_kind == 'revision':
+                    target_revision = revision_id
+                    self._install_revision(revision_id, metadata, bytes)
+                if repo_kind == 'signature':
+                    self._install_signature(revision_id, metadata, bytes)
+            if repo_kind == 'file':
+                if file_id != current_file:
+                    self._install_mp_records(current_versionedfile,
+                        pending_file_records)
+                    current_file = file_id
+                    current_versionedfile = \
+                        self._repository.weave_store.get_weave_or_empty(
+                        file_id, self._repository.get_transaction())
+                    pending_file_records = []
+                if revision_id in current_versionedfile:
+                    continue
+                pending_file_records.append((revision_id, metadata, bytes))
+        self._install_mp_records(current_versionedfile, pending_file_records)
+        return target_revision
+
+    def _handle_info(self, info):
+        """Extract data from an info record"""
+        self._info = info
+        self._source_serializer = self._serializer.get_source_serializer(info)
+        if (info['supports_rich_root'] == 0 and
+            self._repository.supports_rich_root()):
+            self.update_root = True
+        else:
+            self.update_root = False
+
+    def _install_mp_records(self, versionedfile, records):
+        if len(records) == 0:
+            return
+        d_func = multiparent.MultiParent.from_patch
+        vf_records = [(r, m['parents'], m['sha1'], d_func(t)) for r, m, t in
+                      records if r not in versionedfile]
+        versionedfile.add_mpdiffs(vf_records)
+
+    def _install_inventory(self, revision_id, metadata, text):
+        vf = self._repository.get_inventory_weave()
+        if revision_id in vf:
+            return
+        parent_ids = metadata['parents']
+        if self._info['serializer'] == self._repository._serializer.format_num:
+            return self._install_mp_records(vf, [(revision_id, metadata,
+                                                  text)])
+        parents = [self._repository.get_inventory(p)
+                   for p in parent_ids]
+        parent_texts = [self._source_serializer.write_inventory_to_string(p)
+                        for p in parents]
+        target_lines = multiparent.MultiParent.from_patch(text).to_lines(
+            parent_texts)
+        sha1 = osutils.sha_strings(target_lines)
+        if sha1 != metadata['sha1']:
+            raise errors.BadBundle("Can't convert to target format")
+        target_inv = self._source_serializer.read_inventory_from_string(
+            ''.join(target_lines))
+        self._handle_root(target_inv, parent_ids)
+        try:
+            self._repository.add_inventory(revision_id, target_inv, parent_ids)
+        except errors.UnsupportedInventoryKind:
+            raise errors.IncompatibleRevision(repr(self._repository))
+
+    def _handle_root(self, target_inv, parent_ids):
+        revision_id = target_inv.revision_id
+        if self.update_root:
+            target_inv.root.revision = revision_id
+            store = self._repository.weave_store
+            transaction = self._repository.get_transaction()
+            vf = store.get_weave_or_empty(target_inv.root.file_id, transaction)
+            vf.add_lines(revision_id, parent_ids, [])
+        elif not self._repository.supports_rich_root():
+            if target_inv.root.revision != revision_id:
+                raise errors.IncompatibleRevision(repr(self._repository))
+
+
+    def _install_revision(self, revision_id, metadata, text):
+        if self._repository.has_revision(revision_id):
+            return
+        self._repository._add_revision_text(revision_id, text)
+
+    def _install_signature(self, revision_id, metadata, text):
+        transaction = self._repository.get_transaction()
+        if self._repository._revision_store.has_signature(revision_id,
+                                                          transaction):
+            return
+        self._repository._revision_store.add_revision_signature_text(
+            revision_id, text, transaction)

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-07-13 16:35:54 +0000
+++ bzrlib/errors.py	2007-07-17 13:27:14 +0000
@@ -1263,6 +1263,11 @@
         self.file_id = file_id
 
 
+class VersionedFileInvalidChecksum(VersionedFileError):
+
+    _fmt = "Text did not match its checksum: %(message)s"
+
+
 class KnitError(BzrError):
     
     _fmt = "Knit error"
@@ -2114,6 +2119,22 @@
         " branch location."
 
 
+class IllegalMergeDirectivePayload(BzrError):
+    """A merge directive contained something other than a patch or bundle"""
+
+    _fmt = "Bad merge directive payload %(start)r"
+
+    def __init__(self, start):
+        BzrError(self)
+        self.start = start
+
+
+class PatchVerificationFailed(BzrError):
+    """A patch from a merge directive could not be verified"""
+
+    _fmt = "Preview patch does not match requested changes."
+
+
 class PatchMissing(BzrError):
     """Raise a patch type was specified but no patch supplied"""
 

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2007-04-20 04:18:37 +0000
+++ bzrlib/fetch.py	2007-06-22 22:19:13 +0000
@@ -92,7 +92,7 @@
         if to_repository.control_files._transport.base == from_repository.control_files._transport.base:
             # check that last_revision is in 'from' and then return a no-operation.
             if last_revision not in (None, NULL_REVISION):
-                from_repository.get_revision(last_revision)
+                to_repository.get_revision(last_revision)
             return
         self.to_repository = to_repository
         self.from_repository = from_repository

=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2007-06-21 06:32:06 +0000
+++ bzrlib/graph.py	2007-06-27 15:12:32 +0000
@@ -276,6 +276,8 @@
             lca = self.find_lca(*revisions)
             if len(lca) == 1:
                 return lca.pop()
+            if len(lca) == 0:
+                raise errors.NoCommonAncestor(left_revision, right_revision)
             revisions = lca
 
     def iter_topo_order(self, revisions):

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2007-06-29 00:10:13 +0000
+++ bzrlib/knit.py	2007-07-17 13:27:14 +0000
@@ -153,6 +153,32 @@
     def copy(self):
         return KnitContent(self._lines[:])
 
+    @staticmethod
+    def get_line_delta_blocks(knit_delta, source, target):
+        """Extract SequenceMatcher.get_matching_blocks() from a knit delta"""
+        target_len = len(target)
+        s_pos = 0
+        t_pos = 0
+        for s_begin, s_end, t_len, new_text in knit_delta:
+            true_n = s_begin - s_pos
+            n = true_n
+            if n > 0:
+                # knit deltas do not provide reliable info about whether the
+                # last line of a file matches, due to eol handling.
+                if source[s_pos + n -1] != target[t_pos + n -1]:
+                    n-=1
+                if n > 0:
+                    yield s_pos, t_pos, n
+            t_pos += t_len + true_n
+            s_pos = s_end
+        n = target_len - t_pos
+        if n > 0:
+            if source[s_pos + n -1] != target[t_pos + n -1]:
+                n-=1
+            if n > 0:
+                yield s_pos, t_pos, n
+        yield s_pos + (target_len - t_pos), target_len, 0
+
 
 class _KnitFactory(object):
     """Base factory for creating content objects."""
@@ -519,6 +545,12 @@
                                 current_values[3],
                                 new_parents)
 
+    def _extract_blocks(self, version_id, source, target):
+        if self._index.get_method(version_id) != 'line-delta':
+            return None
+        parent, sha1, noeol, delta = self.get_delta(version_id)
+        return KnitContent.get_line_delta_blocks(delta, source, target)
+
     def get_delta(self, version_id):
         """Get a delta for constructing version from some other version."""
         version_id = osutils.safe_revision_id(version_id)
@@ -554,11 +586,14 @@
         return dict(graph_items)
 
     def get_sha1(self, version_id):
+        return self.get_sha1s([version_id])[0]
+
+    def get_sha1s(self, version_ids):
         """See VersionedFile.get_sha1()."""
-        version_id = osutils.safe_revision_id(version_id)
-        record_map = self._get_record_map([version_id])
-        method, content, digest, next = record_map[version_id]
-        return digest 
+        version_ids = [osutils.safe_revision_id(v) for v in version_ids]
+        record_map = self._get_record_map(version_ids)
+        # record entry 2 is the 'digest'.
+        return [record_map[v][2] for v in version_ids]
 
     @staticmethod
     def get_suffixes():
@@ -812,6 +847,8 @@
         text_map, content_map = self._get_content_maps(version_ids)
         return [text_map[v] for v in version_ids]
 
+    _get_lf_split_line_list = get_line_list
+
     def _get_content_maps(self, version_ids):
         """Produce maps of text and KnitContents
         

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2007-07-12 23:56:07 +0000
+++ bzrlib/merge.py	2007-07-17 13:27:14 +0000
@@ -224,6 +224,23 @@
         self.other_tree = self.revision_tree(revision_id)
         self.other_basis = revision_id
 
+    def set_base_revision(self, revision_id, branch):
+        """Set 'base' based on a branch and revision id
+
+        :param revision_id: The revision to use for a tree
+        :param branch: The branch containing this tree
+        """
+        self.base_rev_id = revision_id
+        self.base_branch = branch
+        self._maybe_fetch(branch, self.this_branch, revision_id)
+        self.base_tree = self.revision_tree(revision_id)
+        self.base_is_ancestor = is_ancestor(self.this_basis,
+                                            self.base_rev_id,
+                                            self.this_branch)
+        self.base_is_other_ancestor = is_ancestor(self.other_basis,
+                                                  self.base_rev_id,
+                                                  self.this_branch)
+
     def _maybe_fetch(self, source, target, revision_id):
         if (source.repository.bzrdir.root_transport.base !=
             target.repository.bzrdir.root_transport.base):

=== modified file 'bzrlib/merge_directive.py'
--- bzrlib/merge_directive.py	2007-06-19 21:22:56 +0000
+++ bzrlib/merge_directive.py	2007-07-03 22:44:19 +0000
@@ -17,12 +17,14 @@
 
 from email import Message
 from StringIO import StringIO
+import re
 
 from bzrlib import (
     branch as _mod_branch,
     diff,
     errors,
     gpg,
+    registry,
     revision as _mod_revision,
     rio,
     testament,
@@ -33,27 +35,11 @@
     )
 
 
-class MergeDirective(object):
-
-    """A request to perform a merge into a branch.
-
-    Designed to be serialized and mailed.  It provides all the information
-    needed to perform a merge automatically, by providing at minimum a revision
-    bundle or the location of a branch.
-
-    The serialization format is robust against certain common forms of
-    deterioration caused by mailing.
-
-    The format is also designed to be patch-compatible.  If the directive
-    includes a diff or revision bundle, it should be possible to apply it
-    directly using the standard patch program.
-    """
-
-    _format_string = 'Bazaar merge directive format 1'
+class _BaseMergeDirective(object):
 
     def __init__(self, revision_id, testament_sha1, time, timezone,
-                 target_branch, patch=None, patch_type=None,
-                 source_branch=None, message=None):
+                 target_branch, patch=None, source_branch=None, message=None,
+                 bundle=None):
         """Constructor.
 
         :param revision_id: The revision to merge
@@ -63,69 +49,19 @@
         :param timezone: The timezone offset
         :param target_branch: The branch to apply the merge to
         :param patch: The text of a diff or bundle
-        :param patch_type: None, "diff" or "bundle", depending on the contents
-            of patch
         :param source_branch: A public location to merge the revision from
         :param message: The message to use when committing this merge
         """
-        assert patch_type in (None, 'diff', 'bundle')
-        if patch_type != 'bundle' and source_branch is None:
-            raise errors.NoMergeSource()
-        if patch_type is not None and patch is None:
-            raise errors.PatchMissing(patch_type)
         self.revision_id = revision_id
         self.testament_sha1 = testament_sha1
         self.time = time
         self.timezone = timezone
         self.target_branch = target_branch
         self.patch = patch
-        self.patch_type = patch_type
         self.source_branch = source_branch
         self.message = message
 
-    @classmethod
-    def from_lines(klass, lines):
-        """Deserialize a MergeRequest from an iterable of lines
-
-        :param lines: An iterable of lines
-        :return: a MergeRequest
-        """
-        line_iter = iter(lines)
-        for line in line_iter:
-            if line.startswith('# ' + klass._format_string):
-                break
-        else:
-            if len(lines) > 0:
-                raise errors.NotAMergeDirective(lines[0])
-            else:
-                raise errors.NotAMergeDirective('')
-        stanza = rio.read_patch_stanza(line_iter)
-        patch_lines = list(line_iter)
-        if len(patch_lines) == 0:
-            patch = None
-            patch_type = None
-        else:
-            patch = ''.join(patch_lines)
-            try:
-                bundle_serializer.read_bundle(StringIO(patch))
-            except (errors.NotABundle, errors.BundleNotSupported,
-                    errors.BadBundle):
-                patch_type = 'diff'
-            else:
-                patch_type = 'bundle'
-        time, timezone = timestamp.parse_patch_date(stanza.get('timestamp'))
-        kwargs = {}
-        for key in ('revision_id', 'testament_sha1', 'target_branch',
-                    'source_branch', 'message'):
-            try:
-                kwargs[key] = stanza.get(key)
-            except KeyError:
-                pass
-        kwargs['revision_id'] = kwargs['revision_id'].encode('utf-8')
-        return MergeDirective(time=time, timezone=timezone,
-                              patch_type=patch_type, patch=patch, **kwargs)
-
-    def to_lines(self):
+    def _to_lines(self, base_revision=False):
         """Serialize as a list of lines
 
         :return: a list of lines
@@ -137,13 +73,84 @@
         for key in ('source_branch', 'message'):
             if self.__dict__[key] is not None:
                 stanza.add(key, self.__dict__[key])
+        if base_revision:
+            stanza.add('base_revision_id', self.base_revision_id)
         lines = ['# ' + self._format_string + '\n']
         lines.extend(rio.to_patch_lines(stanza))
         lines.append('# \n')
-        if self.patch is not None:
-            lines.extend(self.patch.splitlines(True))
         return lines
 
+    @classmethod
+    def from_objects(klass, repository, revision_id, time, timezone,
+                 target_branch, patch_type='bundle',
+                 local_target_branch=None, public_branch=None, message=None):
+        """Generate a merge directive from various objects
+
+        :param repository: The repository containing the revision
+        :param revision_id: The revision to merge
+        :param time: The POSIX timestamp of the date the request was issued.
+        :param timezone: The timezone of the request
+        :param target_branch: The url of the branch to merge into
+        :param patch_type: 'bundle', 'diff' or None, depending on the type of
+            patch desired.
+        :param local_target_branch: a local copy of the target branch
+        :param public_branch: location of a public branch containing the target
+            revision.
+        :param message: Message to use when committing the merge
+        :return: The merge directive
+
+        The public branch is always used if supplied.  If the patch_type is
+        not 'bundle', the public branch must be supplied, and will be verified.
+
+        If the message is not supplied, the message from revision_id will be
+        used for the commit.
+        """
+        t_revision_id = revision_id
+        if revision_id == _mod_revision.NULL_REVISION:
+            t_revision_id = None
+        t = testament.StrictTestament3.from_revision(repository, t_revision_id)
+        submit_branch = _mod_branch.Branch.open(target_branch)
+        if submit_branch.get_public_branch() is not None:
+            target_branch = submit_branch.get_public_branch()
+        if patch_type is None:
+            patch = None
+        else:
+            submit_revision_id = submit_branch.last_revision()
+            submit_revision_id = _mod_revision.ensure_null(submit_revision_id)
+            repository.fetch(submit_branch.repository, submit_revision_id)
+            graph = repository.get_graph()
+            ancestor_id = graph.find_unique_lca(revision_id,
+                                                submit_revision_id)
+            type_handler = {'bundle': klass._generate_bundle,
+                            'diff': klass._generate_diff,
+                            None: lambda x, y, z: None }
+            patch = type_handler[patch_type](repository, revision_id,
+                                             ancestor_id)
+
+        if public_branch is not None and patch_type != 'bundle':
+            public_branch_obj = _mod_branch.Branch.open(public_branch)
+            if not public_branch_obj.repository.has_revision(revision_id):
+                raise errors.PublicBranchOutOfDate(public_branch,
+                                                   revision_id)
+
+        return klass(revision_id, t.as_sha1(), time, timezone, target_branch,
+            patch, patch_type, public_branch, message)
+
+    @staticmethod
+    def _generate_diff(repository, revision_id, ancestor_id):
+        tree_1 = repository.revision_tree(ancestor_id)
+        tree_2 = repository.revision_tree(revision_id)
+        s = StringIO()
+        diff.show_diff_trees(tree_1, tree_2, s, old_label='', new_label='')
+        return s.getvalue()
+
+    @staticmethod
+    def _generate_bundle(repository, revision_id, ancestor_id):
+        s = StringIO()
+        bundle_serializer.write_bundle(repository, revision_id,
+                                       ancestor_id, s)
+        return s.getvalue()
+
     def to_signed(self, branch):
         """Serialize as a signed string.
 
@@ -178,91 +185,12 @@
         message.set_payload(body)
         return message
 
-    @classmethod
-    def from_objects(klass, repository, revision_id, time, timezone,
-                 target_branch, patch_type='bundle',
-                 local_target_branch=None, public_branch=None, message=None):
-        """Generate a merge directive from various objects
-
-        :param repository: The repository containing the revision
-        :param revision_id: The revision to merge
-        :param time: The POSIX timestamp of the date the request was issued.
-        :param timezone: The timezone of the request
-        :param target_branch: The url of the branch to merge into
-        :param patch_type: 'bundle', 'diff' or None, depending on the type of
-            patch desired.
-        :param local_target_branch: a local copy of the target branch
-        :param public_branch: location of a public branch containing the target
-            revision.
-        :param message: Message to use when committing the merge
-        :return: The merge directive
-
-        The public branch is always used if supplied.  If the patch_type is
-        not 'bundle', the public branch must be supplied, and will be verified.
-
-        If the message is not supplied, the message from revision_id will be
-        used for the commit.
-        """
-        t_revision_id = revision_id
-        if revision_id == 'null:':
-            t_revision_id = None
-        t = testament.StrictTestament3.from_revision(repository, t_revision_id)
-        submit_branch = _mod_branch.Branch.open(target_branch)
-        if submit_branch.get_public_branch() is not None:
-            target_branch = submit_branch.get_public_branch()
-        if patch_type is None:
-            patch = None
-        else:
-            submit_revision_id = submit_branch.last_revision()
-            submit_revision_id = _mod_revision.ensure_null(submit_revision_id)
-            repository.fetch(submit_branch.repository, submit_revision_id)
-            graph = repository.get_graph()
-            ancestor_id = graph.find_unique_lca(revision_id,
-                                                submit_revision_id)
-            type_handler = {'bundle': klass._generate_bundle,
-                            'diff': klass._generate_diff,
-                            None: lambda x, y, z: None }
-            patch = type_handler[patch_type](repository, revision_id,
-                                             ancestor_id)
-            if patch_type == 'bundle':
-                s = StringIO()
-                bundle_serializer.write_bundle(repository, revision_id,
-                                               ancestor_id, s)
-                patch = s.getvalue()
-            elif patch_type == 'diff':
-                patch = klass._generate_diff(repository, revision_id,
-                                             ancestor_id)
-
-            if public_branch is not None and patch_type != 'bundle':
-                public_branch_obj = _mod_branch.Branch.open(public_branch)
-                if not public_branch_obj.repository.has_revision(revision_id):
-                    raise errors.PublicBranchOutOfDate(public_branch,
-                                                       revision_id)
-
-        return MergeDirective(revision_id, t.as_sha1(), time, timezone,
-                              target_branch, patch, patch_type, public_branch,
-                              message)
-
-    @staticmethod
-    def _generate_diff(repository, revision_id, ancestor_id):
-        tree_1 = repository.revision_tree(ancestor_id)
-        tree_2 = repository.revision_tree(revision_id)
-        s = StringIO()
-        diff.show_diff_trees(tree_1, tree_2, s, old_label='', new_label='')
-        return s.getvalue()
-
-    @staticmethod
-    def _generate_bundle(repository, revision_id, ancestor_id):
-        s = StringIO()
-        bundle_serializer.write_bundle(repository, revision_id,
-                                       ancestor_id, s)
-        return s.getvalue()
-
     def install_revisions(self, target_repo):
         """Install revisions and return the target revision"""
         if not target_repo.has_revision(self.revision_id):
             if self.patch_type == 'bundle':
-                info = bundle_serializer.read_bundle(StringIO(self.patch))
+                info = bundle_serializer.read_bundle(
+                    StringIO(self.get_raw_bundle()))
                 # We don't use the bundle's target revision, because
                 # MergeDirective.revision_id is authoritative.
                 info.install_revisions(target_repo)
@@ -270,3 +198,326 @@
                 source_branch = _mod_branch.Branch.open(self.source_branch)
                 target_repo.fetch(source_branch.repository, self.revision_id)
         return self.revision_id
+
+
+class MergeDirective(_BaseMergeDirective):
+
+    """A request to perform a merge into a branch.
+
+    Designed to be serialized and mailed.  It provides all the information
+    needed to perform a merge automatically, by providing at minimum a revision
+    bundle or the location of a branch.
+
+    The serialization format is robust against certain common forms of
+    deterioration caused by mailing.
+
+    The format is also designed to be patch-compatible.  If the directive
+    includes a diff or revision bundle, it should be possible to apply it
+    directly using the standard patch program.
+    """
+
+    _format_string = 'Bazaar merge directive format 1'
+
+    def __init__(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, patch_type=None,
+                 source_branch=None, message=None, bundle=None):
+        """Constructor.
+
+        :param revision_id: The revision to merge
+        :param testament_sha1: The sha1 of the testament of the revision to
+            merge.
+        :param time: The current POSIX timestamp time
+        :param timezone: The timezone offset
+        :param target_branch: The branch to apply the merge to
+        :param patch: The text of a diff or bundle
+        :param patch_type: None, "diff" or "bundle", depending on the contents
+            of patch
+        :param source_branch: A public location to merge the revision from
+        :param message: The message to use when committing this merge
+        """
+        _BaseMergeDirective.__init__(self, revision_id, testament_sha1, time,
+            timezone, target_branch, patch, source_branch, message)
+        assert patch_type in (None, 'diff', 'bundle'), patch_type
+        if patch_type != 'bundle' and source_branch is None:
+            raise errors.NoMergeSource()
+        if patch_type is not None and patch is None:
+            raise errors.PatchMissing(patch_type)
+        self.patch_type = patch_type
+
+    def clear_payload(self):
+        self.patch = None
+        self.patch_type = None
+
+    def get_raw_bundle(self):
+        return self.bundle
+
+    def _bundle(self):
+        if self.patch_type == 'bundle':
+            return self.patch
+        else:
+            return None
+
+    bundle = property(_bundle)
+
+    @classmethod
+    def from_lines(klass, lines):
+        """Deserialize a MergeRequest from an iterable of lines
+
+        :param lines: An iterable of lines
+        :return: a MergeRequest
+        """
+        line_iter = iter(lines)
+        for line in line_iter:
+            if line.startswith('# Bazaar merge directive format '):
+                break
+        else:
+            if len(lines) > 0:
+                raise errors.NotAMergeDirective(lines[0])
+            else:
+                raise errors.NotAMergeDirective('')
+        return _format_registry.get(line[2:].rstrip())._from_lines(line_iter)
+
+    @classmethod
+    def _from_lines(klass, line_iter):
+        stanza = rio.read_patch_stanza(line_iter)
+        patch_lines = list(line_iter)
+        if len(patch_lines) == 0:
+            patch = None
+            patch_type = None
+        else:
+            patch = ''.join(patch_lines)
+            try:
+                bundle_serializer.read_bundle(StringIO(patch))
+            except (errors.NotABundle, errors.BundleNotSupported,
+                    errors.BadBundle):
+                patch_type = 'diff'
+            else:
+                patch_type = 'bundle'
+        time, timezone = timestamp.parse_patch_date(stanza.get('timestamp'))
+        kwargs = {}
+        for key in ('revision_id', 'testament_sha1', 'target_branch',
+                    'source_branch', 'message'):
+            try:
+                kwargs[key] = stanza.get(key)
+            except KeyError:
+                pass
+        kwargs['revision_id'] = kwargs['revision_id'].encode('utf-8')
+        return MergeDirective(time=time, timezone=timezone,
+                              patch_type=patch_type, patch=patch, **kwargs)
+
+    def to_lines(self):
+        lines = self._to_lines()
+        if self.patch is not None:
+            lines.extend(self.patch.splitlines(True))
+        return lines
+
+    @staticmethod
+    def _generate_bundle(repository, revision_id, ancestor_id):
+        s = StringIO()
+        bundle_serializer.write_bundle(repository, revision_id,
+                                       ancestor_id, s, '0.9')
+        return s.getvalue()
+
+    def get_merge_request(self, repository):
+        """Provide data for performing a merge
+
+        Returns suggested base, suggested target, and patch verification status
+        """
+        return None, self.revision_id, 'inapplicable'
+
+
+class MergeDirective2(_BaseMergeDirective):
+
+    _format_string = 'Bazaar merge directive format 2 (Bazaar 0.18)'
+
+    def __init__(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, source_branch=None, message=None,
+                 bundle=None, base_revision_id=None):
+        if source_branch is None and bundle is None:
+            raise errors.NoMergeSource()
+        _BaseMergeDirective.__init__(self, revision_id, testament_sha1, time,
+            timezone, target_branch, patch, source_branch, message)
+        self.bundle = bundle
+        self.base_revision_id = base_revision_id
+
+    def _patch_type(self):
+        if self.bundle is not None:
+            return 'bundle'
+        elif self.patch is not None:
+            return 'diff'
+        else:
+            return None
+
+    patch_type = property(_patch_type)
+
+    def clear_payload(self):
+        self.patch = None
+        self.bundle = None
+
+    def get_raw_bundle(self):
+        if self.bundle is None:
+            return None
+        else:
+            return self.bundle.decode('base-64')
+
+    @classmethod
+    def _from_lines(klass, line_iter):
+        stanza = rio.read_patch_stanza(line_iter)
+        patch = None
+        bundle = None
+        try:
+            start = line_iter.next()
+        except StopIteration:
+            pass
+        else:
+            if start.startswith('# Begin patch'):
+                patch_lines = []
+                for line in line_iter:
+                    if line.startswith('# Begin bundle'):
+                        start = line
+                        break
+                    patch_lines.append(line)
+                else:
+                    start = None
+                patch = ''.join(patch_lines)
+            if start is not None:
+                if start.startswith('# Begin bundle'):
+                    bundle = ''.join(line_iter)
+                else:
+                    raise errors.IllegalMergeDirectivePayload(start)
+        time, timezone = timestamp.parse_patch_date(stanza.get('timestamp'))
+        kwargs = {}
+        for key in ('revision_id', 'testament_sha1', 'target_branch',
+                    'source_branch', 'message', 'base_revision_id'):
+            try:
+                kwargs[key] = stanza.get(key)
+            except KeyError:
+                pass
+        kwargs['revision_id'] = kwargs['revision_id'].encode('utf-8')
+        kwargs['base_revision_id'] =\
+            kwargs['base_revision_id'].encode('utf-8')
+        return klass(time=time, timezone=timezone, patch=patch, bundle=bundle,
+                     **kwargs)
+
+    def to_lines(self):
+        lines = self._to_lines(base_revision=True)
+        if self.patch is not None:
+            lines.append('# Begin patch\n')
+            lines.extend(self.patch.splitlines(True))
+        if self.bundle is not None:
+            lines.append('# Begin bundle\n')
+            lines.extend(self.bundle.splitlines(True))
+        return lines
+
+    @classmethod
+    def from_objects(klass, repository, revision_id, time, timezone,
+                 target_branch, include_patch=True, include_bundle=True,
+                 local_target_branch=None, public_branch=None, message=None,
+                 base_revision_id=None):
+        """Generate a merge directive from various objects
+
+        :param repository: The repository containing the revision
+        :param revision_id: The revision to merge
+        :param time: The POSIX timestamp of the date the request was issued.
+        :param timezone: The timezone of the request
+        :param target_branch: The url of the branch to merge into
+        :param include_patch: If true, include a preview patch
+        :param include_bundle: If true, include a bundle
+        :param local_target_branch: a local copy of the target branch
+        :param public_branch: location of a public branch containing the target
+            revision.
+        :param message: Message to use when committing the merge
+        :return: The merge directive
+
+        The public branch is always used if supplied.  If no bundle is
+        included, the public branch must be supplied, and will be verified.
+
+        If the message is not supplied, the message from revision_id will be
+        used for the commit.
+        """
+        locked = []
+        try:
+            repository.lock_write()
+            locked.append(repository)
+            t_revision_id = revision_id
+            if revision_id == 'null:':
+                t_revision_id = None
+            t = testament.StrictTestament3.from_revision(repository,
+                t_revision_id)
+            submit_branch = _mod_branch.Branch.open(target_branch)
+            submit_branch.lock_read()
+            locked.append(submit_branch)
+            if submit_branch.get_public_branch() is not None:
+                target_branch = submit_branch.get_public_branch()
+            submit_revision_id = submit_branch.last_revision()
+            submit_revision_id = _mod_revision.ensure_null(submit_revision_id)
+            graph = repository.get_graph(submit_branch.repository)
+            ancestor_id = graph.find_unique_lca(revision_id,
+                                                submit_revision_id)
+            if base_revision_id is None:
+                base_revision_id = ancestor_id
+            if (include_patch, include_bundle) != (False, False):
+                repository.fetch(submit_branch.repository, submit_revision_id)
+            if include_patch:
+                patch = klass._generate_diff(repository, revision_id,
+                                             base_revision_id)
+            else:
+                patch = None
+
+            if include_bundle:
+                bundle = klass._generate_bundle(repository, revision_id,
+                    ancestor_id).encode('base-64')
+            else:
+                bundle = None
+
+            if public_branch is not None and not include_bundle:
+                public_branch_obj = _mod_branch.Branch.open(public_branch)
+                public_branch_obj.lock_read()
+                locked.append(public_branch_obj)
+                if not public_branch_obj.repository.has_revision(
+                    revision_id):
+                    raise errors.PublicBranchOutOfDate(public_branch,
+                                                       revision_id)
+        finally:
+            for entry in reversed(locked):
+                entry.unlock()
+        return klass(revision_id, t.as_sha1(), time, timezone, target_branch,
+            patch, public_branch, message, bundle, base_revision_id)
+
+    def _verify_patch(self, repository):
+        calculated_patch = self._generate_diff(repository, self.revision_id,
+                                               self.base_revision_id)
+        # Convert line-endings to UNIX
+        stored_patch = re.sub('\r\n?', '\n', self.patch)
+        # Strip trailing whitespace
+        calculated_patch = re.sub(' *\n', '\n', calculated_patch)
+        stored_patch = re.sub(' *\n', '\n', stored_patch)
+        return (calculated_patch == stored_patch)
+
+    def get_merge_request(self, repository):
+        """Provide data for performing a merge
+
+        Returns suggested base, suggested target, and patch verification status
+        """
+        verified = self._maybe_verify(repository)
+        return self.base_revision_id, self.revision_id, verified
+
+    def _maybe_verify(self, repository):
+        if self.patch is not None:
+            if self._verify_patch(repository):
+                return 'verified'
+            else:
+                return 'failed'
+        else:
+            return 'inapplicable'
+
+
+class MergeDirectiveFormatRegistry(registry.Registry):
+
+    def register(self, directive):
+        registry.Registry.register(self, directive._format_string, directive)
+
+
+_format_registry = MergeDirectiveFormatRegistry()
+_format_registry.register(MergeDirective)
+_format_registry.register(MergeDirective2)

=== added file 'bzrlib/multiparent.py'
--- bzrlib/multiparent.py	1970-01-01 00:00:00 +0000
+++ bzrlib/multiparent.py	2007-07-04 03:38:28 +0000
@@ -0,0 +1,636 @@
+# Copyright (C) 2007 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+from bzrlib.lazy_import import lazy_import
+
+lazy_import(globals(), """
+import errno
+import itertools
+import os
+from StringIO import StringIO
+
+from bzrlib import (
+    patiencediff,
+    trace,
+    ui,
+    )
+from bzrlib.util import bencode
+""")
+from bzrlib.tuned_gzip import GzipFile
+
+
+def topo_iter(vf, versions=None):
+    seen = set()
+    descendants = {}
+    if versions is None:
+        versions = vf.versions()
+    def pending_parents(version):
+        return [v for v in vf.get_parents(version) if v in versions and
+                v not in seen]
+    for version_id in versions:
+        for parent_id in vf.get_parents(version_id):
+            descendants.setdefault(parent_id, []).append(version_id)
+    cur = [v for v in versions if len(pending_parents(v)) == 0]
+    while len(cur) > 0:
+        next = []
+        for version_id in cur:
+            if version_id in seen:
+                continue
+            if len(pending_parents(version_id)) != 0:
+                continue
+            next.extend(descendants.get(version_id, []))
+            yield version_id
+            seen.add(version_id)
+        cur = next
+    assert len(seen) == len(versions)
+
+
+class MultiParent(object):
+    """A multi-parent diff"""
+
+    def __init__(self, hunks=None):
+        if hunks is not None:
+            self.hunks = hunks
+        else:
+            self.hunks = []
+
+    def __repr__(self):
+        return "MultiParent(%r)" % self.hunks
+
+    def __eq__(self, other):
+        if self.__class__ is not other.__class__:
+            return False
+        return (self.hunks == other.hunks)
+
+    @staticmethod
+    def from_lines(text, parents=(), left_blocks=None):
+        """Produce a MultiParent from a list of lines and parents"""
+        def compare(parent):
+            matcher = patiencediff.PatienceSequenceMatcher(None, parent,
+                                                           text)
+            return matcher.get_matching_blocks()
+        if len(parents) > 0:
+            if left_blocks is None:
+                left_blocks = compare(parents[0])
+            parent_comparisons = [left_blocks] + [compare(p) for p in
+                                                  parents[1:]]
+        else:
+            parent_comparisons = []
+        cur_line = 0
+        new_text = NewText([])
+        parent_text = []
+        block_iter = [iter(i) for i in parent_comparisons]
+        diff = MultiParent([])
+        def next_block(p):
+            try:
+                return block_iter[p].next()
+            except StopIteration:
+                return None
+        cur_block = [next_block(p) for p, i in enumerate(block_iter)]
+        while cur_line < len(text):
+            best_match = None
+            for p, block in enumerate(cur_block):
+                if block is None:
+                    continue
+                i, j, n = block
+                while j + n < cur_line:
+                    block = cur_block[p] = next_block(p)
+                    if block is None:
+                        break
+                    i, j, n = block
+                if block is None:
+                    continue
+                if j > cur_line:
+                    continue
+                offset = cur_line - j
+                i += offset
+                j = cur_line
+                n -= offset
+                if n == 0:
+                    continue
+                if best_match is None or n > best_match.num_lines:
+                    best_match = ParentText(p, i, j, n)
+            if best_match is None:
+                new_text.lines.append(text[cur_line])
+                cur_line += 1
+            else:
+                if len(new_text.lines) > 0:
+                    diff.hunks.append(new_text)
+                    new_text = NewText([])
+                diff.hunks.append(best_match)
+                cur_line += best_match.num_lines
+        if len(new_text.lines) > 0:
+            diff.hunks.append(new_text)
+        return diff
+
+    def to_lines(self, parents=()):
+        """Contruct a fulltext from this diff and its parents"""
+        mpvf = MultiMemoryVersionedFile()
+        for num, parent in enumerate(parents):
+            mpvf.add_version(StringIO(parent).readlines(), num, [])
+        mpvf.add_diff(self, 'a', range(len(parents)))
+        return mpvf.get_line_list(['a'])[0]
+
+    @classmethod
+    def from_texts(cls, text, parents=()):
+        """Produce a MultiParent from a text and list of parent text"""
+        return cls.from_lines(StringIO(text).readlines(),
+                              [StringIO(p).readlines() for p in parents])
+
+    def to_patch(self):
+        """Yield text lines for a patch"""
+        for hunk in self.hunks:
+            for line in hunk.to_patch():
+                yield line
+
+    def patch_len(self):
+        return len(''.join(self.to_patch()))
+
+    def zipped_patch_len(self):
+        return len(gzip_string(self.to_patch()))
+
+    @classmethod
+    def from_patch(cls, text):
+        """Create a MultiParent from its string form"""
+        return cls._from_patch(StringIO(text))
+
+    @staticmethod
+    def _from_patch(lines):
+        """This is private because it is essential to split lines on \n only"""
+        line_iter = iter(lines)
+        hunks = []
+        cur_line = None
+        while(True):
+            try:
+                cur_line = line_iter.next()
+            except StopIteration:
+                break
+            if cur_line[0] == 'i':
+                num_lines = int(cur_line.split(' ')[1])
+                hunk_lines = [line_iter.next() for x in xrange(num_lines)]
+                hunk_lines[-1] = hunk_lines[-1][:-1]
+                hunks.append(NewText(hunk_lines))
+            elif cur_line[0] == '\n':
+                hunks[-1].lines[-1] += '\n'
+            else:
+                assert cur_line[0] == 'c', cur_line[0]
+                parent, parent_pos, child_pos, num_lines =\
+                    [int(v) for v in cur_line.split(' ')[1:]]
+                hunks.append(ParentText(parent, parent_pos, child_pos,
+                                        num_lines))
+        return MultiParent(hunks)
+
+    def range_iterator(self):
+        """Iterate through the hunks, with range indicated
+
+        kind is "new" or "parent".
+        for "new", data is a list of lines.
+        for "parent", data is (parent, parent_start, parent_end)
+        :return: a generator of (start, end, kind, data)
+        """
+        start = 0
+        for hunk in self.hunks:
+            if isinstance(hunk, NewText):
+                kind = 'new'
+                end = start + len(hunk.lines)
+                data = hunk.lines
+            else:
+                kind = 'parent'
+                start = hunk.child_pos
+                end = start + hunk.num_lines
+                data = (hunk.parent, hunk.parent_pos, hunk.parent_pos +
+                        hunk.num_lines)
+            yield start, end, kind, data
+            start = end
+
+    def num_lines(self):
+        """The number of lines in the output text"""
+        extra_n = 0
+        for hunk in reversed(self.hunks):
+            if isinstance(hunk, ParentText):
+               return hunk.child_pos + hunk.num_lines + extra_n
+            extra_n += len(hunk.lines)
+        return extra_n
+
+    def is_snapshot(self):
+        """Return true of this hunk is effectively a fulltext"""
+        if len(self.hunks) != 1:
+            return False
+        return (isinstance(self.hunks[0], NewText))
+
+
+class NewText(object):
+    """The contents of text that is introduced by this text"""
+
+    def __init__(self, lines):
+        self.lines = lines
+
+    def __eq__(self, other):
+        if self.__class__ is not other.__class__:
+            return False
+        return (other.lines == self.lines)
+
+    def __repr__(self):
+        return 'NewText(%r)' % self.lines
+
+    def to_patch(self):
+        yield 'i %d\n' % len(self.lines)
+        for line in self.lines:
+            yield line
+        yield '\n'
+
+
+class ParentText(object):
+    """A reference to text present in a parent text"""
+
+    def __init__(self, parent, parent_pos, child_pos, num_lines):
+        self.parent = parent
+        self.parent_pos = parent_pos
+        self.child_pos = child_pos
+        self.num_lines = num_lines
+
+    def __repr__(self):
+        return 'ParentText(%(parent)r, %(parent_pos)r, %(child_pos)r,'\
+            ' %(num_lines)r)' % self.__dict__
+
+    def __eq__(self, other):
+        if self.__class__ != other.__class__:
+            return False
+        return (self.__dict__ == other.__dict__)
+
+    def to_patch(self):
+        yield 'c %(parent)d %(parent_pos)d %(child_pos)d %(num_lines)d\n'\
+            % self.__dict__
+
+
+class BaseVersionedFile(object):
+    """Pseudo-VersionedFile skeleton for MultiParent"""
+
+    def __init__(self, snapshot_interval=25, max_snapshots=None):
+        self._lines = {}
+        self._parents = {}
+        self._snapshots = set()
+        self.snapshot_interval = snapshot_interval
+        self.max_snapshots = max_snapshots
+
+    def versions(self):
+        return iter(self._parents)
+
+    def has_version(self, version):
+        return version in self._parents
+
+    def do_snapshot(self, version_id, parent_ids):
+        """Determine whether to perform a a snapshot for this version"""
+        if self.snapshot_interval is None:
+            return False
+        if self.max_snapshots is not None and\
+            len(self._snapshots) == self.max_snapshots:
+            return False
+        if len(parent_ids) == 0:
+            return True
+        for ignored in xrange(self.snapshot_interval):
+            if len(parent_ids) == 0:
+                return False
+            version_ids = parent_ids
+            parent_ids = []
+            for version_id in version_ids:
+                if version_id not in self._snapshots:
+                    parent_ids.extend(self._parents[version_id])
+        else:
+            return True
+
+    def add_version(self, lines, version_id, parent_ids,
+                    force_snapshot=None, single_parent=False):
+        """Add a version to the versionedfile
+
+        :param lines: The list of lines to add.  Must be split on '\n'.
+        :param version_id: The version_id of the version to add
+        :param force_snapshot: If true, force this version to be added as a
+            snapshot version.  If false, force this version to be added as a
+            diff.  If none, determine this automatically.
+        :param single_parent: If true, use a single parent, rather than
+            multiple parents.
+        """
+        if force_snapshot is None:
+            do_snapshot = self.do_snapshot(version_id, parent_ids)
+        else:
+            do_snapshot = force_snapshot
+        if do_snapshot:
+            self._snapshots.add(version_id)
+            diff = MultiParent([NewText(lines)])
+        else:
+            if single_parent:
+                parent_lines = self.get_line_list(parent_ids[:1])
+            else:
+                parent_lines = self.get_line_list(parent_ids)
+            diff = MultiParent.from_lines(lines, parent_lines)
+            if diff.is_snapshot():
+                self._snapshots.add(version_id)
+        self.add_diff(diff, version_id, parent_ids)
+        self._lines[version_id] = lines
+
+    def get_parents(self, version_id):
+        return self._parents[version_id]
+
+    def make_snapshot(self, version_id):
+        snapdiff = MultiParent([NewText(self.cache_version(version_id))])
+        self.add_diff(snapdiff, version_id, self._parents[version_id])
+        self._snapshots.add(version_id)
+
+    def import_versionedfile(self, vf, snapshots, no_cache=True,
+                             single_parent=False, verify=False):
+        """Import all revisions of a versionedfile
+
+        :param vf: The versionedfile to import
+        :param snapshots: If provided, the revisions to make snapshots of.
+            Otherwise, this will be auto-determined
+        :param no_cache: If true, clear the cache after every add.
+        :param single_parent: If true, omit all but one parent text, (but
+            retain parent metadata).
+        """
+        assert no_cache or not verify
+        revisions = set(vf.versions())
+        total = len(revisions)
+        pb = ui.ui_factory.nested_progress_bar()
+        try:
+            while len(revisions) > 0:
+                added = set()
+                for revision in revisions:
+                    parents = vf.get_parents(revision)
+                    if [p for p in parents if p not in self._parents] != []:
+                        continue
+                    lines = [a + ' ' + l for a, l in
+                             vf.annotate_iter(revision)]
+                    if snapshots is None:
+                        force_snapshot = None
+                    else:
+                        force_snapshot = (revision in snapshots)
+                    self.add_version(lines, revision, parents, force_snapshot,
+                                     single_parent)
+                    added.add(revision)
+                    if no_cache:
+                        self.clear_cache()
+                        vf.clear_cache()
+                        if verify:
+                            assert lines == self.get_line_list([revision])[0]
+                            self.clear_cache()
+                    pb.update('Importing revisions',
+                              (total - len(revisions)) + len(added), total)
+                revisions = [r for r in revisions if r not in added]
+        finally:
+            pb.finished()
+
+    def select_snapshots(self, vf):
+        """Determine which versions to add as snapshots"""
+        build_ancestors = {}
+        descendants = {}
+        snapshots = set()
+        for version_id in topo_iter(vf):
+            potential_build_ancestors = set(vf.get_parents(version_id))
+            parents = vf.get_parents(version_id)
+            if len(parents) == 0:
+                snapshots.add(version_id)
+                build_ancestors[version_id] = set()
+            else:
+                for parent in vf.get_parents(version_id):
+                    potential_build_ancestors.update(build_ancestors[parent])
+                if len(potential_build_ancestors) > self.snapshot_interval:
+                    snapshots.add(version_id)
+                    build_ancestors[version_id] = set()
+                else:
+                    build_ancestors[version_id] = potential_build_ancestors
+        return snapshots
+
+    def select_by_size(self, num):
+        """Select snapshots for minimum output size"""
+        num -= len(self._snapshots)
+        new_snapshots = self.get_size_ranking()[-num:]
+        return [v for n, v in new_snapshots]
+
+    def get_size_ranking(self):
+        """Get versions ranked by size"""
+        versions = []
+        new_snapshots = set()
+        for version_id in self.versions():
+            if version_id in self._snapshots:
+                continue
+            diff_len = self.get_diff(version_id).patch_len()
+            snapshot_len = MultiParent([NewText(
+                self.cache_version(version_id))]).patch_len()
+            versions.append((snapshot_len - diff_len, version_id))
+        versions.sort()
+        return versions
+
+    def import_diffs(self, vf):
+        """Import the diffs from another pseudo-versionedfile"""
+        for version_id in vf.versions():
+            self.add_diff(vf.get_diff(version_id), version_id,
+                          vf._parents[version_id])
+
+    def get_build_ranking(self):
+        """Return revisions sorted by how much they reduce build complexity"""
+        could_avoid = {}
+        referenced_by = {}
+        for version_id in topo_iter(self):
+            could_avoid[version_id] = set()
+            if version_id not in self._snapshots:
+                for parent_id in self._parents[version_id]:
+                    could_avoid[version_id].update(could_avoid[parent_id])
+                could_avoid[version_id].update(self._parents)
+                could_avoid[version_id].discard(version_id)
+            for avoid_id in could_avoid[version_id]:
+                referenced_by.setdefault(avoid_id, set()).add(version_id)
+        available_versions = list(self.versions())
+        ranking = []
+        while len(available_versions) > 0:
+            available_versions.sort(key=lambda x:
+                len(could_avoid[x]) *
+                len(referenced_by.get(x, [])))
+            selected = available_versions.pop()
+            ranking.append(selected)
+            for version_id in referenced_by[selected]:
+                could_avoid[version_id].difference_update(
+                    could_avoid[selected])
+            for version_id in could_avoid[selected]:
+                referenced_by[version_id].difference_update(
+                    referenced_by[selected]
+                )
+        return ranking
+
+    def clear_cache(self):
+        self._lines.clear()
+
+    def get_line_list(self, version_ids):
+        return [self.cache_version(v) for v in version_ids]
+
+    def cache_version(self, version_id):
+        try:
+            return self._lines[version_id]
+        except KeyError:
+            pass
+        diff = self.get_diff(version_id)
+        lines = []
+        reconstructor = _Reconstructor(self, self._lines,
+                                       self._parents)
+        reconstructor.reconstruct_version(lines, version_id)
+        self._lines[version_id] = lines
+        return lines
+
+
+class MultiMemoryVersionedFile(BaseVersionedFile):
+    """Memory-backed pseudo-versionedfile"""
+
+    def __init__(self, snapshot_interval=25, max_snapshots=None):
+        BaseVersionedFile.__init__(self, snapshot_interval, max_snapshots)
+        self._diffs = {}
+
+    def add_diff(self, diff, version_id, parent_ids):
+        self._diffs[version_id] = diff
+        self._parents[version_id] = parent_ids
+
+    def get_diff(self, version_id):
+        return self._diffs[version_id]
+
+    def destroy(self):
+        self._diffs = {}
+
+
+class MultiVersionedFile(BaseVersionedFile):
+    """Disk-backed pseudo-versionedfile"""
+
+    def __init__(self, filename, snapshot_interval=25, max_snapshots=None):
+        BaseVersionedFile.__init__(self, snapshot_interval, max_snapshots)
+        self._filename = filename
+        self._diff_offset = {}
+
+    def get_diff(self, version_id):
+        start, count = self._diff_offset[version_id]
+        infile = open(self._filename + '.mpknit', 'rb')
+        try:
+            infile.seek(start)
+            sio = StringIO(infile.read(count))
+        finally:
+            infile.close()
+        zip_file = GzipFile(None, mode='rb', fileobj=sio)
+        try:
+            file_version_id = zip_file.readline()
+            return MultiParent.from_patch(zip_file.read())
+        finally:
+            zip_file.close()
+
+    def add_diff(self, diff, version_id, parent_ids):
+        outfile = open(self._filename + '.mpknit', 'ab')
+        try:
+            start = outfile.tell()
+            try:
+                zipfile = GzipFile(None, mode='ab', fileobj=outfile)
+                zipfile.writelines(itertools.chain(
+                    ['version %s\n' % version_id], diff.to_patch()))
+            finally:
+                zipfile.close()
+            end = outfile.tell()
+        finally:
+            outfile.close()
+        self._diff_offset[version_id] = (start, end-start)
+        self._parents[version_id] = parent_ids
+
+    def destroy(self):
+        try:
+            os.unlink(self._filename + '.mpknit')
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+        try:
+            os.unlink(self._filename + '.mpidx')
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+
+    def save(self):
+        open(self._filename + '.mpidx', 'wb').write(bencode.bencode(
+            (self._parents, list(self._snapshots), self._diff_offset)))
+
+    def load(self):
+        self._parents, snapshots, self._diff_offset = bencode.bdecode(
+            open(self._filename + '.mpidx', 'rb').read())
+        self._snapshots = set(snapshots)
+
+
+class _Reconstructor(object):
+    """Build a text from the diffs, ancestry graph and cached lines"""
+
+    def __init__(self, diffs, lines, parents):
+        self.diffs = diffs
+        self.lines = lines
+        self.parents = parents
+        self.cursor = {}
+
+    def reconstruct(self, lines, parent_text, version_id):
+        """Append the lines referred to by a ParentText to lines"""
+        parent_id = self.parents[version_id][parent_text.parent]
+        end = parent_text.parent_pos + parent_text.num_lines
+        return self._reconstruct(lines, parent_id, parent_text.parent_pos,
+                                 end)
+
+    def _reconstruct(self, lines, req_version_id, req_start, req_end):
+        """Append lines for the requested version_id range"""
+        # stack of pending range requests
+        if req_start == req_end:
+            return
+        pending_reqs = [(req_version_id, req_start, req_end)]
+        while len(pending_reqs) > 0:
+            req_version_id, req_start, req_end = pending_reqs.pop()
+            # lazily allocate cursors for versions
+            try:
+                start, end, kind, data, iterator = self.cursor[req_version_id]
+            except KeyError:
+                iterator = self.diffs.get_diff(req_version_id).range_iterator()
+                start, end, kind, data = iterator.next()
+            if start > req_start:
+                iterator = self.diffs.get_diff(req_version_id).range_iterator()
+                start, end, kind, data = iterator.next()
+
+            # find the first hunk relevant to the request
+            while end <= req_start:
+                start, end, kind, data = iterator.next()
+            self.cursor[req_version_id] = start, end, kind, data, iterator
+            # if the hunk can't satisfy the whole request, split it in two,
+            # and leave the second half for later.
+            if req_end > end:
+                pending_reqs.append((req_version_id, end, req_end))
+                req_end = end
+            if kind == 'new':
+                lines.extend(data[req_start - start: (req_end - start)])
+            else:
+                # If the hunk is a ParentText, rewrite it as a range request
+                # for the parent, and make it the next pending request.
+                parent, parent_start, parent_end = data
+                new_version_id = self.parents[req_version_id][parent]
+                new_start = parent_start + req_start - start
+                new_end = parent_end + req_end - end
+                pending_reqs.append((new_version_id, new_start, new_end))
+
+    def reconstruct_version(self, lines, version_id):
+        length = self.diffs.get_diff(version_id).num_lines()
+        return self._reconstruct(lines, version_id, 0, length)
+
+
+def gzip_string(lines):
+    sio = StringIO()
+    data_file = GzipFile(None, mode='wb', fileobj=sio)
+    data_file.writelines(lines)
+    data_file.close()
+    return sio.getvalue()

=== added file 'bzrlib/plugins/multiparent.py'
--- bzrlib/plugins/multiparent.py	1970-01-01 00:00:00 +0000
+++ bzrlib/plugins/multiparent.py	2007-07-19 15:03:23 +0000
@@ -0,0 +1,175 @@
+"""Implementation of multiparent diffs for versionedfile-like storage
+
+Provides mp-regen and mp-extract commands.
+Focus is on comparing size/performance to knits.
+"""
+
+from bzrlib.lazy_import import lazy_import
+
+lazy_import(globals(), """
+import (
+        errno,
+        os,
+        sys,
+        time,
+        )
+
+from bzrlib import (
+    commands,
+    urlutils
+    )
+from bzrlib.workingtree import WorkingTree
+from bzrlib.tests import TestUtil
+
+from bzrlib.plugins.multiparent.multiparent import (
+    MultiVersionedFile,
+    MultiMemoryVersionedFile,
+    )
+""")
+
+class cmd_mp_regen(commands.Command):
+    """Generate a multiparent versionedfile"""
+
+    hidden = True
+
+    takes_args = ['file?']
+
+    takes_options = [commands.Option('sync-snapshots',
+                                     help='Snapshots follow source.'),
+                     commands.Option('snapshot-interval', type=int,
+                                     help='Take snapshots every x revisions.'),
+                     commands.Option('outfile', type=unicode,
+                                     help='Write pseudo-knit to this file.'),
+                     commands.Option('memory', help='Use memory, not disk.'),
+                     commands.Option('extract', help='Test extract time.'),
+                     commands.Option('single', help='Use a single parent.'),
+                     commands.Option('verify', help='Verify added texts.'),
+                     commands.Option('cache', help='Aggresively cache.'),
+                     commands.Option('size', help='Aggressive size.'),
+                     commands.Option('build', help='Aggressive build.'),
+                    ]
+    hidden = True
+
+    def run(self, file=None, sync_snapshots=False, snapshot_interval=26,
+            lsprof_timed=False, dump=False, extract=False, single=False,
+            verify=False, outfile=None, memory=False, cache=False,
+            size=False, build=False):
+        file_weave = get_file_weave(file)
+        url = file_weave.transport.abspath(file_weave.filename)
+        print >> sys.stderr, 'Importing: %s' % \
+            urlutils.local_path_from_url(url)
+        if sync_snapshots:
+            print >> sys.stderr, 'Snapshots follow input'
+        else:
+            print >> sys.stderr, 'Snapshot interval: %d' % snapshot_interval
+        if not memory:
+            if outfile is None:
+                filename = 'pknit'
+            else:
+                filename = outfile
+            vf = MultiVersionedFile(filename, snapshot_interval)
+        else:
+            vf = MultiMemoryVersionedFile(snapshot_interval)
+        vf.destroy()
+        old_snapshots = set(r for r in file_weave.versions() if
+                        file_weave._index.get_method(r) == 'fulltext')
+        if sync_snapshots:
+            to_sync = old_snapshots
+        elif size or build:
+            assert memory
+            to_sync = set()
+        else:
+            to_sync = vf.select_snapshots(file_weave)
+        print >> sys.stderr, "%d fulltext(s)" % len(old_snapshots)
+        print >> sys.stderr, "%d planned snapshots" % len(to_sync)
+
+        try:
+            vf.import_versionedfile(file_weave, to_sync, single_parent=single,
+                                    verify=verify, no_cache=not cache)
+            if size:
+                snapshots = vf.select_by_size(len(old_snapshots))
+                for version_id in snapshots:
+                    vf.make_snapshot(version_id)
+            if build:
+                ranking = vf.get_build_ranking()
+                snapshots = ranking[:len(old_snapshots) -\
+                    len(vf._snapshots)]
+                for version_id in snapshots:
+                    vf.make_snapshot(version_id)
+        except:
+            vf.destroy()
+            raise
+        try:
+            print >> sys.stderr, "%d actual snapshots" % len(vf._snapshots)
+            if not cache:
+                vf.clear_cache()
+            if memory:
+                if outfile is not None:
+                    vf_file = MultiVersionedFile(outfile)
+                vf_file.import_diffs(vf)
+            else:
+                vf_file = vf
+        finally:
+            if outfile is None:
+                vf.destroy()
+            else:
+                vf_file.save()
+
+class cmd_mp_extract(commands.Command):
+    """Test extraction time multiparent knits"""
+
+    hidden = True
+
+    takes_options = [
+        commands.Option('lsprof-timed', help='Use lsprof.'),
+        commands.Option('parallel', help='Extract multiple versions at once.'),
+        commands.Option('count', help='Number of cycles to do.', type=int),
+        ]
+
+    takes_args = ['filename', 'vfile?']
+
+    def run(self, filename, vfile=None, lsprof_timed=False, count=1000,
+            parallel=False):
+        vf = MultiVersionedFile(filename)
+        vf.load()
+        snapshots = [r for r in vf.versions() if vf.get_diff(r).is_snapshot()]
+        print '%d snapshots' % len(snapshots)
+        revisions = list(vf.versions())
+        revisions = revisions[-count:]
+        print 'Testing extract time of %d revisions' % len(revisions)
+        if parallel:
+            revisions_list = [revisions]
+        else:
+            revisions_list = [[r] for r in revisions]
+        start = time.clock()
+        for revisions in revisions_list:
+            vf = MultiVersionedFile(filename)
+            vf.load()
+            vf.get_line_list(revisions)
+        print >> sys.stderr, time.clock() - start
+        if lsprof_timed:
+            from bzrlib.lsprof import profile
+            vf.clear_cache()
+            ret, stats = profile(vf.get_line_list, revisions_list[-1][-1])
+            stats.sort()
+            stats.pprint()
+        start = time.clock()
+        for revisions in revisions_list:
+            file_weave = get_file_weave(vfile)
+            file_weave.get_line_list(revisions)
+        print >> sys.stderr, time.clock() - start
+
+
+def get_file_weave(filename=None, wt=None):
+    if filename is None:
+        wt, path = WorkingTree.open_containing('.')
+        return wt.branch.repository.get_inventory_weave()
+    else:
+        wt, path = WorkingTree.open_containing(filename)
+        file_id = wt.path2id(path)
+        bt = wt.branch.repository.revision_tree(wt.last_revision())
+        return bt.get_weave(file_id)
+
+
+commands.register_command(cmd_mp_regen)
+commands.register_command(cmd_mp_extract)

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2007-07-12 12:07:13 +0000
+++ bzrlib/remote.py	2007-07-17 13:27:14 +0000
@@ -470,6 +470,10 @@
         self._ensure_real()
         return self._real_repository.revision_tree(revision_id)
 
+    def get_serializer_format(self):
+        self._ensure_real()
+        return self._real_repository.get_serializer_format()
+
     def get_commit_builder(self, branch, parents, config, timestamp=None,
                            timezone=None, committer=None, revprops=None,
                            revision_id=None):
@@ -527,6 +531,10 @@
         return self._real_repository.fetch(
             source, revision_id=revision_id, pb=pb)
 
+    def create_bundle(self, target, base, fileobj, format=None):
+        self._ensure_real()
+        self._real_repository.create_bundle(target, base, fileobj, format)
+
     @property
     def control_weaves(self):
         self._ensure_real()

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2007-07-12 13:35:54 +0000
+++ bzrlib/repository.py	2007-07-17 13:27:14 +0000
@@ -40,6 +40,7 @@
     transactions,
     ui,
     )
+from bzrlib.bundle import serializer
 from bzrlib.revisiontree import RevisionTree
 from bzrlib.store.versioned import VersionedFileStore
 from bzrlib.store.text import TextStore
@@ -143,6 +144,12 @@
                 self.add_inventory(revision_id, inv, rev.parent_ids)
         self._revision_store.add_revision(rev, self.get_transaction())
 
+    def _add_revision_text(self, revision_id, text):
+        revision = self._revision_store._serializer.read_revision_from_string(
+            text)
+        self._revision_store._add_revision(revision, StringIO(text),
+                                           self.get_transaction())
+
     @needs_read_lock
     def _all_possible_ids(self):
         """Return all the possible revisions that we could find."""
@@ -365,6 +372,9 @@
         except NotImplementedError:
             raise errors.IncompatibleRepositories(source, self)
 
+    def create_bundle(self, target, base, fileobj, format=None):
+        return serializer.write_bundle(self, target, base, fileobj, format)
+
     def get_commit_builder(self, branch, parents, config, timestamp=None, 
                            timezone=None, committer=None, revprops=None, 
                            revision_id=None):
@@ -643,6 +653,9 @@
     def serialise_inventory(self, inv):
         return self._serializer.write_inventory_to_string(inv)
 
+    def get_serializer_format(self):
+        return self._serializer.format_num
+
     @needs_read_lock
     def get_inventory_xml(self, revision_id):
         """Get inventory XML as a file object."""

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2007-07-18 23:19:46 +0000
+++ bzrlib/tests/__init__.py	2007-07-19 13:13:06 +0000
@@ -2304,6 +2304,7 @@
                    'bzrlib.tests.test_merge_directive',
                    'bzrlib.tests.test_missing',
                    'bzrlib.tests.test_msgeditor',
+                   'bzrlib.tests.test_multiparent',
                    'bzrlib.tests.test_nonascii',
                    'bzrlib.tests.test_options',
                    'bzrlib.tests.test_osutils',

=== modified file 'bzrlib/tests/blackbox/__init__.py'
--- bzrlib/tests/blackbox/__init__.py	2007-07-12 12:07:13 +0000
+++ bzrlib/tests/blackbox/__init__.py	2007-07-17 13:27:14 +0000
@@ -50,7 +50,6 @@
                      'bzrlib.tests.blackbox.test_break_lock',
                      'bzrlib.tests.blackbox.test_breakin',
                      'bzrlib.tests.blackbox.test_bound_branches',
-                     'bzrlib.tests.blackbox.test_bundle',
                      'bzrlib.tests.blackbox.test_cat',
                      'bzrlib.tests.blackbox.test_cat_revision',
                      'bzrlib.tests.blackbox.test_checkout',
@@ -98,6 +97,7 @@
                      'bzrlib.tests.blackbox.test_sign_my_commits',
                      'bzrlib.tests.blackbox.test_split',
                      'bzrlib.tests.blackbox.test_status',
+                     'bzrlib.tests.blackbox.test_submit',
                      'bzrlib.tests.blackbox.test_tags',
                      'bzrlib.tests.blackbox.test_testament',
                      'bzrlib.tests.blackbox.test_too_much',

=== modified file 'bzrlib/tests/blackbox/test_merge.py'
--- bzrlib/tests/blackbox/test_merge.py	2007-06-27 19:13:50 +0000
+++ bzrlib/tests/blackbox/test_merge.py	2007-07-17 13:27:14 +0000
@@ -21,6 +21,7 @@
 
 import os
 
+from bzrlib import merge_directive
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import BzrDir
 from bzrlib.conflicts import ConflictList, ContentsConflict
@@ -289,3 +290,44 @@
         self.run_bzr('merge ../tree_a', retcode=1)
         self.assertEqual(tree_b.conflicts(),
                          [ContentsConflict('file', file_id='file-id')])
+
+    def test_directive_cherrypick(self):
+        source = self.make_branch_and_tree('source')
+        self.build_tree(['source/a'])
+        source.add('a')
+        source.commit('Added a', rev_id='rev1')
+        self.build_tree(['source/b'])
+        source.add('b')
+        source.commit('Added b', rev_id='rev2')
+        target = self.make_branch_and_tree('target')
+        target.commit('empty commit')
+        self.write_directive('directive', source.branch, 'target', 'rev2',
+                             'rev1')
+        self.run_bzr('merge -d target directive')
+        self.failIfExists('target/a')
+        self.failUnlessExists('target/b')
+
+    def write_directive(self, filename, source, target, revision_id,
+                        base_revision_id=None, mangle_patch=False):
+        md = merge_directive.MergeDirective2.from_objects(
+            source.repository, revision_id, 0, 0, target,
+            base_revision_id=base_revision_id)
+        if mangle_patch:
+            md.patch = 'asdf\n'
+        self.build_tree_contents([(filename, ''.join(md.to_lines()))])
+
+    def test_directive_verify_warning(self):
+        source = self.make_branch_and_tree('source')
+        self.build_tree(['source/a'])
+        source.add('a')
+        source.commit('Added a', rev_id='rev1')
+        target = self.make_branch_and_tree('target')
+        target.commit('empty commit')
+        self.write_directive('directive', source.branch, 'target', 'rev1')
+        err = self.run_bzr('merge -d target directive')[1]
+        self.assertNotContainsRe(err, 'Preview patch does not match changes')
+        target.revert([])
+        self.write_directive('directive', source.branch, 'target', 'rev1',
+                             mangle_patch=True)
+        err = self.run_bzr('merge -d target directive')[1]
+        self.assertContainsRe(err, 'Preview patch does not match changes')

=== modified file 'bzrlib/tests/blackbox/test_merge_directive.py'
--- bzrlib/tests/blackbox/test_merge_directive.py	2007-07-05 00:12:25 +0000
+++ bzrlib/tests/blackbox/test_merge_directive.py	2007-07-17 13:27:14 +0000
@@ -19,6 +19,7 @@
 
 from bzrlib import (
     gpg,
+    merge_directive,
     tests,
     workingtree,
     )
@@ -28,7 +29,7 @@
 From: J. Random Hacker <jrandom@example.com>
 Subject: bar
 
-# Bazaar merge directive format 1
+# Bazaar merge directive format 2 \\(Bazaar 0.18\\)
 # revision_id: bar-id
 # target_branch: ../tree2
 # testament_sha1: .*
@@ -45,7 +46,7 @@
         self.tree1.branch.get_config().set_user_option('email',
             'J. Random Hacker <jrandom@example.com>')
         self.tree1.add('file')
-        self.tree1.commit('foo')
+        self.tree1.commit('foo', rev_id='foo-id')
         self.tree2 = self.tree1.bzrdir.sprout('tree2').open_workingtree()
         self.build_tree_contents([('tree1/file', 'a\nb\nc\nd\ne\n')])
         self.tree1.commit('bar', rev_id='bar-id')
@@ -58,6 +59,11 @@
         self.assertContainsRe(md_text, "\\+e")
         md_text = self.run_bzr('merge-directive -r -2 ../tree2')[0]
         self.assertNotContainsRe(md_text, "\\+e")
+        md_text = self.run_bzr('merge-directive -r -1..-2 ../tree2')[0]
+        md2 = merge_directive.MergeDirective.from_lines(
+            md_text.splitlines(True))
+        self.assertEqual('foo-id', md2.revision_id)
+        self.assertEqual('bar-id', md2.base_revision_id)
 
     def test_submit_branch(self):
         self.prepare_merge_directive()
@@ -77,10 +83,10 @@
     def test_patch_types(self):
         self.prepare_merge_directive()
         md_text = self.run_bzr('merge-directive ../tree2')[0]
-        self.assertContainsRe(md_text, "Bazaar revision bundle")
+        self.assertContainsRe(md_text, "# Begin bundle")
         self.assertContainsRe(md_text, "\\+e")
         md_text = self.run_bzr('merge-directive ../tree2 --diff .')[0]
-        self.assertNotContainsRe(md_text, "Bazaar revision bundle")
+        self.assertNotContainsRe(md_text, "# Begin bundle")
         self.assertContainsRe(md_text, "\\+e")
         md_text = self.run_bzr('merge-directive --plain')[0]
         self.assertNotContainsRe(md_text, "\\+e")

=== renamed file 'bzrlib/tests/blackbox/test_bundle.py' => 'bzrlib/tests/blackbox/test_submit.py'
--- bzrlib/tests/blackbox/test_bundle.py	2007-06-26 20:32:49 +0000
+++ bzrlib/tests/blackbox/test_submit.py	2007-07-17 13:27:14 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2006 Canonical Ltd
+# Copyright (C) 2006, 2007 Canonical Ltd
 # Authors: Aaron Bentley
 #
 # This program is free software; you can redistribute it and/or modify
@@ -19,12 +19,18 @@
 import os
 from StringIO import StringIO
 
-from bzrlib.bundle.serializer import read_bundle
+from bzrlib import merge_directive
+from bzrlib.bundle import serializer
 from bzrlib.bzrdir import BzrDir
 from bzrlib import tests
 
 
-class TestBundle(tests.TestCaseWithTransport):
+def read_bundle(fileobj):
+    md = merge_directive.MergeDirective.from_lines(fileobj.readlines())
+    return serializer.read_bundle(StringIO(md.get_raw_bundle()))
+
+
+class TestSubmit(tests.TestCaseWithTransport):
 
     def make_trees(self):
         grandparent_tree = BzrDir.create_standalone_workingtree('grandparent')
@@ -39,36 +45,37 @@
         """Parent location is used as a basis by default"""
         self.make_trees()
         os.chdir('grandparent')
-        errmsg = self.run_bzr('bundle', retcode=3)[1]
-        self.assertContainsRe(errmsg, 'No base branch known or specified')
+        errmsg = self.run_bzr('submit', retcode=3)[1]
+        self.assertContainsRe(errmsg, 'No submit branch known or specified')
         os.chdir('../branch')
-        stdout, stderr = self.run_bzr('bundle')
+        stdout, stderr = self.run_bzr('submit')
         self.assertEqual(stderr.count('Using saved location'), 1)
         br = read_bundle(StringIO(stdout))
         self.assertRevisions(br, ['revision3'])
 
     def assertRevisions(self, bi, expected):
-        self.assertEqual([r.revision_id for r in bi.revisions], expected)
+        self.assertEqual(set(r.revision_id for r in bi.revisions),
+            set(expected))
 
     def test_uses_submit(self):
         """Submit location can be used and set"""
         self.make_trees()
         os.chdir('branch')
-        br = read_bundle(StringIO(self.run_bzr('bundle')[0]))
+        br = read_bundle(StringIO(self.run_bzr('submit')[0]))
         self.assertRevisions(br, ['revision3'])
-        br = read_bundle(StringIO(self.run_bzr('bundle ../grandparent')[0]))
+        br = read_bundle(StringIO(self.run_bzr('submit ../grandparent')[0]))
         self.assertRevisions(br, ['revision3', 'revision2'])
         # submit location should be auto-remembered
-        br = read_bundle(StringIO(self.run_bzr('bundle')[0]))
-        self.assertRevisions(br, ['revision3', 'revision2'])
-        self.run_bzr('bundle ../parent')
-        br = read_bundle(StringIO(self.run_bzr('bundle')[0]))
-        self.assertRevisions(br, ['revision3', 'revision2'])
-        self.run_bzr('bundle ../parent --remember')
+        br = read_bundle(StringIO(self.run_bzr('submit')[0]))
+        self.assertRevisions(br, ['revision3', 'revision2'])
+        self.run_bzr('submit ../parent')
+        br = read_bundle(StringIO(self.run_bzr('bundle')[0]))
+        self.assertRevisions(br, ['revision3', 'revision2'])
+        self.run_bzr('submit ../parent --remember')
         br = read_bundle(StringIO(self.run_bzr('bundle')[0]))
         self.assertRevisions(br, ['revision3'])
-        err = self.run_bzr('bundle --remember', retcode=3)[1]
-        self.assertContainsRe(err,
+        err = self.run_bzr('submit --remember', retcode=3)[1]
+        self.assertContainsRe(err, 
                               '--remember requires a branch to be specified.')
 
     def test_revision_branch_interaction(self):
@@ -76,25 +83,70 @@
         os.chdir('branch')
         bi = read_bundle(StringIO(self.run_bzr('bundle ../grandparent')[0]))
         self.assertRevisions(bi, ['revision3', 'revision2'])
-        out = StringIO(self.run_bzr('bundle ../grandparent -r -2')[0])
+        out = StringIO(self.run_bzr('submit ../grandparent -r -2')[0])
         bi = read_bundle(out)
         self.assertRevisions(bi, ['revision2'])
-        bi = read_bundle(StringIO(self.run_bzr('bundle -r -2..-1')[0]))
-        self.assertRevisions(bi, ['revision3'])
-        self.run_bzr('bundle ../grandparent -r -2..-1', retcode=3)
+        sio = StringIO(self.run_bzr('submit -r -2..-1')[0])
+        md = merge_directive.MergeDirective.from_lines(sio.readlines())
+        self.assertEqual('revision2', md.base_revision_id)
+        self.assertEqual('revision3', md.revision_id)
+        sio.seek(0)
+        bi = read_bundle(sio)
+        self.assertRevisions(bi, ['revision2', 'revision3'])
+        self.run_bzr('submit ../grandparent -r -2..-1')
 
     def test_output(self):
         # check output for consistency
         # win32 stdout converts LF to CRLF,
-        # and this is breaks the created bundle
-        self.make_trees()
+        # which would break patch-based bundles
+        self.make_trees()        
         os.chdir('branch')
-        stdout = self.run_bzr_subprocess('bundle')[0]
+        stdout = self.run_bzr_subprocess('submit')[0]
         br = read_bundle(StringIO(stdout))
         self.assertRevisions(br, ['revision3'])
 
     def test_no_common_ancestor(self):
         foo = self.make_branch_and_tree('foo')
+        foo.commit('rev a')
         bar = self.make_branch_and_tree('bar')
+        bar.commit('rev b')
         os.chdir('foo')
-        self.run_bzr('bundle ../bar')
+        self.run_bzr('submit ../bar')
+
+    def submit_directive(self, args):
+        sio = StringIO(self.run_bzr(['submit'] + args)[0])
+        return merge_directive.MergeDirective.from_lines(sio.readlines())
+
+    def test_content_options(self):
+        """--no-patch and --no-bundle should work and be independant"""
+        self.make_trees()
+        os.chdir('branch')
+        md = self.submit_directive([])
+        self.assertIsNot(None, md.bundle)
+        self.assertIsNot(None, md.patch)
+        md = self.submit_directive(['--no-patch'])
+        self.assertIsNot(None, md.bundle)
+        self.assertIs(None, md.patch)
+        md = self.submit_directive(['--no-bundle', '.', '.'])
+        self.assertIs(None, md.bundle)
+        self.assertIsNot(None, md.patch)
+        md = self.submit_directive(['--no-bundle', '--no-patch', '.', '.'])
+        self.assertIs(None, md.bundle)
+        self.assertIs(None, md.patch)
+
+    def test_from_option(self):
+        self.make_trees()
+        self.run_bzr('submit', retcode=3)
+        md = self.submit_directive(['--from', 'branch'])
+        self.assertEqual('revision3', md.revision_id)
+        md = self.submit_directive(['-f', 'branch'])
+        self.assertEqual('revision3', md.revision_id)
+
+    def test_output_option(self):
+        self.make_trees()
+        stdout = self.run_bzr('submit -f branch --output file1')[0]
+        self.assertEqual('', stdout)
+        md_file = open('file1', 'rb')
+        self.addCleanup(md_file.close)
+        self.assertContainsRe(md_file.read(), 'revision3')
+

=== modified file 'bzrlib/tests/repository_implementations/test_repository.py'
--- bzrlib/tests/repository_implementations/test_repository.py	2007-06-22 05:05:22 +0000
+++ bzrlib/tests/repository_implementations/test_repository.py	2007-06-28 20:41:30 +0000
@@ -16,6 +16,7 @@
 
 """Tests for bzrdir implementations - tests a bzrdir format."""
 
+from cStringIO import StringIO
 import re
 
 import bzrlib
@@ -415,6 +416,11 @@
         repo._format.rich_root_data
         repo._format.supports_tree_reference
 
+    def test_get_serializer_format(self):
+        repo = self.make_repository('.')
+        format = repo.get_serializer_format()
+        self.assertEqual(repo._serializer.format_num, format)
+
 
 class TestRepositoryLocking(TestCaseWithRepository):
 
@@ -679,3 +685,11 @@
         revtree = branch.repository.revision_tree(REV_ID)
         contents = revtree.get_file_text(FOO_ID)
         self.assertEqual(contents, 'contents of repo/foo\n')
+
+    def test_create_bundle(self):
+        wt = self.make_branch_and_tree('repo')
+        self.build_tree(['repo/file1'])
+        wt.add('file1')
+        wt.commit('file1', rev_id='rev1')
+        fileobj = StringIO()
+        wt.branch.repository.create_bundle('rev1', NULL_REVISION, fileobj)

=== modified file 'bzrlib/tests/test_bundle.py'
--- bzrlib/tests/test_bundle.py	2007-07-11 16:42:37 +0000
+++ bzrlib/tests/test_bundle.py	2007-07-17 13:27:14 +0000
@@ -31,9 +31,10 @@
 from bzrlib.bzrdir import BzrDir
 from bzrlib.bundle.apply_bundle import install_bundle, merge_bundle
 from bzrlib.bundle.bundle_data import BundleTree
-from bzrlib.bundle.serializer import write_bundle, read_bundle
+from bzrlib.bundle.serializer import write_bundle, read_bundle, v09, v4
 from bzrlib.bundle.serializer.v08 import BundleSerializerV08
 from bzrlib.bundle.serializer.v09 import BundleSerializerV09
+from bzrlib.bundle.serializer.v4 import BundleSerializerV4
 from bzrlib.branch import Branch
 from bzrlib.diff import internal_diff
 from bzrlib.errors import (BzrError, TestamentMismatch, NotABundle, BadBundle, 
@@ -42,7 +43,7 @@
 from bzrlib.repofmt import knitrepo
 from bzrlib.osutils import has_symlinks, sha_file
 from bzrlib.tests import (TestCaseInTempDir, TestCaseWithTransport,
-                          TestCase, TestSkipped)
+                          TestCase, TestSkipped, test_commit)
 from bzrlib.transform import TreeTransform
 from bzrlib.workingtree import WorkingTree
 
@@ -348,9 +349,7 @@
                           target.repository, read_bundle(text))
 
 
-class V08BundleTester(TestCaseWithTransport):
-
-    format = '0.8'
+class BundleTester(object):
 
     def bzrdir_format(self):
         format = bzrdir.BzrDirMetaFormat1()
@@ -379,8 +378,6 @@
         rev = self.b1.repository.get_revision(rev_id)
         self.assertEqual(bundle_txt.readline().decode('utf-8'),
                          u'# message:\n')
-
-        open(',,bundle', 'wb').write(bundle_txt.getvalue())
         bundle_txt.seek(0)
         return bundle_txt, rev_ids
 
@@ -551,12 +548,6 @@
         self.tree1.commit('add one', rev_id='a@cset-0-1')
 
         bundle = self.get_valid_bundle('null:', 'a@cset-0-1')
-        # FIXME: The current write_bundle api no longer supports
-        #        setting a custom summary message
-        #        We should re-introduce the ability, and update
-        #        the tests to make sure it works.
-        # bundle = self.get_valid_bundle(None, 'a@cset-0-1',
-        #         message='With a specialized message')
 
         # Make sure we can handle files with spaces, tabs, other
         # bogus characters
@@ -575,9 +566,11 @@
         tt = TreeTransform(self.tree1)
         tt.new_file('executable', tt.root, '#!/bin/sh\n', 'exe-1', True)
         tt.apply()
+        # have to fix length of file-id so that we can predictably rewrite
+        # a (length-prefixed) record containing it later.
+        self.tree1.add('with space.txt', 'withspace-id')
         self.tree1.add([
-                'with space.txt'
-                , 'dir'
+                  'dir'
                 , 'dir/filein subdir.c'
                 , 'dir/WithCaps.txt'
                 , 'dir/ pre space'
@@ -607,8 +600,9 @@
         self.tree1.commit('removed', rev_id='a@cset-0-3')
         
         bundle = self.get_valid_bundle('a@cset-0-2', 'a@cset-0-3')
-        self.assertRaises(TestamentMismatch, self.get_invalid_bundle, 
-                          'a@cset-0-2', 'a@cset-0-3')
+        self.assertRaises((TestamentMismatch,
+            errors.VersionedFileInvalidChecksum), self.get_invalid_bundle,
+            'a@cset-0-2', 'a@cset-0-3')
         # Check a rollup bundle 
         bundle = self.get_valid_bundle('null:', 'a@cset-0-3')
 
@@ -622,7 +616,8 @@
 
         # Modified files
         open('b1/sub/dir/WithCaps.txt', 'ab').write('\nAdding some text\n')
-        open('b1/sub/dir/ pre space', 'ab').write('\r\nAdding some\r\nDOS format lines\r\n')
+        open('b1/sub/dir/ pre space', 'ab').write(
+             '\r\nAdding some\r\nDOS format lines\r\n')
         open('b1/sub/dir/nolastnewline.txt', 'ab').write('\n')
         self.tree1.rename_one('sub/dir/ pre space', 
                               'sub/ start space')
@@ -687,7 +682,8 @@
         
         # Add
         tt.new_file('file', tt.root, '\x00\n\x00\r\x01\n\x02\r\xff', 'binary-1')
-        tt.new_file('file2', tt.root, '\x01\n\x02\r\x03\n\x04\r\xff', 'binary-2')
+        tt.new_file('file2', tt.root, '\x01\n\x02\r\x03\n\x04\r\xff',
+            'binary-2')
         tt.apply()
         self.tree1.commit('add binary', rev_id='b@cset-0-1')
         self.get_valid_bundle('null:', 'b@cset-0-1')
@@ -766,8 +762,20 @@
         rev_ids = write_bundle(self.tree1.branch.repository, 'a@cset-0-3',
                                'a@cset-0-1', bundle_file, format=self.format)
         self.assertNotContainsRe(bundle_file.getvalue(), '\btwo\b')
-        self.assertContainsRe(bundle_file.getvalue(), 'one')
-        self.assertContainsRe(bundle_file.getvalue(), 'three')
+        self.assertContainsRe(self.get_raw(bundle_file), 'one')
+        self.assertContainsRe(self.get_raw(bundle_file), 'three')
+
+    def test_bundle_same_basis(self):
+        """Ensure using the basis as the target doesn't cause an error"""
+        self.tree1 = self.make_branch_and_tree('b1')
+        self.tree1.commit('add file', rev_id='a@cset-0-1')
+        bundle_file = StringIO()
+        rev_ids = write_bundle(self.tree1.branch.repository, 'a@cset-0-1',
+                               'a@cset-0-1', bundle_file)
+
+    @staticmethod
+    def get_raw(bundle_file):
+        return bundle_file.getvalue()
 
     def test_unicode_bundle(self):
         # Handle international characters
@@ -892,7 +900,7 @@
         self.b1 = self.tree1.branch
         self.tree1.commit('message', rev_id='revid1')
         bundle = self.get_valid_bundle('null:', 'revid1')
-        tree = bundle.revision_tree(self.b1.repository, 'revid1')
+        tree = self.get_bundle_tree(bundle, 'revid1')
         self.assertEqual('revid1', tree.inventory.root.revision)
 
     def test_install_revisions(self):
@@ -915,6 +923,163 @@
         tree.commit('One', revprops={'one':'two', 'empty':''}, rev_id='rev1')
         self.b1 = tree.branch
         bundle_sio, revision_ids = self.create_bundle_text('null:', 'rev1')
+        bundle = read_bundle(bundle_sio)
+        revision_info = bundle.revisions[0]
+        self.assertEqual('rev1', revision_info.revision_id)
+        rev = revision_info.as_revision()
+        self.assertEqual({'branch-nick':'tree', 'empty':'', 'one':'two'},
+                         rev.properties)
+
+    def test_bundle_sorted_properties(self):
+        """For stability the writer should write properties in sorted order."""
+        tree = self.make_branch_and_memory_tree('tree')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+
+        tree.add([''], ['TREE_ROOT'])
+        tree.commit('One', rev_id='rev1',
+                    revprops={'a':'4', 'b':'3', 'c':'2', 'd':'1'})
+        self.b1 = tree.branch
+        bundle_sio, revision_ids = self.create_bundle_text('null:', 'rev1')
+        bundle = read_bundle(bundle_sio)
+        revision_info = bundle.revisions[0]
+        self.assertEqual('rev1', revision_info.revision_id)
+        rev = revision_info.as_revision()
+        self.assertEqual({'branch-nick':'tree', 'a':'4', 'b':'3', 'c':'2',
+                          'd':'1'}, rev.properties)
+
+    def test_bundle_unicode_properties(self):
+        """We should be able to round trip a non-ascii property."""
+        tree = self.make_branch_and_memory_tree('tree')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+
+        tree.add([''], ['TREE_ROOT'])
+        # Revisions themselves do not require anything about revision property
+        # keys, other than that they are a basestring, and do not contain
+        # whitespace.
+        # However, Testaments assert than they are str(), and thus should not
+        # be Unicode.
+        tree.commit('One', rev_id='rev1',
+                    revprops={'omega':u'\u03a9', 'alpha':u'\u03b1'})
+        self.b1 = tree.branch
+        bundle_sio, revision_ids = self.create_bundle_text('null:', 'rev1')
+        bundle = read_bundle(bundle_sio)
+        revision_info = bundle.revisions[0]
+        self.assertEqual('rev1', revision_info.revision_id)
+        rev = revision_info.as_revision()
+        self.assertEqual({'branch-nick':'tree', 'omega':u'\u03a9',
+                          'alpha':u'\u03b1'}, rev.properties)
+
+    def test_bundle_with_ghosts(self):
+        tree = self.make_branch_and_tree('tree')
+        self.b1 = tree.branch
+        self.build_tree_contents([('tree/file', 'content1')])
+        tree.add(['file'])
+        tree.commit('rev1')
+        self.build_tree_contents([('tree/file', 'content2')])
+        tree.add_parent_tree_id('ghost')
+        tree.commit('rev2', rev_id='rev2')
+        bundle = self.get_valid_bundle('null:', 'rev2')
+
+    def make_simple_tree(self, format=None):
+        tree = self.make_branch_and_tree('b1', format=format)
+        self.b1 = tree.branch
+        self.build_tree(['b1/file'])
+        tree.add('file')
+        return tree
+
+    def test_across_serializers(self):
+        tree = self.make_simple_tree('knit')
+        tree.commit('hello', rev_id='rev1')
+        tree.commit('hello', rev_id='rev2')
+        bundle = read_bundle(self.create_bundle_text('null:', 'rev2')[0])
+        repo = self.make_repository('repo', format='dirstate-with-subtree')
+        bundle.install_revisions(repo)
+        inv_text = repo.get_inventory_xml('rev2')
+        self.assertNotContainsRe(inv_text, 'format="5"')
+        self.assertContainsRe(inv_text, 'format="7"')
+
+    def test_across_models(self):
+        tree = self.make_simple_tree('knit')
+        tree.commit('hello', rev_id='rev1')
+        tree.commit('hello', rev_id='rev2')
+        bundle = read_bundle(self.create_bundle_text('null:', 'rev2')[0])
+        repo = self.make_repository('repo', format='dirstate-with-subtree')
+        bundle.install_revisions(repo)
+        inv = repo.get_inventory('rev2')
+        self.assertEqual('rev2', inv.root.revision)
+        root_vf = repo.weave_store.get_weave(inv.root.file_id,
+                                             repo.get_transaction())
+        self.assertEqual(root_vf.versions(), ['rev1', 'rev2'])
+
+    def test_across_models_incompatible(self):
+        tree = self.make_simple_tree('dirstate-with-subtree')
+        tree.commit('hello', rev_id='rev1')
+        tree.commit('hello', rev_id='rev2')
+        try:
+            bundle = read_bundle(self.create_bundle_text('null:', 'rev1')[0])
+        except errors.IncompatibleBundleFormat:
+            raise TestSkipped("Format 0.8 doesn't work with knit3")
+        repo = self.make_repository('repo', format='knit')
+        bundle.install_revisions(repo)
+
+        bundle = read_bundle(self.create_bundle_text('null:', 'rev2')[0])
+        self.assertRaises(errors.IncompatibleRevision,
+                          bundle.install_revisions, repo)
+
+    def test_get_merge_request(self):
+        tree = self.make_simple_tree()
+        tree.commit('hello', rev_id='rev1')
+        tree.commit('hello', rev_id='rev2')
+        bundle = read_bundle(self.create_bundle_text('null:', 'rev1')[0])
+        result = bundle.get_merge_request(tree.branch.repository)
+        self.assertEqual((None, 'rev1', 'inapplicable'), result)
+
+    def test_with_subtree(self):
+        tree = self.make_branch_and_tree('tree',
+                                         format='dirstate-with-subtree')
+        self.b1 = tree.branch
+        subtree = self.make_branch_and_tree('tree/subtree',
+                                            format='dirstate-with-subtree')
+        tree.add('subtree')
+        tree.commit('hello', rev_id='rev1')
+        try:
+            bundle = read_bundle(self.create_bundle_text('null:', 'rev1')[0])
+        except errors.IncompatibleBundleFormat:
+            raise TestSkipped("Format 0.8 doesn't work with knit3")
+        if isinstance(bundle, v09.BundleInfo09):
+            raise TestSkipped("Format 0.9 doesn't work with subtrees")
+        repo = self.make_repository('repo', format='knit')
+        self.assertRaises(errors.IncompatibleRevision,
+                          bundle.install_revisions, repo)
+        repo2 = self.make_repository('repo2', format='dirstate-with-subtree')
+        bundle.install_revisions(repo2)
+
+    def test_revision_id_with_slash(self):
+        self.tree1 = self.make_branch_and_tree('tree')
+        self.b1 = self.tree1.branch
+        try:
+            self.tree1.commit('Revision/id/with/slashes', rev_id='rev/id')
+        except ValueError:
+            raise TestSkipped("Repository doesn't support revision ids with"
+                              " slashes")
+        bundle = self.get_valid_bundle('null:', 'rev/id')
+
+
+class V08BundleTester(BundleTester, TestCaseWithTransport):
+
+    format = '0.8'
+
+    def test_bundle_empty_property(self):
+        """Test serializing revision properties with an empty value."""
+        tree = self.make_branch_and_memory_tree('tree')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.add([''], ['TREE_ROOT'])
+        tree.commit('One', revprops={'one':'two', 'empty':''}, rev_id='rev1')
+        self.b1 = tree.branch
+        bundle_sio, revision_ids = self.create_bundle_text('null:', 'rev1')
         self.assertContainsRe(bundle_sio.getvalue(),
                               '# properties:\n'
                               '#   branch-nick: tree\n'
@@ -928,6 +1093,10 @@
         self.assertEqual({'branch-nick':'tree', 'empty':'', 'one':'two'},
                          rev.properties)
 
+    def get_bundle_tree(self, bundle, revision_id):
+        repository = self.make_repository('repo')
+        return bundle.revision_tree(repository, 'revid1')
+
     def test_bundle_empty_property_alt(self):
         """Test serializing revision properties with an empty value.
 
@@ -1037,7 +1206,157 @@
         return format
 
 
-class MungedBundleTester(TestCaseWithTransport):
+class V4BundleTester(BundleTester, TestCaseWithTransport):
+
+    format = '4alpha'
+
+    def get_valid_bundle(self, base_rev_id, rev_id, checkout_dir=None):
+        """Create a bundle from base_rev_id -> rev_id in built-in branch.
+        Make sure that the text generated is valid, and that it
+        can be applied against the base, and generate the same information.
+        
+        :return: The in-memory bundle 
+        """
+        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)
+
+        # This should also validate the generated bundle 
+        bundle = read_bundle(bundle_txt)
+        repository = self.b1.repository
+        for bundle_rev in bundle.real_revisions:
+            # These really should have already been checked when we read the
+            # bundle, since it computes the sha1 hash for the revision, which
+            # only will match if everything is okay, but lets be explicit about
+            # it
+            branch_rev = repository.get_revision(bundle_rev.revision_id)
+            for a in ('inventory_sha1', 'revision_id', 'parent_ids',
+                      'timestamp', 'timezone', 'message', 'committer', 
+                      'parent_ids', 'properties'):
+                self.assertEqual(getattr(branch_rev, a), 
+                                 getattr(bundle_rev, a))
+            self.assertEqual(len(branch_rev.parent_ids), 
+                             len(bundle_rev.parent_ids))
+        self.assertEqual(set(rev_ids),
+                         set([r.revision_id for r in bundle.real_revisions]))
+        self.valid_apply_bundle(base_rev_id, bundle,
+                                   checkout_dir=checkout_dir)
+
+        return bundle
+
+    def get_invalid_bundle(self, base_rev_id, rev_id):
+        """Create a bundle from base_rev_id -> rev_id in built-in branch.
+        Munge the text so that it's invalid.
+
+        :return: The in-memory bundle
+        """
+        from bzrlib.bundle import serializer
+        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)
+        new_text = self.get_raw(StringIO(''.join(bundle_txt)))
+        new_text = new_text.replace('<file file_id="exe-1"',
+                                    '<file executable="y" file_id="exe-1"')
+        new_text = new_text.replace('B372', 'B387')
+        bundle_txt = StringIO()
+        bundle_txt.write(serializer._get_bundle_header('4alpha'))
+        bundle_txt.write('\n')
+        bundle_txt.write(new_text.encode('bz2'))
+        bundle_txt.seek(0)
+        bundle = read_bundle(bundle_txt)
+        self.valid_apply_bundle(base_rev_id, bundle)
+        return bundle
+
+    def create_bundle_text(self, base_rev_id, rev_id):
+        bundle_txt = StringIO()
+        rev_ids = write_bundle(self.b1.repository, rev_id, base_rev_id, 
+                               bundle_txt, format=self.format)
+        bundle_txt.seek(0)
+        self.assertEqual(bundle_txt.readline(), 
+                         '# Bazaar revision bundle v%s\n' % self.format)
+        self.assertEqual(bundle_txt.readline(), '#\n')
+        rev = self.b1.repository.get_revision(rev_id)
+        bundle_txt.seek(0)
+        return bundle_txt, rev_ids
+
+    def get_bundle_tree(self, bundle, revision_id):
+        repository = self.make_repository('repo')
+        bundle.install_revisions(repository)
+        return repository.revision_tree(revision_id)
+
+    def test_creation(self):
+        tree = self.make_branch_and_tree('tree')
+        self.build_tree_contents([('tree/file', 'contents1\nstatic\n')])
+        tree.add('file', 'fileid-2')
+        tree.commit('added file', rev_id='rev1')
+        self.build_tree_contents([('tree/file', 'contents2\nstatic\n')])
+        tree.commit('changed file', rev_id='rev2')
+        s = StringIO()
+        serializer = BundleSerializerV4('1.0')
+        serializer.write(tree.branch.repository, ['rev1', 'rev2'], {}, s)
+        s.seek(0)
+        tree2 = self.make_branch_and_tree('target')
+        target_repo = tree2.branch.repository
+        install_bundle(target_repo, serializer.read(s))
+        vf = target_repo.weave_store.get_weave('fileid-2',
+            target_repo.get_transaction())
+        self.assertEqual('contents1\nstatic\n', vf.get_text('rev1'))
+        self.assertEqual('contents2\nstatic\n', vf.get_text('rev2'))
+        rtree = target_repo.revision_tree('rev2')
+        inventory_vf = target_repo.get_inventory_weave()
+        self.assertEqual(['rev1'], inventory_vf.get_parents('rev2'))
+        self.assertEqual('changed file',
+                         target_repo.get_revision('rev2').message)
+
+    @staticmethod
+    def get_raw(bundle_file):
+        bundle_file.seek(0)
+        line = bundle_file.readline()
+        line = bundle_file.readline()
+        lines = bundle_file.readlines()
+        return ''.join(lines).decode('bz2')
+
+    def test_copy_signatures(self):
+        tree_a = self.make_branch_and_tree('tree_a')
+        import bzrlib.gpg
+        import bzrlib.commit as commit
+        oldstrategy = bzrlib.gpg.GPGStrategy
+        branch = tree_a.branch
+        repo_a = branch.repository
+        tree_a.commit("base", allow_pointless=True, rev_id='A')
+        self.failIf(branch.repository.has_signature_for_revision_id('A'))
+        try:
+            from bzrlib.testament import Testament
+            # monkey patch gpg signing mechanism
+            bzrlib.gpg.GPGStrategy = bzrlib.gpg.LoopbackGPGStrategy
+            new_config = test_commit.MustSignConfig(branch)
+            commit.Commit(config=new_config).commit(message="base",
+                                                    allow_pointless=True,
+                                                    rev_id='B',
+                                                    working_tree=tree_a)
+            def sign(text):
+                return bzrlib.gpg.LoopbackGPGStrategy(None).sign(text)
+            self.assertTrue(repo_a.has_signature_for_revision_id('B'))
+        finally:
+            bzrlib.gpg.GPGStrategy = oldstrategy
+        tree_b = self.make_branch_and_tree('tree_b')
+        repo_b = tree_b.branch.repository
+        s = StringIO()
+        serializer = BundleSerializerV4('4alpha')
+        serializer.write(tree_a.branch.repository, ['A', 'B'], {}, s)
+        s.seek(0)
+        install_bundle(repo_b, serializer.read(s))
+        self.assertTrue(repo_b.has_signature_for_revision_id('B'))
+        self.assertEqual(repo_b.get_signature_text('B'),
+                         repo_a.get_signature_text('B'))
+        s.seek(0)
+        # ensure repeat installs are harmless
+        install_bundle(repo_b, serializer.read(s))
+
+
+class V4WeaveBundleTester(V4BundleTester):
+
+    def bzrdir_format(self):
+        return 'metaweave'
+
+
+class MungedBundleTester(object):
 
     def build_test_bundle(self):
         wt = self.make_branch_and_tree('b1')
@@ -1052,8 +1371,8 @@
 
         bundle_txt = StringIO()
         rev_ids = write_bundle(wt.branch.repository, 'a@cset-0-2',
-                               'a@cset-0-1', bundle_txt)
-        self.assertEqual(['a@cset-0-2'], rev_ids)
+                               'a@cset-0-1', bundle_txt, self.format)
+        self.assertEqual(set(['a@cset-0-2']), set(rev_ids))
         bundle_txt.seek(0, 0)
         return bundle_txt
 
@@ -1088,6 +1407,11 @@
         bundle = read_bundle(bundle_txt)
         self.check_valid(bundle)
 
+
+class MungedBundleTesterV09(TestCaseWithTransport, MungedBundleTester):
+
+    format = '0.9'
+
     def test_missing_trailing_whitespace(self):
         bundle_txt = self.build_test_bundle()
 
@@ -1121,3 +1445,58 @@
         bundle = read_bundle(bundle_txt)
         self.check_valid(bundle)
 
+
+class MungedBundleTesterV4(TestCaseWithTransport, MungedBundleTester):
+
+    format = '4alpha'
+
+
+class TestBundleWriterReader(TestCase):
+
+    def test_roundtrip_record(self):
+        fileobj = StringIO()
+        writer = v4.BundleWriter(fileobj)
+        writer.begin()
+        writer.add_info_record(foo='bar')
+        writer._add_record("Record body", {'parents': ['1', '3'],
+            'storage_kind':'fulltext'}, 'file', 'revid', 'fileid')
+        writer.end()
+        fileobj.seek(0)
+        record_iter = v4.BundleReader(fileobj).iter_records()
+        record = record_iter.next()
+        self.assertEqual((None, {'foo': 'bar', 'storage_kind': 'header'},
+            'info', None, None), record)
+        record = record_iter.next()
+        self.assertEqual(("Record body", {'storage_kind': 'fulltext',
+                          'parents': ['1', '3']}, 'file', 'revid', 'fileid'),
+                          record)
+
+    def test_encode_name(self):
+        self.assertEqual('revision/rev1',
+            v4.BundleWriter.encode_name('revision', 'rev1'))
+        self.assertEqual('file/rev//1/file-id-1',
+            v4.BundleWriter.encode_name('file', 'rev/1', 'file-id-1'))
+        self.assertEqual('info',
+            v4.BundleWriter.encode_name('info', None, None))
+
+    def test_decode_name(self):
+        self.assertEqual(('revision', 'rev1', None),
+            v4.BundleReader.decode_name('revision/rev1'))
+        self.assertEqual(('file', 'rev/1', 'file-id-1'),
+            v4.BundleReader.decode_name('file/rev//1/file-id-1'))
+        self.assertEqual(('info', None, None),
+                         v4.BundleReader.decode_name('info'))
+
+    def test_too_many_names(self):
+        fileobj = StringIO()
+        writer = v4.BundleWriter(fileobj)
+        writer.begin()
+        writer.add_info_record(foo='bar')
+        writer._container.add_bytes_record('blah', ['two', 'names'])
+        writer.end()
+        fileobj.seek(0)
+        record_iter = v4.BundleReader(fileobj).iter_records()
+        record = record_iter.next()
+        self.assertEqual((None, {'foo': 'bar', 'storage_kind': 'header'},
+            'info', None, None), record)
+        self.assertRaises(BadBundle, record_iter.next)

=== modified file 'bzrlib/tests/test_graph.py'
--- bzrlib/tests/test_graph.py	2007-06-21 01:58:29 +0000
+++ bzrlib/tests/test_graph.py	2007-06-27 15:12:32 +0000
@@ -181,6 +181,12 @@
         self.assertEqual(set(['rev1']), graph.find_lca('rev1', 'rev1'))
         self.assertEqual(set(['rev1']), graph.find_lca('rev2a', 'rev2b'))
 
+    def test_no_unique_lca(self):
+        """Test error when one revision is not in the graph"""
+        graph = self.make_graph(ancestry_1)
+        self.assertRaises(errors.NoCommonAncestor, graph.find_unique_lca,
+                          'rev1', '1rev')
+
     def test_lca_criss_cross(self):
         """Test least-common-ancestor after a criss-cross merge."""
         graph = self.make_graph(criss_cross)

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2007-07-05 19:39:28 +0000
+++ bzrlib/tests/test_knit.py	2007-07-17 13:27:14 +0000
@@ -41,6 +41,7 @@
     _KnitData,
     _KnitIndex,
     WeaveToKnit,
+    KnitSequenceMatcher,
     )
 from bzrlib.osutils import split_lines
 from bzrlib.tests import TestCase, TestCaseWithTransport, Feature
@@ -916,12 +917,57 @@
     def test_delta(self):
         """Expression of knit delta as lines"""
         k = self.make_test_knit()
+        KnitContent
         td = list(line_delta(TEXT_1.splitlines(True),
                              TEXT_1A.splitlines(True)))
         self.assertEqualDiff(''.join(td), delta_1_1a)
         out = apply_line_delta(TEXT_1.splitlines(True), td)
         self.assertEqualDiff(''.join(out), TEXT_1A)
 
+    def assertDerivedBlocksEqual(self, source, target, noeol=False):
+        """Assert that the derived matching blocks match real output"""
+        source_lines = source.splitlines(True)
+        target_lines = target.splitlines(True)
+        def nl(line):
+            if noeol and not line.endswith('\n'):
+                return line + '\n'
+            else:
+                return line
+        source_content = KnitContent([(None, nl(l)) for l in source_lines])
+        target_content = KnitContent([(None, nl(l)) for l in target_lines])
+        line_delta = source_content.line_delta(target_content)
+        delta_blocks = list(KnitContent.get_line_delta_blocks(line_delta,
+            source_lines, target_lines))
+        matcher = KnitSequenceMatcher(None, source_lines, target_lines)
+        matcher_blocks = list(list(matcher.get_matching_blocks()))
+        self.assertEqual(matcher_blocks, delta_blocks)
+
+    def test_get_line_delta_blocks(self):
+        self.assertDerivedBlocksEqual('a\nb\nc\n', 'q\nc\n')
+        self.assertDerivedBlocksEqual(TEXT_1, TEXT_1)
+        self.assertDerivedBlocksEqual(TEXT_1, TEXT_1A)
+        self.assertDerivedBlocksEqual(TEXT_1, TEXT_1B)
+        self.assertDerivedBlocksEqual(TEXT_1B, TEXT_1A)
+        self.assertDerivedBlocksEqual(TEXT_1A, TEXT_1B)
+        self.assertDerivedBlocksEqual(TEXT_1A, '')
+        self.assertDerivedBlocksEqual('', TEXT_1A)
+        self.assertDerivedBlocksEqual('', '')
+        self.assertDerivedBlocksEqual('a\nb\nc', 'a\nb\nc\nd')
+
+    def test_get_line_delta_blocks_noeol(self):
+        """Handle historical knit deltas safely
+
+        Some existing knit deltas don't consider the last line to differ
+        when the only difference whether it has a final newline.
+
+        New knit deltas appear to always consider the last line to differ
+        in this case.
+        """
+        self.assertDerivedBlocksEqual('a\nb\nc', 'a\nb\nc\nd\n', noeol=True)
+        self.assertDerivedBlocksEqual('a\nb\nc\nd\n', 'a\nb\nc', noeol=True)
+        self.assertDerivedBlocksEqual('a\nb\nc\n', 'a\nb\nc', noeol=True)
+        self.assertDerivedBlocksEqual('a\nb\nc', 'a\nb\nc\n', noeol=True)
+
     def test_add_with_parents(self):
         """Store in knit with parents"""
         k = self.make_test_knit()

=== modified file 'bzrlib/tests/test_merge_directive.py'
--- bzrlib/tests/test_merge_directive.py	2007-06-19 21:22:56 +0000
+++ bzrlib/tests/test_merge_directive.py	2007-07-03 22:44:19 +0000
@@ -31,6 +31,15 @@
 #\x20
 booga"""
 
+OUTPUT1_2 = """# Bazaar merge directive format 2 (Bazaar 0.18)
+# revision_id: example:
+# target_branch: http://example.com
+# testament_sha1: sha
+# timestamp: 1970-01-01 00:09:33 +0002
+# base_revision_id: null:
+#\x20
+# Begin bundle
+booga"""
 
 OUTPUT2 = """# Bazaar merge directive format 1
 # revision_id: example:
@@ -42,6 +51,17 @@
 #\x20
 booga"""
 
+OUTPUT2_2 = """# Bazaar merge directive format 2 (Bazaar 0.18)
+# revision_id: example:
+# target_branch: http://example.com
+# testament_sha1: sha
+# timestamp: 1970-01-01 00:09:33 +0002
+# source_branch: http://example.org
+# message: Hi mom!
+# base_revision_id: null:
+#\x20
+# Begin patch
+booga"""
 
 INPUT1 = """
 I was thinking today about creating a merge directive.
@@ -65,47 +85,60 @@
 booga""".splitlines(True)
 
 
-class TestMergeDirective(tests.TestCase):
+INPUT1_2 = """
+I was thinking today about creating a merge directive.
+
+So I did.
+
+Here it is.
+
+(I've pasted it in the body of this message)
+
+Aaron
+
+# Bazaar merge directive format 2 (Bazaar 0.18)\r
+# revision_id: example:
+# target_branch: http://example.com
+# testament_sha1: sha
+# timestamp: 1970-01-01 00:09:33 +0002
+# source_branch: http://example.org
+# base_revision_id: null:
+# message: Hi mom!
+#\x20
+# Begin patch
+booga""".splitlines(True)
+
+
+class TestMergeDirective(object):
 
     def test_merge_source(self):
         time = 500000.0
         timezone = 5 * 3600
-        self.assertRaises(errors.NoMergeSource, merge_directive.MergeDirective,
+        self.assertRaises(errors.NoMergeSource, self.make_merge_directive,
             'example:', 'sha', time, timezone, 'http://example.com')
-        self.assertRaises(errors.NoMergeSource, merge_directive.MergeDirective,
+        self.assertRaises(errors.NoMergeSource, self.make_merge_directive,
             'example:', 'sha', time, timezone, 'http://example.com',
             patch_type='diff')
-        merge_directive.MergeDirective('example:', 'sha', time, timezone,
+        self.make_merge_directive('example:', 'sha', time, timezone,
             'http://example.com', source_branch='http://example.org')
-        md = merge_directive.MergeDirective('null:', 'sha', time, timezone,
+        md = self.make_merge_directive('null:', 'sha', time, timezone,
             'http://example.com', patch='blah', patch_type='bundle')
         self.assertIs(None, md.source_branch)
-        md2 = merge_directive.MergeDirective('null:', 'sha', time, timezone,
+        md2 = self.make_merge_directive('null:', 'sha', time, timezone,
             'http://example.com', patch='blah', patch_type='bundle',
             source_branch='bar')
         self.assertEqual('bar', md2.source_branch)
 
-    def test_require_patch(self):
-        time = 500.0
-        timezone = 120
-        self.assertRaises(errors.PatchMissing, merge_directive.MergeDirective,
-            'example:', 'sha', time, timezone, 'http://example.com',
-            patch_type='bundle')
-        md = merge_directive.MergeDirective('example:', 'sha1', time, timezone,
-            'http://example.com', source_branch="http://example.org",
-            patch='', patch_type='diff')
-        self.assertEqual(md.patch, '')
-
     def test_serialization(self):
         time = 453
         timezone = 120
-        md = merge_directive.MergeDirective('example:', 'sha', time, timezone,
+        md = self.make_merge_directive('example:', 'sha', time, timezone,
             'http://example.com', patch='booga', patch_type='bundle')
-        self.assertEqualDiff(OUTPUT1, ''.join(md.to_lines()))
-        md = merge_directive.MergeDirective('example:', 'sha', time, timezone,
+        self.assertEqualDiff(self.OUTPUT1, ''.join(md.to_lines()))
+        md = self.make_merge_directive('example:', 'sha', time, timezone,
             'http://example.com', source_branch="http://example.org",
             patch='booga', patch_type='diff', message="Hi mom!")
-        self.assertEqualDiff(OUTPUT2, ''.join(md.to_lines()))
+        self.assertEqualDiff(self.OUTPUT2, ''.join(md.to_lines()))
 
     def test_deserialize_junk(self):
         time = 501
@@ -117,7 +150,7 @@
                           merge_directive.MergeDirective.from_lines, [])
 
     def test_deserialize_leading_junk(self):
-        md = merge_directive.MergeDirective.from_lines(INPUT1)
+        md = merge_directive.MergeDirective.from_lines(self.INPUT1)
         self.assertEqual('example:', md.revision_id)
         self.assertEqual('sha', md.testament_sha1)
         self.assertEqual('http://example.com', md.target_branch)
@@ -131,7 +164,7 @@
     def test_roundtrip(self):
         time = 500000
         timezone = 7.5 * 3600
-        md = merge_directive.MergeDirective('example:', 'sha', time, timezone,
+        md = self.make_merge_directive('example:', 'sha', time, timezone,
             'http://example.com', source_branch="http://example.org",
             patch='booga', patch_type='diff')
         md2 = merge_directive.MergeDirective.from_lines(md.to_lines())
@@ -145,20 +178,79 @@
         self.assertEqual('diff', md2.patch_type)
         self.assertEqual('booga', md2.patch)
         self.assertEqual(None, md2.message)
-        md.patch = "# Bazaar revision bundle v0.9\n#\n"
+        self.set_bundle(md, "# Bazaar revision bundle v0.9\n#\n")
         md.message = "Hi mom!"
-        md3 = merge_directive.MergeDirective.from_lines(md.to_lines())
-        self.assertEqual("# Bazaar revision bundle v0.9\n#\n", md3.patch)
+        lines = md.to_lines()
+        md3 = merge_directive.MergeDirective.from_lines(lines)
+        self.assertEqual("# Bazaar revision bundle v0.9\n#\n", md3.bundle)
         self.assertEqual("bundle", md3.patch_type)
         self.assertContainsRe(md3.to_lines()[0],
             '^# Bazaar merge directive format ')
         self.assertEqual("Hi mom!", md3.message)
-        md3.patch_type = None
-        md3.patch = None
+        md3.clear_payload()
+        self.assertIs(None, md3.get_raw_bundle())
         md4 = merge_directive.MergeDirective.from_lines(md3.to_lines())
         self.assertIs(None, md4.patch_type)
 
 
+class TestMergeDirective1(tests.TestCase, TestMergeDirective):
+    """Test merge directive format 1"""
+
+    INPUT1 = INPUT1
+
+    OUTPUT1 = OUTPUT1
+
+    OUTPUT2 = OUTPUT2
+
+    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, patch_type=None,
+                 source_branch=None, message=None):
+        return merge_directive.MergeDirective(revision_id, testament_sha1,
+                 time, timezone, target_branch, patch, patch_type,
+                 source_branch, message)
+
+    @staticmethod
+    def set_bundle(md, value):
+        md.patch = value
+
+    def test_require_patch(self):
+        time = 500.0
+        timezone = 120
+        self.assertRaises(errors.PatchMissing, merge_directive.MergeDirective,
+            'example:', 'sha', time, timezone, 'http://example.com',
+            patch_type='bundle')
+        md = merge_directive.MergeDirective('example:', 'sha1', time, timezone,
+            'http://example.com', source_branch="http://example.org",
+            patch='', patch_type='diff')
+        self.assertEqual(md.patch, '')
+
+
+class TestMergeDirective2(tests.TestCase, TestMergeDirective):
+    """Test merge directive format 2"""
+
+    INPUT1 = INPUT1_2
+
+    OUTPUT1 = OUTPUT1_2
+
+    OUTPUT2 = OUTPUT2_2
+
+    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, patch_type=None,
+                 source_branch=None, message=None, base_revision_id='null:'):
+        if patch_type == 'bundle':
+            bundle = patch
+            patch = None
+        else:
+            bundle = None
+        return merge_directive.MergeDirective2(revision_id, testament_sha1,
+            time, timezone, target_branch, patch, source_branch, message,
+            bundle, base_revision_id)
+
+    @staticmethod
+    def set_bundle(md, value):
+        md.bundle = value
+
+
 EMAIL1 = """To: pqm@example.com
 From: J. Random Hacker <jrandom@example.com>
 Subject: Commit of rev2a
@@ -172,6 +264,19 @@
 """
 
 
+EMAIL1_2 = """To: pqm@example.com
+From: J. Random Hacker <jrandom@example.com>
+Subject: Commit of rev2a
+
+# Bazaar merge directive format 2 \\(Bazaar 0.18\\)
+# revision_id: rev2a
+# target_branch: (.|\n)*
+# testament_sha1: .*
+# timestamp: 1970-01-01 00:08:56 \\+0001
+# source_branch: (.|\n)*
+"""
+
+
 EMAIL2 = """To: pqm@example.com
 From: J. Random Hacker <jrandom@example.com>
 Subject: Commit of rev2a with special message
@@ -185,8 +290,20 @@
 # message: Commit of rev2a with special message
 """
 
-
-class TestMergeDirectiveBranch(tests.TestCaseWithTransport):
+EMAIL2_2 = """To: pqm@example.com
+From: J. Random Hacker <jrandom@example.com>
+Subject: Commit of rev2a with special message
+
+# Bazaar merge directive format 2 \\(Bazaar 0.18\\)
+# revision_id: rev2a
+# target_branch: (.|\n)*
+# testament_sha1: .*
+# timestamp: 1970-01-01 00:08:56 \\+0001
+# source_branch: (.|\n)*
+# message: Commit of rev2a with special message
+"""
+
+class TestMergeDirectiveBranch(object):
 
     def make_trees(self):
         tree_a = self.make_branch_and_tree('tree_a')
@@ -198,22 +315,22 @@
         tree_b = tree_a.bzrdir.sprout('tree_b').open_workingtree()
         branch_c = tree_a.bzrdir.sprout('branch_c').open_branch()
         tree_b.commit('message', rev_id='rev2b')
-        self.build_tree_contents([('tree_a/file', 'content_a\ncontent_c\n')])
+        self.build_tree_contents([('tree_a/file', 'content_a\ncontent_c \n')])
         tree_a.commit('Commit of rev2a', rev_id='rev2a')
         return tree_a, tree_b, branch_c
 
     def test_empty_target(self):
         tree_a, tree_b, branch_c = self.make_trees()
         tree_d = self.make_branch_and_tree('tree_d')
-        md2 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 120, tree_d.branch.base,
-            patch_type='diff', public_branch=tree_a.branch.base)
+        md2 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 120,
+            tree_d.branch.base, patch_type='diff',
+            public_branch=tree_a.branch.base)
 
     def test_generate_patch(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md2 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 120, tree_b.branch.base,
-            patch_type='diff', public_branch=tree_a.branch.base)
+        md2 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 120,
+            tree_b.branch.base, patch_type='diff',
+            public_branch=tree_a.branch.base)
         self.assertNotContainsRe(md2.patch, 'Bazaar revision bundle')
         self.assertContainsRe(md2.patch, '\\+content_c')
         self.assertNotContainsRe(md2.patch, '\\+\\+\\+ b/')
@@ -222,40 +339,38 @@
     def test_public_branch(self):
         tree_a, tree_b, branch_c = self.make_trees()
         self.assertRaises(errors.PublicBranchOutOfDate,
-            merge_directive.MergeDirective.from_objects,
-            tree_a.branch.repository, 'rev2a', 500, 144, tree_b.branch.base,
-            public_branch=branch_c.base, patch_type='diff')
+            self.from_objects, tree_a.branch.repository, 'rev2a', 500, 144,
+            tree_b.branch.base, public_branch=branch_c.base, patch_type='diff')
+        self.assertRaises(errors.PublicBranchOutOfDate,
+            self.from_objects, tree_a.branch.repository, 'rev2a', 500, 144,
+            tree_b.branch.base, public_branch=branch_c.base, patch_type=None)
         # public branch is not checked if patch format is bundle.
-        md1 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 144, tree_b.branch.base,
-            public_branch=branch_c.base)
+        md1 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 144,
+            tree_b.branch.base, public_branch=branch_c.base)
         # public branch is provided with a bundle, despite possibly being out
         # of date, because it's not required if a bundle is present.
         self.assertEqual(md1.source_branch, branch_c.base)
         # Once we update the public branch, we can generate a diff.
         branch_c.pull(tree_a.branch)
-        md3 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 144, tree_b.branch.base,
-            patch_type=None, public_branch=branch_c.base)
+        md3 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 144,
+            tree_b.branch.base, patch_type=None, public_branch=branch_c.base)
 
     def test_use_public_submit_branch(self):
         tree_a, tree_b, branch_c = self.make_trees()
         branch_c.pull(tree_a.branch)
-        md = merge_directive.MergeDirective.from_objects(
-             tree_a.branch.repository, 'rev2a', 500, 144, tree_b.branch.base,
-             patch_type=None, public_branch=branch_c.base)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 144,
+            tree_b.branch.base, patch_type=None, public_branch=branch_c.base)
         self.assertEqual(md.target_branch, tree_b.branch.base)
         tree_b.branch.set_public_branch('http://example.com')
-        md2 = merge_directive.MergeDirective.from_objects(
+        md2 = self.from_objects(
               tree_a.branch.repository, 'rev2a', 500, 144, tree_b.branch.base,
               patch_type=None, public_branch=branch_c.base)
         self.assertEqual(md2.target_branch, 'http://example.com')
 
     def test_message(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md3 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 120, tree_b.branch.base,
-            patch_type=None, public_branch=branch_c.base,
+        md3 = self.from_objects(tree_a.branch.repository, 'rev1', 500, 120,
+            tree_b.branch.base, patch_type=None, public_branch=branch_c.base,
             message='Merge message')
         md3.to_lines()
         self.assertIs(None, md3.patch)
@@ -263,10 +378,10 @@
 
     def test_generate_bundle(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md1 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 120, tree_b.branch.base,
-            public_branch=branch_c.base)
-        self.assertContainsRe(md1.patch, 'Bazaar revision bundle')
+        md1 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 120,
+            tree_b.branch.base, public_branch=branch_c.base)
+
+        self.assertContainsRe(md1.get_raw_bundle(), 'Bazaar revision bundle')
         self.assertContainsRe(md1.patch, '\\+content_c')
         self.assertNotContainsRe(md1.patch, '\\+content_a')
         self.assertContainsRe(md1.patch, '\\+content_c')
@@ -274,9 +389,8 @@
 
     def test_broken_bundle(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md1 = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 120, tree_b.branch.base,
-            public_branch=branch_c.base)
+        md1 = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 120,
+            tree_b.branch.base, public_branch=branch_c.base)
         lines = md1.to_lines()
         lines = [l.replace('\n', '\r\n') for l in lines]
         md2 = merge_directive.MergeDirective.from_lines(lines)
@@ -290,7 +404,7 @@
                 return self
             def gpg_signing_command(self):
                 return 'loopback'
-        md = merge_directive.MergeDirective('example:', 'sha', time, timezone,
+        md = self.make_merge_directive('example:', 'sha', time, timezone,
             'http://example.com', source_branch="http://example.org",
             patch='booga', patch_type='diff')
         old_strategy = gpg.GPGStrategy
@@ -305,30 +419,79 @@
 
     def test_email(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 476, 60, tree_b.branch.base,
-            patch_type=None, public_branch=tree_a.branch.base)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 476, 60,
+            tree_b.branch.base, patch_type=None,
+            public_branch=tree_a.branch.base)
         message = md.to_email('pqm@example.com', tree_a.branch)
-        self.assertContainsRe(message.as_string(), EMAIL1)
+        self.assertContainsRe(message.as_string(), self.EMAIL1)
         md.message = 'Commit of rev2a with special message'
         message = md.to_email('pqm@example.com', tree_a.branch)
-        self.assertContainsRe(message.as_string(), EMAIL2)
+        self.assertContainsRe(message.as_string(), self.EMAIL2)
 
     def test_install_revisions_branch(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 36, tree_b.branch.base,
-            patch_type=None, public_branch=tree_a.branch.base)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type=None,
+            public_branch=tree_a.branch.base)
         self.assertFalse(tree_b.branch.repository.has_revision('rev2a'))
         revision = md.install_revisions(tree_b.branch.repository)
         self.assertEqual('rev2a', revision)
         self.assertTrue(tree_b.branch.repository.has_revision('rev2a'))
 
+    def test_get_merge_request(self):
+        tree_a, tree_b, branch_c = self.make_trees()
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type='bundle',
+            public_branch=tree_a.branch.base)
+        self.assertFalse(tree_b.branch.repository.has_revision('rev2a'))
+        md.install_revisions(tree_b.branch.repository)
+        base, revision, verified = md.get_merge_request(
+            tree_b.branch.repository)
+        if isinstance(md, merge_directive.MergeDirective):
+            self.assertIs(None, base)
+            self.assertEqual('inapplicable', verified)
+        else:
+            self.assertEqual('rev1', base)
+            self.assertEqual('verified', verified)
+        self.assertEqual('rev2a', revision)
+        self.assertTrue(tree_b.branch.repository.has_revision('rev2a'))
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type=None,
+            public_branch=tree_a.branch.base)
+        base, revision, verified = md.get_merge_request(
+            tree_b.branch.repository)
+        if isinstance(md, merge_directive.MergeDirective):
+            self.assertIs(None, base)
+            self.assertEqual('inapplicable', verified)
+        else:
+            self.assertEqual('rev1', base)
+            self.assertEqual('inapplicable', verified)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type='diff',
+            public_branch=tree_a.branch.base)
+        base, revision, verified = md.get_merge_request(
+            tree_b.branch.repository)
+        if isinstance(md, merge_directive.MergeDirective):
+            self.assertIs(None, base)
+            self.assertEqual('inapplicable', verified)
+        else:
+            self.assertEqual('rev1', base)
+            self.assertEqual('verified', verified)
+        md.patch='asdf'
+        base, revision, verified = md.get_merge_request(
+            tree_b.branch.repository)
+        if isinstance(md, merge_directive.MergeDirective):
+            self.assertIs(None, base)
+            self.assertEqual('inapplicable', verified)
+        else:
+            self.assertEqual('rev1', base)
+            self.assertEqual('failed', verified)
+
     def test_install_revisions_bundle(self):
         tree_a, tree_b, branch_c = self.make_trees()
-        md = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 36, tree_b.branch.base,
-            patch_type='bundle', public_branch=tree_a.branch.base)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type='bundle',
+            public_branch=tree_a.branch.base)
         self.assertFalse(tree_b.branch.repository.has_revision('rev2a'))
         revision = md.install_revisions(tree_b.branch.repository)
         self.assertEqual('rev2a', revision)
@@ -337,9 +500,98 @@
     def test_get_target_revision_nofetch(self):
         tree_a, tree_b, branch_c = self.make_trees()
         tree_b.branch.fetch(tree_a.branch)
-        md = merge_directive.MergeDirective.from_objects(
-            tree_a.branch.repository, 'rev2a', 500, 36, tree_b.branch.base,
-            patch_type=None, public_branch=tree_a.branch.base)
+        md = self.from_objects( tree_a.branch.repository, 'rev2a', 500, 36,
+            tree_b.branch.base, patch_type=None,
+            public_branch=tree_a.branch.base)
         md.source_branch = '/dev/null'
         revision = md.install_revisions(tree_b.branch.repository)
         self.assertEqual('rev2a', revision)
+
+
+class TestMergeDirective1Branch(tests.TestCaseWithTransport,
+    TestMergeDirectiveBranch):
+    """Test merge directive format 1 with a branch"""
+
+    EMAIL1 = EMAIL1
+
+    EMAIL2 = EMAIL2
+
+    def from_objects(self, repository, revision_id, time, timezone,
+        target_branch, patch_type='bundle', local_target_branch=None,
+        public_branch=None, message=None):
+        return merge_directive.MergeDirective.from_objects(
+            repository, revision_id, time, timezone, target_branch,
+            patch_type, local_target_branch, public_branch, message)
+
+    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, patch_type=None,
+                 source_branch=None, message=None):
+        return merge_directive.MergeDirective(revision_id, testament_sha1,
+                 time, timezone, target_branch, patch, patch_type,
+                 source_branch, message)
+
+
+class TestMergeDirective2Branch(tests.TestCaseWithTransport,
+    TestMergeDirectiveBranch):
+    """Test merge directive format 2 with a branch"""
+
+    EMAIL1 = EMAIL1_2
+
+    EMAIL2 = EMAIL2_2
+
+    def from_objects(self, repository, revision_id, time, timezone,
+        target_branch, patch_type='bundle', local_target_branch=None,
+        public_branch=None, message=None, base_revision_id=None):
+        include_patch = (patch_type in ('bundle', 'diff'))
+        include_bundle = (patch_type == 'bundle')
+        assert patch_type in ('bundle', 'diff', None)
+        return merge_directive.MergeDirective2.from_objects(
+            repository, revision_id, time, timezone, target_branch,
+            include_patch, include_bundle, local_target_branch, public_branch,
+            message, base_revision_id)
+
+    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
+                 target_branch, patch=None, patch_type=None,
+                 source_branch=None, message=None, base_revision_id='null:'):
+        if patch_type == 'bundle':
+            bundle = patch
+            patch = None
+        else:
+            bundle = None
+        return merge_directive.MergeDirective2(revision_id, testament_sha1,
+            time, timezone, target_branch, patch, source_branch, message,
+            bundle, base_revision_id)
+
+    def test_base_revision(self):
+        tree_a, tree_b, branch_c = self.make_trees()
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 60,
+            tree_b.branch.base, patch_type='bundle',
+            public_branch=tree_a.branch.base, base_revision_id=None)
+        self.assertEqual('rev1', md.base_revision_id)
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 60,
+            tree_b.branch.base, patch_type='bundle',
+            public_branch=tree_a.branch.base, base_revision_id='null:')
+        self.assertEqual('null:', md.base_revision_id)
+        lines = md.to_lines()
+        md2 = merge_directive.MergeDirective.from_lines(lines)
+        self.assertEqual(md2.base_revision_id, md.base_revision_id)
+
+    def test_patch_verification(self):
+        tree_a, tree_b, branch_c = self.make_trees()
+        md = self.from_objects(tree_a.branch.repository, 'rev2a', 500, 60,
+            tree_b.branch.base, patch_type='bundle',
+            public_branch=tree_a.branch.base)
+        lines = md.to_lines()
+        md2 = merge_directive.MergeDirective.from_lines(lines)
+        md2._verify_patch(tree_a.branch.repository)
+        # Stript trailing whitespace
+        md2.patch = md2.patch.replace(' \n', '\n')
+        md2._verify_patch(tree_a.branch.repository)
+        # Convert to Mac line-endings
+        md2.patch = md2.patch.replace('\n', '\r')
+        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
+        # Convert to DOS line-endings
+        md2.patch = md2.patch.replace('\r', '\r\n')
+        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
+        md2.patch = md2.patch.replace('content_c', 'content_d')
+        self.assertFalse(md2._verify_patch(tree_a.branch.repository))

=== added file 'bzrlib/tests/test_multiparent.py'
--- bzrlib/tests/test_multiparent.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_multiparent.py	2007-06-26 20:05:54 +0000
@@ -0,0 +1,261 @@
+# Copyright (C) 2007 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+from unittest import TestCase
+
+from bzrlib import (
+    knit,
+    multiparent,
+    patiencediff,
+    tests,
+    )
+
+
+LINES_1 = "a\nb\nc\nd\ne\n".splitlines(True)
+LINES_2 = "a\nc\nd\ne\n".splitlines(True)
+LINES_3 = "a\nb\nc\nd\n".splitlines(True)
+LF_SPLIT_LINES = ['\x00\n', '\x00\r\x01\n', '\x02\r\xff']
+
+
+class Mock(object):
+
+    def __init__(self, **kwargs):
+        self.__dict__ = kwargs
+
+
+class TestMulti(TestCase):
+
+    def test_compare_no_parent(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1)
+        self.assertEqual([multiparent.NewText(LINES_1)], diff.hunks)
+
+    def test_compare_one_parent(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2])
+        self.assertEqual([multiparent.ParentText(0, 0, 0, 1),
+                          multiparent.NewText(['b\n']),
+                          multiparent.ParentText(0, 1, 2, 3)],
+                         diff.hunks)
+
+    def test_compare_two_parents(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
+        self.assertEqual([multiparent.ParentText(1, 0, 0, 4),
+                          multiparent.ParentText(0, 3, 4, 1)],
+                         diff.hunks)
+
+    def test_compare_two_parents_blocks(self):
+        matcher = patiencediff.PatienceSequenceMatcher(None, LINES_2, LINES_1)
+        blocks = matcher.get_matching_blocks()
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3],
+                                                  left_blocks=blocks)
+        self.assertEqual([multiparent.ParentText(1, 0, 0, 4),
+                          multiparent.ParentText(0, 3, 4, 1)],
+                         diff.hunks)
+
+    def test_range_iterator(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
+        diff.hunks.append(multiparent.NewText(['q\n']))
+        self.assertEqual([(0, 4, 'parent', (1, 0, 4)),
+                          (4, 5, 'parent', (0, 3, 4)),
+                          (5, 6, 'new', ['q\n'])],
+                         list(diff.range_iterator()))
+
+    def test_eq(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1)
+        diff2 = multiparent.MultiParent.from_lines(LINES_1)
+        self.assertEqual(diff, diff2)
+        diff3 = multiparent.MultiParent.from_lines(LINES_2)
+        self.assertFalse(diff == diff3)
+        self.assertFalse(diff == Mock(hunks=[multiparent.NewText(LINES_1)]))
+        self.assertEqual(multiparent.MultiParent(
+                         [multiparent.NewText(LINES_1),
+                          multiparent.ParentText(0, 1, 2, 3)]),
+                         multiparent.MultiParent(
+                         [multiparent.NewText(LINES_1),
+                          multiparent.ParentText(0, 1, 2, 3)]))
+
+    def test_to_patch(self):
+        self.assertEqual(['i 1\n', 'a\n', '\n', 'c 0 1 2 3\n'],
+            list(multiparent.MultiParent([multiparent.NewText(['a\n']),
+            multiparent.ParentText(0, 1, 2, 3)]).to_patch()))
+
+    def test_from_patch(self):
+        self.assertEqual(multiparent.MultiParent(
+            [multiparent.NewText(['a\n']),
+             multiparent.ParentText(0, 1, 2, 3)]),
+             multiparent.MultiParent.from_patch('i 1\na\n\nc 0 1 2 3'))
+        self.assertEqual(multiparent.MultiParent(
+            [multiparent.NewText(['a']),
+             multiparent.ParentText(0, 1, 2, 3)]),
+             multiparent.MultiParent.from_patch('i 1\na\nc 0 1 2 3\n'))
+
+    def test_binary_content(self):
+        patch = list(
+            multiparent.MultiParent.from_lines(LF_SPLIT_LINES).to_patch())
+        multiparent.MultiParent.from_patch(''.join(patch))
+
+    def test_make_patch_from_binary(self):
+        patch = multiparent.MultiParent.from_texts(''.join(LF_SPLIT_LINES))
+        expected = multiparent.MultiParent([
+            multiparent.NewText(LF_SPLIT_LINES)])
+        self.assertEqual(expected, patch)
+
+    def test_num_lines(self):
+        mp = multiparent.MultiParent([multiparent.NewText(['a\n'])])
+        self.assertEqual(1, mp.num_lines())
+        mp.hunks.append(multiparent.NewText(['b\n', 'c\n']))
+        self.assertEqual(3, mp.num_lines())
+        mp.hunks.append(multiparent.ParentText(0, 0, 3, 2))
+        self.assertEqual(5, mp.num_lines())
+        mp.hunks.append(multiparent.NewText(['f\n', 'g\n']))
+        self.assertEqual(7, mp.num_lines())
+
+    def test_to_lines(self):
+        mpdiff = multiparent.MultiParent.from_texts('a\nb\nc\n', ('b\nc\n',))
+        lines = mpdiff.to_lines(('b\ne\n',))
+        self.assertEqual(['a\n', 'b\n', 'e\n'], lines)
+
+
+class TestNewText(TestCase):
+
+    def test_eq(self):
+        self.assertEqual(multiparent.NewText([]), multiparent.NewText([]))
+        self.assertFalse(multiparent.NewText(['a']) ==
+                         multiparent.NewText(['b']))
+        self.assertFalse(multiparent.NewText(['a']) == Mock(lines=['a']))
+
+    def test_to_patch(self):
+        self.assertEqual(['i 0\n', '\n'],
+                         list(multiparent.NewText([]).to_patch()))
+        self.assertEqual(['i 1\n', 'a', '\n'],
+                         list(multiparent.NewText(['a']).to_patch()))
+        self.assertEqual(['i 1\n', 'a\n', '\n'],
+                         list(multiparent.NewText(['a\n']).to_patch()))
+
+
+class TestParentText(TestCase):
+
+    def test_eq(self):
+        self.assertEqual(multiparent.ParentText(1, 2, 3, 4),
+                         multiparent.ParentText(1, 2, 3, 4))
+        self.assertFalse(multiparent.ParentText(1, 2, 3, 4) ==
+                         multiparent.ParentText(2, 2, 3, 4))
+        self.assertFalse(multiparent.ParentText(1, 2, 3, 4) ==
+                         Mock(parent=1, parent_pos=2, child_pos=3,
+                              num_lines=4))
+
+    def test_to_patch(self):
+        self.assertEqual(['c 0 1 2 3\n'],
+                         list(multiparent.ParentText(0, 1, 2, 3).to_patch()))
+
+
+REV_A = ['a\n', 'b\n', 'c\n', 'd\n']
+REV_B = ['a\n', 'c\n', 'd\n', 'e\n']
+REV_C = ['a\n', 'b\n', 'e\n', 'f\n']
+
+
+class TestVersionedFile(TestCase):
+
+    def add_version(self, vf, text, version_id, parent_ids):
+        vf.add_version([(t+'\n') for t in text], version_id, parent_ids)
+
+    def make_vf(self):
+        vf = multiparent.MultiMemoryVersionedFile()
+        self.add_version(vf, 'abcd', 'rev-a', [])
+        self.add_version(vf, 'acde', 'rev-b', [])
+        self.add_version(vf, 'abef', 'rev-c', ['rev-a', 'rev-b'])
+        return vf
+
+    def test_add_version(self):
+        vf = self.make_vf()
+        self.assertEqual(REV_A, vf._lines['rev-a'])
+        vf.clear_cache()
+        self.assertEqual(vf._lines, {})
+
+    def test_get_line_list(self):
+        vf = self.make_vf()
+        vf.clear_cache()
+        self.assertEqual(REV_A, vf.get_line_list(['rev-a'])[0])
+        self.assertEqual([REV_B, REV_C], vf.get_line_list(['rev-b', 'rev-c']))
+
+    def test_reconstruct_empty(self):
+        vf = multiparent.MultiMemoryVersionedFile()
+        vf.add_version([], 'a', [])
+        self.assertEqual([], self.reconstruct_version(vf, 'a'))
+
+    @staticmethod
+    def reconstruct(vf, revision_id, start, end):
+        reconstructor = multiparent._Reconstructor(vf, vf._lines,
+                                                   vf._parents)
+        lines = []
+        reconstructor._reconstruct(lines, revision_id, start, end)
+        return lines
+
+    @staticmethod
+    def reconstruct_version(vf, revision_id):
+        reconstructor = multiparent._Reconstructor(vf, vf._lines,
+                                                   vf._parents)
+        lines = []
+        reconstructor.reconstruct_version(lines, revision_id)
+        return lines
+
+    def test_reconstructor(self):
+        vf = self.make_vf()
+        self.assertEqual(['a\n', 'b\n'], self.reconstruct(vf, 'rev-a',  0, 2))
+        self.assertEqual(['c\n', 'd\n'], self.reconstruct(vf, 'rev-a',  2, 4))
+        self.assertEqual(['e\n', 'f\n'], self.reconstruct(vf, 'rev-c',  2, 4))
+        self.assertEqual(['a\n', 'b\n', 'e\n', 'f\n'],
+                          self.reconstruct(vf, 'rev-c',  0, 4))
+        self.assertEqual(['a\n', 'b\n', 'e\n', 'f\n'],
+                          self.reconstruct_version(vf, 'rev-c'))
+
+    def test_reordered(self):
+        """Check for a corner case that requires re-starting the cursor"""
+        vf = multiparent.MultiMemoryVersionedFile()
+        # rev-b must have at least two hunks, so split a and b with c.
+        self.add_version(vf, 'c', 'rev-a', [])
+        self.add_version(vf, 'acb', 'rev-b', ['rev-a'])
+        # rev-c and rev-d must each have a line from a different rev-b hunk
+        self.add_version(vf, 'b', 'rev-c', ['rev-b'])
+        self.add_version(vf, 'a', 'rev-d', ['rev-b'])
+        # The lines from rev-c and rev-d must appear in the opposite order
+        self.add_version(vf, 'ba', 'rev-e', ['rev-c', 'rev-d'])
+        vf.clear_cache()
+        lines = vf.get_line_list(['rev-e'])[0]
+        self.assertEqual(['b\n', 'a\n'], lines)
+
+
+class TestMultiVersionedFile(tests.TestCaseInTempDir):
+
+    def test_save_load(self):
+        vf = multiparent.MultiVersionedFile('foop')
+        vf.add_version('a\nb\nc\nd'.splitlines(True), 'a', [])
+        vf.add_version('a\ne\nd\n'.splitlines(True), 'b', ['a'])
+        vf.save()
+        newvf = multiparent.MultiVersionedFile('foop')
+        newvf.load()
+        self.assertEqual('a\nb\nc\nd', ''.join(newvf.get_line_list(['a'])[0]))
+        self.assertEqual('a\ne\nd\n', ''.join(newvf.get_line_list(['b'])[0]))
+
+    def test_filenames(self):
+        vf = multiparent.MultiVersionedFile('foop')
+        vf.add_version('a\nb\nc\nd'.splitlines(True), 'a', [])
+        self.failUnlessExists('foop.mpknit')
+        self.failIfExists('foop.mpidx')
+        vf.save()
+        self.failUnlessExists('foop.mpidx')
+        vf.destroy()
+        self.failIfExists('foop.mpknit')
+        self.failIfExists('foop.mpidx')

=== modified file 'bzrlib/tests/test_read_bundle.py'
--- bzrlib/tests/test_read_bundle.py	2007-07-11 16:42:37 +0000
+++ bzrlib/tests/test_read_bundle.py	2007-07-17 13:27:14 +0000
@@ -69,8 +69,8 @@
             return
         info = bzrlib.bundle.read_bundle_from_url(
                     unicode(self.get_url('test_bundle')))
-        bundle_tree = info.revision_tree(wt.branch.repository, info.target)
-        self.assertEqual('commit-1', bundle_tree.revision_id)
+        revision = info.real_revisions[-1]
+        self.assertEqual('commit-1', revision.revision_id)
 
     def test_read_fail(self):
         # Trying to read from a directory, or non-bundle file

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2007-06-17 17:07:04 +0000
+++ bzrlib/tests/test_versionedfile.py	2007-06-23 05:14:25 +0000
@@ -247,6 +247,18 @@
         self.assertEqual(['line'], f.get_lines('eolbeforefirstparent'))
         #self.assertTrue(deltas['eolbeforefirstparent'] in expected_deltas)
 
+    def test_make_mpdiffs(self):
+        from bzrlib import multiparent
+        vf = self.get_file('foo')
+        sha1s = self._setup_for_deltas(vf)
+        new_vf = self.get_file('bar')
+        for version in multiparent.topo_iter(vf):
+            mpdiff = vf.make_mpdiffs([version])[0]
+            new_vf.add_mpdiffs([(version, vf.get_parents(version),
+                                 vf.get_sha1(version), mpdiff)])
+            self.assertEqualDiff(vf.get_text(version),
+                                 new_vf.get_text(version))
+
     def _setup_for_deltas(self, f):
         self.assertRaises(errors.RevisionNotPresent, f.get_delta, 'base')
         # add texts that should trip the knit maximum delta chain threshold
@@ -776,6 +788,11 @@
             '3f786850e387550fdab836ed7e6dc881de23001b', vf.get_sha1('b'))
         self.assertEqual(
             '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8', vf.get_sha1('c'))
+
+        self.assertEqual(['3f786850e387550fdab836ed7e6dc881de23001b',
+                          '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8',
+                          '3f786850e387550fdab836ed7e6dc881de23001b'],
+                          vf.get_sha1s(['a', 'c', 'b']))
         
 
 class TestWeave(TestCaseWithTransport, VersionedFileTestMixIn):

=== modified file 'bzrlib/tests/test_xml.py'
--- bzrlib/tests/test_xml.py	2007-02-23 05:47:26 +0000
+++ bzrlib/tests/test_xml.py	2007-06-26 19:31:00 +0000
@@ -20,6 +20,7 @@
     errors, 
     inventory, 
     xml7,
+    xml_serializer,
     )
 from bzrlib.tests import TestCase
 from bzrlib.inventory import Inventory, InventoryEntry
@@ -397,6 +398,16 @@
 
         self.assertEqual(len(expected), len(actual))
 
+    def test_registry(self):
+        self.assertIs(serializer_v4,
+                      xml_serializer.format_registry.get('4'))
+        self.assertIs(bzrlib.xml5.serializer_v5,
+                      xml_serializer.format_registry.get('5'))
+        self.assertIs(bzrlib.xml6.serializer_v6,
+                      xml_serializer.format_registry.get('6'))
+        self.assertIs(bzrlib.xml7.serializer_v7,
+                      xml_serializer.format_registry.get('7'))
+
 
 class TestEncodeAndEscape(TestCase):
     """Whitebox testing of the _encode_and_escape function."""

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2007-06-28 06:16:19 +0000
+++ bzrlib/versionedfile.py	2007-07-04 04:11:36 +0000
@@ -25,6 +25,7 @@
 from bzrlib import (
     errors,
     osutils,
+    multiparent,
     tsort,
     revision,
     ui,
@@ -32,6 +33,8 @@
 from bzrlib.transport.memory import MemoryTransport
 """)
 
+from cStringIO import StringIO
+
 from bzrlib.inter import InterObject
 from bzrlib.textmerge import TextMerge
 from bzrlib.symbol_versioning import (deprecated_function,
@@ -264,6 +267,50 @@
             result[version_id] = self.get_delta(version_id)
         return result
 
+    def make_mpdiffs(self, version_ids):
+        """Create multiparent diffs for specified versions"""
+        knit_versions = set()
+        for version_id in version_ids:
+            knit_versions.add(version_id)
+            knit_versions.update(self.get_parents(version_id))
+        lines = dict(zip(knit_versions,
+            self._get_lf_split_line_list(knit_versions)))
+        diffs = []
+        for version_id in version_ids:
+            target = lines[version_id]
+            parents = [lines[p] for p in self.get_parents(version_id)]
+            if len(parents) > 0:
+                left_parent_blocks = self._extract_blocks(version_id,
+                                                          parents[0], target)
+            else:
+                left_parent_blocks = None
+            diffs.append(multiparent.MultiParent.from_lines(target, parents,
+                         left_parent_blocks))
+        return diffs
+
+    def _extract_blocks(self, version_id, source, target):
+        return None
+
+    def add_mpdiffs(self, records):
+        """Add mpdiffs to this versionedfile
+
+        Records should be iterables of version, parents, expected_sha1,
+        mpdiff.  mpdiff should be a MultiParent instance.
+        """
+        vf_parents = {}
+        for version, parents, expected_sha1, mpdiff in records:
+            mpvf = multiparent.MultiMemoryVersionedFile()
+            needed_parents = [p for p in parents if not mpvf.has_version(p)]
+            parent_lines = self._get_lf_split_line_list(needed_parents)
+            for parent_id, lines in zip(needed_parents, parent_lines):
+                mpvf.add_version(lines, parent_id, [])
+            mpvf.add_diff(mpdiff, version, parents)
+            lines = mpvf.get_line_list([version])[0]
+            version_text = self.add_lines(version, parents, lines, vf_parents)
+            vf_parents[version] = version_text
+            if expected_sha1 != self.get_sha1(version):
+                raise errors.VersionedFileInvalidChecksum(version)
+
     def get_sha1(self, version_id):
         """Get the stored sha1 sum for the given revision.
         
@@ -271,6 +318,14 @@
         """
         raise NotImplementedError(self.get_sha1)
 
+    def get_sha1s(self, version_ids):
+        """Get the stored sha1 sums for the given revisions.
+
+        :param version_ids: The names of the versions to lookup
+        :return: a list of sha1s in order according to the version_ids
+        """
+        raise NotImplementedError(self.get_sha1)
+
     def get_suffixes(self):
         """Return the file suffixes associated with this versioned file."""
         raise NotImplementedError(self.get_suffixes)
@@ -300,6 +355,9 @@
         """
         raise NotImplementedError(self.get_lines)
 
+    def _get_lf_split_line_list(self, version_ids):
+        return [StringIO(t).readlines() for t in self.get_texts(version_ids)]
+
     def get_ancestry(self, version_ids, topo_sorted=True):
         """Return a list of all ancestors of given version(s). This
         will not include the null revision.

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2007-06-17 17:07:04 +0000
+++ bzrlib/weave.py	2007-06-23 05:14:25 +0000
@@ -891,6 +891,10 @@
         """See VersionedFile.get_sha1()."""
         return self._sha1s[self._lookup(version_id)]
 
+    def get_sha1s(self, version_ids):
+        """See VersionedFile.get_sha1s()."""
+        return [self._sha1s[self._lookup(v)] for v in version_ids]
+
     @deprecated_method(zero_eight)
     def numversions(self):
         """How many versions are in this weave?

=== modified file 'bzrlib/xml5.py'
--- bzrlib/xml5.py	2007-07-11 14:58:54 +0000
+++ bzrlib/xml5.py	2007-07-17 13:27:14 +0000
@@ -151,6 +151,7 @@
     # of the versionedfile, without doing XML parsing.
 
     supported_kinds = set(['file', 'directory', 'symlink'])
+    format_num = '5'
 
     def write_inventory_to_string(self, inv):
         """Just call write_inventory with a StringIO and return the value"""

=== modified file 'bzrlib/xml_serializer.py'
--- bzrlib/xml_serializer.py	2006-10-05 05:37:25 +0000
+++ bzrlib/xml_serializer.py	2007-06-26 19:31:00 +0000
@@ -22,6 +22,7 @@
 # importing this module is fairly slow because it has to load several
 # ElementTree bits
 
+from bzrlib import registry
 from bzrlib.trace import mutter, warning
 
 try:
@@ -155,3 +156,14 @@
         elementtree.ElementTree._raise_serialization_error(text)
 
 elementtree.ElementTree._escape_cdata = _escape_cdata
+
+
+class SerializerRegistry(registry.Registry):
+    """Registry for serializer objects"""
+
+
+format_registry = SerializerRegistry()
+format_registry.register_lazy('4', 'bzrlib.xml4', 'serializer_v4')
+format_registry.register_lazy('5', 'bzrlib.xml5', 'serializer_v5')
+format_registry.register_lazy('6', 'bzrlib.xml6', 'serializer_v6')
+format_registry.register_lazy('7', 'bzrlib.xml7', 'serializer_v7')

=== added file 'doc/developers/bundle-format4.txt'
--- doc/developers/bundle-format4.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/bundle-format4.txt	2007-07-04 23:38:26 +0000
@@ -0,0 +1,272 @@
+============================================
+Merge Directive format 2 and Bundle format 4
+============================================
+
+:Date: 2007-06-21
+
+Motivation
+----------
+Merge Directive format 2 represents a request to perform a certain merge.  It
+provides access to all the data necessary to perform that merge, by including
+a branch URL or a bundle payload.  It typically will include a preview of
+what applying the patch would do.
+
+Bundle Format 4 is designed to be a compact format for storing revision
+metadata that can be generated quickly and installed into a repository
+efficiently.  It is not intended to be human-readable.
+
+Note
+----
+These two formats, taken together, can be viewed as the successor of Bundle
+format 0.9, so their specifications are combined.  It is expected that in the
+future, bundle and merge-directive formats will vary independently.
+
+
+Bundle Format Name
+------------------
+This is the fourth bundle format to see public use.  Previous versions were
+0.7, 0.8, and 0.9.  Only 0.7's version number was aligned with a Bazaar
+release.
+
+
+Dependencies
+------------
+- Container format 1
+- Multiparent diffs
+- Bencode
+- Patch-RIO
+
+
+Description
+-----------
+Merge Directives fulfil the role previous bundle formats had of requesting a
+merge to be performed, but are a more flexible way of doing so.  With the
+introduction of these two formats, there is a clear split between "directive",
+which is a request to merge (and therefore signable), and "bundle", which is
+just data.
+
+Merge Directive format 2 may provide a patch preview of the change being
+requested.  If a preview is supplied, the receiving client will verify that
+the actual change matches the preview.
+
+Merge Directive format 2 also includes a testament hash, to ensure that if a
+branch is used, the branch cannot be subverted to cause the wrong changes to be
+applied.
+
+Bundle format 4 is designed to trade human-readability for speed and
+compactness.  It does not contain a human-readable "prelude" patch.
+
+Merge Directive 2 Contents
+--------------------------
+This format consists of three sections, in the following order.
+
+
+Patch-RIO command section
+~~~~~~~~~~~~~~~~~~~~~~~~~
+This section is identical to the corresponding section in Format 1 merge
+directives, except as noted below.  It is mandatory.  It is terminated by a
+line reading ``#`` that is not preceeded by a line ending with ``\``.
+
+In order to support cherry-picking and patch comparison, this format adds a new
+piece of information, the ``base_revision_id``.  This is a suggested base
+revision for merging.  It may be supplied by the user.  If not, it is
+calculated using the standard merge base algorithm, with the ``revision_id``
+and target branch's ``last_revision`` as its inputs.
+
+When merging, clients should use the ``base_revision_id`` when it is not
+already present in the ancestry of the ``last_revision`` of the target branch.
+If it is already present, clients should calculate a merge base in the normal
+way.
+
+
+Patch preview section
+~~~~~~~~~~~~~~~~~~~~~
+This section is optional.  It begins with the line ``# Begin patch``.  It is
+terminated by the end-of-file or by the beginning of a bundle section.
+
+Its contents are a unified diff, as per the ``bzr diff`` command.  The FROM
+revision is the ``base_revision_id`` specified in the Patch-RIO section.
+
+
+Bundle section
+~~~~~~~~~~~~~~
+This section is optional, but if it is not supplied, a source_branch must be
+supplied.  It begins with the line ``# Begin bundle``, and is terminated by the
+end-of-file.
+
+The contents are a base-64 encoded bundle.  This may be any bundle format, but
+formats 4+ are strongly recommended.  The base revision is the newest revision
+in the source branch which is an ancestor of all revisions not present in
+target which are ancestors of revision_id.
+
+This base revision may or may not be the same as the ``base_revision_id``.  In
+particular, the ``base_revision_id`` may specify a cherry-pick, but all the
+ancestors of the ``base_revision_id`` should be installed in the target
+repository before performing such a merge.
+
+
+Bundle 4 Contents
+-----------------
+Bazaar revision bundles  begin with a format marker that reads
+``# Bazaar revision bundle v4`` in plaintext.  The remainder of the file is a
+``Bazaar pack format 1`` container.  The container is compressed using bzip2.
+
+Putting the format marker in plaintext ensures that old clients will give good
+diagnostics, but renders the file unreadable by standard bzip2 utilities.
+
+Serialization
+~~~~~~~~~~~~~
+Format 4 records revision and inventory records in their repository
+serialization format.  This minimizes translation and compression costs
+in the common case, where the sender and receiver use the same serialization
+format for their repository. Steps have been taken to ensure a faithful
+conversion when serialization formats are mismatched.
+
+
+Bundle Records
+~~~~~~~~~~~~~~
+The bundle format creates a single bundle-level record out of two container
+records.  The first container record contains metainfo as a Bencoded dict.  The
+second container record contains the body.
+
+The bundle record name is associated with the metainfo record.  The body record
+is anonymous.
+
+
+Record metainfo
+~~~~~~~~~~~~~~~
+
+:record_kind: The storage strategy of the record.  May be ``fulltext`` (the
+    record body contains the full text of the value), ``mpdiff`` (the record
+    body contains a multi-parent diff of the value), or ``header`` (no record
+    body).
+:parents: Used in fulltext and mpdiff records.  The revisions that should be
+    noted as parents of this revision in the repository.  For mpdiffs, this is
+    also the list of build-parents.
+:sha1: Used in mpdiff records.  The sha-1 hash of the full-text value.
+
+
+Bundle record naming
+~~~~~~~~~~~~~~~~~~~~~
+All bundle records have a single name, which is associated with the metainfo
+container record.  Records are named according to the body's content-kind,
+revision-id, and file-id.
+
+Content-kind may be one of:
+
+:file: a version of a user file
+:inventory: the tree inventory
+:revision: the revision metadata for a revision
+:signature: the revision signature for a revision
+
+Names are constructed like so: ``content-kind/revision-id/file-id``.  Values
+are iterpreted left-to-right, so if two values are present, they are
+content-kind and revision-id.
+A record has a file-id if-and-only-if it is a file record.
+Info records have no revision or file-id.
+Inventory, revision and signature all have content-kind and revision-id, but
+no file-id.
+
+Layout
+~~~~~~
+The first record is an info/header record.
+
+The subsequent records are mpdiff file records.  The are ordered first by file
+id, then in topological order by revision-id.
+
+The next records are mpdiff inventory records.  They are topologically sorted.
+
+The next records are revision and signature fulltexts.  They are interleaved
+and topologically sorted.
+
+Info record
+~~~~~~~~~~~
+The info record has type ``header``.  It has no revision_id or file_id.
+Its metadata contains:
+
+:serializer: A string describing the serialization format used for inventory
+    and revision data.  May be ``xml5``, ``xml6`` or ``xml7``.
+:supports_rich_root: 1 if the source repository supports rich roots,
+    0 otherwise.
+
+
+Implementation notes
+~~~~~~~~~~~~~~~~~~~~
+- knit deltas contain almost enough information to extract the original
+  SequenceMatcher.get_matching_blocks() call used to produce them.  Combining
+  that information with the relevant fulltexts allows us to avoid performing
+  sequence matching on any fulltexts for which we have deltas.
+
+- MultiParent deltas contain ``get_matching_blocks`` output almost verbatim,
+  but if there is more than one parent, the information about the leftmost
+  parent may be incomplete.  However, for single-parent multiparent diffs, we
+  can extract the ``SequenceMatcher.get_matching_blocks`` output, and therefore
+  ``the SequenceMatcher.get_opcodes`` output used to create knit deltas.
+
+
+Installing data across serialization mismatches
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+In practice, there cannot be revision serialization mismatches, because the
+serialization of revisions has been consistent in serializations 5-7
+
+If there is a mismatch in inventory serialization formats, the receiver can
+
+  1. extract the inventory objects for the parents
+  2. serialize them using the bundle serialize
+  3. apply the mpdiff
+  4. calculate the fulltext sha1
+  5. compare the calculated sha1 to the expected sha1
+  6. deserialize using the bundle serializer
+  7. serialize using the repository serializer
+  8. add to the repository
+
+This is much slower, of course.  But since the since the fulltext is verified
+at step 5, it should be just as safe as any other conversion.
+
+Model differences
+~~~~~~~~~~~~~~~~~
+
+Note that there may be model differences requiring additional changes.  These
+differences are described by the "supports_rich_root" value in the info record.
+
+A subset of xml6 and xml7 records are compatible with xml5 (i.e. those that
+were converted from xml5 originally).
+
+When installing from a bundle whose serializer supports tree references to a
+repository that does not support tree references, clients should halt if they
+encounter a record containing a tree reference.
+
+When installing from a supports_rich_root bundle to a repository that does not
+support rich roots, clients should halt if they encounter an inventory record
+whose root directory revision-id does not match the inventory revision id.
+
+When installing from a bundle that does not support rich roots to a repository
+that does, additional knits should be added for the root directory, with a
+revision for each inventory revision.
+
+Validating preview patches
+~~~~~~~~~~~~~~~~~~~~~~~~~~
+When applying a merge directive that includes a preview, clients should
+verify that the preview matches the changes requested by the merge directive.
+
+In order to do this, the client should generate a diff from the
+``base_revision_id`` to the ``revision_id``.  This diff should be compared
+against the preview patch, making allowances for the fact that whitespace
+munging may have occurred.
+
+One form of whitespace munging that has been observed is line-ending
+conversion.  Certain mail clients such as Evolution do not respect the
+line-endings of text attachments.  Since line-ending conversion is unlikely to
+alter the meaning of a patch, it seems safe to ignore line endings when
+comparing the preview patch.
+
+Another form of whitespace munging that has been observed is
+trailing-whitespace stripping.  Again, it seems unlikely that stripping
+trailing whitespace could alter the meaning of a patch.  Such a distinction
+is also invisible to readers, so ignoring it does not create a new threat.  So
+it seems reasonable to ignore trailing whitespace when comparing the patches.
+
+Other mungings are possible, but it is recommended not to implement support
+for them until they have been observed.  Each of these changes makes the
+comparison more approximate, and the more approximate it becomes, the easier it
+is to provide a preview patch that does not match the requested changes.

=== modified file 'doc/developers/bundles.txt'
--- doc/developers/bundles.txt	2007-06-21 03:06:32 +0000
+++ doc/developers/bundles.txt	2007-07-04 23:38:26 +0000
@@ -41,8 +41,10 @@
 
 June 2007 Design
 ================
-
-To be documented.
+`Bundle Format 4 spec`_
+
+.. _Bundle Format 4 spec: bundle-format4.htm
+
 
 Future Plans
 ============

